/*-
 * Copyright 2008 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/module.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/selinfo.h>
#include <sys/poll.h>
#include <sys/fcntl.h>
#include <dev/sound/pcm/sound.h>
#include "mixer_if.h"
#include "cx88_common.h"
#include "cx88_if.h"
#include "iicdev.h"
#include "cx88_i2c.h"
#include "cx88_i2c_if.h"
#include "cx88_regs.h"
#include "cx88_audio_regs.h"
#include "cx88_audio_ioctl.h"

#define CX88_AUDIO_MINVER 1
#define CX88_AUDIO_MODVER 1
#define CX88_AUDIO_MAXVER 1

#define MIN_AUDIO_BUFFER_SIZE     (1 << 12)
#define DEFAULT_AUDIO_BUFFER_SIZE (1 << 14)
#define MAX_AUDIO_BUFFER_SIZE     (1 << 16)
#define CX88_AUDIO_SPEED 48000

/* Maximum line size, in bytes (dictated by capacity of line length register) */
#define MAX_AUDIO_LINE_SIZE 0xFFF

/* Driver states */
typedef enum
{
   CX88_AUDIO_INVALID = 0, /* Uninitialized */
   CX88_AUDIO_UNCONFIGURED, /* Not configured or formatted */
   CX88_AUDIO_CONFIGURED, /* Configured (I2C devices, buffers, etc.) */
   CX88_AUDIO_CAPTURING, /* Configured and capturing */
   CX88_AUDIO_PCM
} cx88_audio_state_t;

typedef struct
{
   cx88_common_softc common;
   cx88_channel channel;
   cx88_i2c_softc i2c;
   device_t pcm_dev;
   struct pcm_channel *pcm_chan;
   struct snd_dbuf *pcm_buf;
   struct cdev *devnode;
   struct mtx isrlock;
   struct sx sxlock;
   struct selinfo selectors;
   uint8_t broadcast_format;
   boolean_t in_use;
   boolean_t reconfig_requested;
   boolean_t user_configured;
   cx88_audio_state_t state;
   boolean_t data_available; /* Indicates an interrupt has completed and data is available */
   volatile uint8_t isr_status; /* Error status from interrupt service routine */
   volatile uint8_t buffer_available;
   uint8_t stereo_format;
   uint8_t second_lang;
   uint8_t input;
   uint8_t i2s_config;
   uint8_t volume_reg;
   boolean_t mute;
   int8_t balance_reg;
   cx88_audio_volume volume;
   unsigned int pcm_bufsize;
   uint32_t pcm_blocksize;
   void *pcm_databuf;
} cx88_audio_softc;

static uint32_t cx88_audio_formats[] = 
{
   AFMT_S16_LE,
   AFMT_STEREO | AFMT_S16_LE,
   0
};

static struct pcmchan_caps cx88_audio_caps = 
{
   .minspeed = CX88_AUDIO_SPEED,
   .maxspeed = CX88_AUDIO_SPEED,
   .fmtlist = cx88_audio_formats, 
   .caps = 0
};

static int cx88_audio_probe(device_t dev);
static int cx88_audio_attach(device_t dev);
static int cx88_audio_setup(void *arg);
static int cx88_audio_teardown(void *arg);
static void cx88_audio_intr(void *arg);
static cx88_i2c_bus *cx88_audio_get_i2c_bus(device_t dev);
static cx88_i2c_dev *cx88_audio_get_i2c_dev(device_t dev, uint8_t addr);
static void cx88_audio_reconfigure(cx88_audio_softc *sc);
static int cx88_audio_request_reconfig(device_t dev);
static int cx88_audio_dmaprog(cx88_buffer *buffer, uint8_t bufindex);
static void cx88_audio_reset(cx88_audio_softc *sc, cx88_audio_state_t state);
static int cx88_audio_configure(
   cx88_audio_softc *sc,
   cx88_audio_config *conf,
   uint32_t blocksize,
   struct thread *context);
static void cx88_audio_close_session(cx88_audio_softc *sc);
static int cx88_audio_load_format(cx88_audio_softc *sc);
static int cx88_audio_setformat(cx88_audio_softc *sc, cx88_audio_format *fmt);
static int cx88_audio_start_capture(cx88_audio_softc *sc);
static int cx88_audio_transfer(cx88_audio_softc *sc, cx88_audio_wait *wait, boolean_t nonblocking);
static int cx88_audio_set_volume(cx88_audio_softc *sc, cx88_audio_volume *vol);
static void cx88_audio_mute(cx88_audio_softc *sc);
static void cx88_audio_get_status(cx88_audio_softc *sc, cx88_audio_status *status);
static int cx88_audio_fm_radio(cx88_audio_softc *sc, uint32_t *ctl_reg);
static int cx88_audio_btsc(cx88_audio_softc *sc, uint32_t *ctl_reg);
static int cx88_audio_eiaj(cx88_audio_softc *sc, uint32_t *ctl_reg);
static int cx88_audio_a2(cx88_audio_softc *sc, uint32_t *ctl_reg);
static int cx88_audio_nicam(cx88_audio_softc *sc, uint32_t *ctl_reg);
static int cx88_audio_shutdown(device_t dev);
static int cx88_audio_suspend(device_t dev);

static const cx88_client_driver cx88_audio_client = 
{
   .softc_size               = sizeof(cx88_audio_softc),
   .function                 = CX88_FUNC_AUDIO,
   .setup                    = cx88_audio_setup,
   .teardown                 = cx88_audio_teardown,
   .interrupt_handler        = cx88_audio_intr,
   .description              = "Conexant CX2388x Analog Audio"
};

static devclass_t cx88_audio_devclass;

static d_open_t cx88_audio_open;
static d_close_t cx88_audio_close;
static d_mmap_t cx88_audio_mmap;
static d_ioctl_t cx88_audio_ioctl;
static d_poll_t cx88_audio_poll;

static struct cdevsw cx88_audio_cdevsw =
{
   .d_version = D_VERSION,
   .d_flags = D_TRACKCLOSE,
   .d_name = "cx88audio",
   .d_open = cx88_audio_open,   /* open */
   .d_close = cx88_audio_close, /* close */
   .d_mmap = cx88_audio_mmap,   /* mmap */
   .d_ioctl = cx88_audio_ioctl, /* ioctl */
   .d_poll = cx88_audio_poll,   /* poll */
};

typedef int (*broadcast_format_func) (cx88_audio_softc*, uint32_t*);

typedef struct
{
   uint32_t init_reg;
   broadcast_format_func set_format;
} cx88_audio_broadcast_format;

static cx88_audio_broadcast_format broadcast_formats [] = 
{
   /* init_reg             set_format */
   {AUDIO_INIT_FM_RADIO,   cx88_audio_fm_radio}, /* CX88_AUDIO_FMT_FM_RADIO */
   {AUDIO_INIT_FM_RADIO,   cx88_audio_fm_radio}, /* CX88_AUDIO_FMT_FM_RADIO_50 */
   {AUDIO_INIT_FM_RADIO,   cx88_audio_fm_radio}, /* CX88_AUDIO_FMT_FM_RADIO_75 */
   {AUDIO_INIT_BTSC,       cx88_audio_fm_radio}, /* CX88_AUDIO_FMT_BTSC_FM */
   {AUDIO_INIT_BTSC,       cx88_audio_btsc},     /* CX88_AUDIO_FMT_BTSC */
   {AUDIO_INIT_EIAJ,       cx88_audio_eiaj},     /* CX88_AUDIO_FMT_EIAJ */
   {AUDIO_INIT_A2,         cx88_audio_a2},       /* CX88_AUDIO_FMT_AM_L */
   {AUDIO_INIT_A2,         cx88_audio_a2},       /* CX88_AUDIO_FMT_A2_BG */
   {AUDIO_INIT_A2,         cx88_audio_a2},       /* CX88_AUDIO_FMT_A2_DK */
   {AUDIO_INIT_A2,         cx88_audio_a2},       /* CX88_AUDIO_FMT_A1_I */
   {AUDIO_INIT_NICAM,      cx88_audio_nicam},    /* CX88_AUDIO_FMT_NICAM_BGDK */
   {AUDIO_INIT_NICAM,      cx88_audio_nicam},    /* CX88_AUDIO_FMT_NICAM_I */
   {AUDIO_INIT_NICAM,      cx88_audio_nicam}     /* CX88_AUDIO_FMT_NICAM_L */
};

static int cx88_audio_setup(void *arg)
{   
   cx88_audio_softc *sc = arg;
   int error = 0;
   sc->reconfig_requested = FALSE;
   sc->in_use = FALSE;
   sc->user_configured = FALSE;
   sc->pcm_dev = NULL;
   sc->pcm_chan = NULL;
   sc->pcm_bufsize = 0;
   sc->data_available = FALSE;
   sc->buffer_available = 0;
   sc->broadcast_format = CX88_AUDIO_FMT_FM_RADIO;
   sc->stereo_format = CX88_AUDIO_STEREO_AUTO;
   sc->input = CX88_AUDIO_INPUT_BCAST;
   sc->i2s_config = 0;
   sc->second_lang = 0;
   sc->volume_reg = 0;
   sc->mute = FALSE;
   sc->balance_reg = 0;
   mtx_init(&(sc->isrlock), "CX88 Audio ISR lock", NULL, MTX_DEF);
   sx_init(&(sc->sxlock), "CX88 Audio SX lock");
   sc->volume.in_left = 0;
   sc->volume.in_right = 0;
   if ((error = cx88_init_channel(&(sc->channel), &(sc->common),
      CX88_CH_AUDIO_IN, MAX_AUDIO_BUFFER_SIZE)))
   {
      goto fail0;
   }
   /* Create I2C bus instance */
   if ((error = cx88_i2c_attach(&(sc->i2c), &(sc->common))))
   {
      goto fail1;
   }
   sc->pcm_dev = device_add_child(sc->common.dev, "pcm", -1);
   if (sc->pcm_dev == NULL)
   {
      error = ENXIO;
      goto fail2;
   }
   if ((error = device_probe_and_attach(sc->pcm_dev)))
   {
      goto fail3;  
   }
   /* Create device file--this should be done as the last step in initialization */
   sc->devnode = make_dev(&cx88_audio_cdevsw, device_get_unit(sc->common.dev), UID_ROOT, GID_WHEEL, 0644, device_get_nameunit(sc->common.dev));
   if (sc->devnode == NULL)
   {
      goto fail4;
   }
   cx88_audio_reset(sc, CX88_AUDIO_UNCONFIGURED);
   goto success;
   fail4:
      device_detach(sc->pcm_dev);
   fail3:
      device_delete_child(sc->common.dev, sc->pcm_dev);
   fail2:
      cx88_i2c_detach(&(sc->i2c));
   fail1:
      cx88_destroy_channel(&(sc->channel));  
   fail0:
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
   success:
      return error;
}

static int cx88_audio_teardown(void *arg)
{
   int error = 0;
   cx88_audio_softc *sc = arg;
   sx_slock(&(sc->sxlock));
   cx88_audio_reset(sc, CX88_AUDIO_INVALID);
   sx_sunlock(&(sc->sxlock));
   sx_xlock(&(sc->sxlock));
   cx88_release_channel(&(sc->channel));
   sc->pcm_blocksize = 0;
   sx_xunlock(&(sc->sxlock));
   if (sc->devnode != NULL)
   {
      destroy_dev(sc->devnode);
      sc->devnode = NULL;  
   }
   if (sc->pcm_dev != NULL)
   {
      error = device_detach(sc->pcm_dev);
      if (!error)
      {
         device_delete_child(sc->common.dev, sc->pcm_dev);
         sc->pcm_dev = NULL;
      }
   }
   if (!error)
   {
      error = cx88_i2c_detach(&(sc->i2c));
   }
   if (!error)
   {
      cx88_destroy_channel (&(sc->channel));
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
   }
   return error;
}

static void cx88_audio_reset(cx88_audio_softc *sc, cx88_audio_state_t state)
{
   mtx_lock(&(sc->isrlock));
   /* Mute all audio */
   cx88_write_reg(&(sc->common), 
      AUDIO_REG_VOL_CTL, VOL_CTL_SRC_MUTE | VOL_CTL_I2S_MUTE | VOL_CTL_DAC_MUTE);
   /* Stop any data transfer */
   cx88_write_reg(&(sc->common), AUDIO_REG_DMA_CONTROL, 0);
   /* Disable interrupts */
   cx88_write_reg(&(sc->common), AUDIO_REG_INT_MASK, 0);
   /* Disable chroma ADC gain */
   cx88_write_reg(&(sc->common), AUDIO_REG_AFE_12DB, 0);
   
   /* Reset capture status values */
   sc->isr_status = CX88_AUDIO_WAIT_ABORTED;
   
   /* Our state value should never be higher than CX88_AUDIO_CONFIGURED, as we won't be capturing once reset is done */
   sc->data_available = FALSE;
   sc->state = CX88_AUDIO_CONFIGURED;
   if (state < sc->state)
   {
      /* Reset alone should never raise the state value */
      sc->state = state;  
   }
   
   wakeup(&(sc->data_available));
   selwakeup(&(sc->selectors));
   mtx_unlock(&(sc->isrlock));
   
   /* Protect access to the shared PCI interrupt register. */
   cx88_lock_board(&(sc->common));
   cx88_write_reg(&(sc->common), CX88_REG_PCI_INTMASK,
      (cx88_read_reg(&(sc->common), CX88_REG_PCI_INTMASK) & ~CX88_PCI_INT_AUDIO));
   cx88_unlock_board(&(sc->common));
   
}

static int cx88_audio_transfer(cx88_audio_softc *sc, cx88_audio_wait *wait, boolean_t nonblocking)
{
   int error = 0;
   mtx_lock(&(sc->isrlock));
   if (sc->state != CX88_AUDIO_CAPTURING)
   {
      mtx_unlock(&(sc->isrlock));
      return ENXIO;
   }
   while (!error && !(sc->isr_status & CX88_AUDIO_CAP_FATAL) && !(sc->data_available))
   {
      if (!nonblocking)
      {
         error = msleep(&(sc->data_available), &(sc->isrlock), PZERO | PCATCH, 
            "cx88au", wait->in_timeout_ms);
      }
      else
      {
         error = EWOULDBLOCK;
      }
   }
   if (!error)
   {
      sc->data_available = FALSE;
   }
   wait->out_status = sc->isr_status;
   wait->out_buffer_available = sc->buffer_available;
   wait->out_stereo_format = sc->stereo_format;
   if (wait->out_stereo_format == CX88_AUDIO_STEREO_AUTO)
   {
      uint32_t stereo_mode = cx88_read_field(
         cx88_read_reg(&(sc->common), AUDIO_REG_STATUS), AUDIO_STATUS_STEREO_MODE);
      if (stereo_mode == AUDIO_STEREO_MODE_STEREO)
      {
         wait->out_stereo_format = CX88_AUDIO_STEREO_ON;
      }
      else
      {
         wait->out_stereo_format = CX88_AUDIO_STEREO_OFF;  
      }
   }
   sc->isr_status = CX88_AUDIO_CAP_NOERR;
   uint32_t stat = cx88_read_reg(&(sc->common), AUDIO_REG_INT_STAT);
   if (stat & AUDIO_INT_BER)
   {
      wait->out_status |= CX88_AUDIO_BERTHRESH;
   }
   if (stat & AUDIO_INT_IN_OVERFLOW)
   {
      wait->out_status |= CX88_AUDIO_OVERFLOW;
   }
   mtx_unlock(&(sc->isrlock));
   /* For blocking I/O, msleep() timed out */
   if ((error == EWOULDBLOCK) && wait->in_timeout_ms)
   {
      wait->out_status |= CX88_AUDIO_CAP_TIMEOUT;
      error = 0;
   }
   else if (error)
   {
      wait->out_status |= CX88_AUDIO_WAIT_ABORTED;
   }
   return error;
}  

static int cx88_audio_fm_radio(cx88_audio_softc *sc, uint32_t *ctl_reg)
{
   uint32_t src_sel = 0;
   switch (sc->stereo_format)
   {
      case CX88_AUDIO_STEREO_OFF:
         src_sel = AUD_SRC_FM_RADIO_FORCE_MONO;
         break;
      case CX88_AUDIO_STEREO_ON:
         src_sel = AUD_SRC_FM_RADIO_FORCE_STEREO;
         break;
      case CX88_AUDIO_STEREO_AUTO:
         src_sel = AUD_SRC_FM_RADIO_AUTO_STEREO;
         break;
   }
   *ctl_reg = cx88_update_field(*ctl_reg, AUD_CTL_SRC_SEL, src_sel);
   static const cx88_reg regs_fm_radio_75[] = 
   {
      {AUDIO_REG_DEEMPH0_G0, 0x0000091B},
      {AUDIO_REG_DEEMPH0_A0, 0x00006B68},
      {AUDIO_REG_DEEMPH0_B0, 0x000011EC},
      {AUDIO_REG_DEEMPH1_G0, 0x00000AA0},
      {AUDIO_REG_DEEMPH1_A0, 0x00006B68},
      {AUDIO_REG_DEEMPH1_B0, 0x000011EC},
   };
   static const cx88_reg regs_fm_radio_50[] = 
   {
      {AUDIO_REG_DEEMPH0_G0, 0x00000C45},
      {AUDIO_REG_DEEMPH0_A0, 0x00006262},
      {AUDIO_REG_DEEMPH0_B0, 0x00001C29},
      {AUDIO_REG_DEEMPH1_G0, 0x00000D80},
      {AUDIO_REG_DEEMPH1_A0, 0x00006262},
      {AUDIO_REG_DEEMPH1_B0, 0x00001C29}
   };
   if ((sc->broadcast_format == CX88_AUDIO_FMT_BTSC_FM) ||
       (sc->broadcast_format == CX88_AUDIO_FMT_FM_RADIO_75))
   {
      cx88_write_regs(&(sc->common), regs_fm_radio_75, 
         sizeof(regs_fm_radio_75) / sizeof(cx88_reg));
   }
   else if (sc->broadcast_format == CX88_AUDIO_FMT_FM_RADIO_50)
   {
      cx88_write_regs(&(sc->common), regs_fm_radio_50,
         sizeof(regs_fm_radio_50) / sizeof(cx88_reg));  
   }
   cx88_write_reg(&(sc->common), AUDIO_REG_DEEMPH1_A1, 0x0003FC66);
   cx88_write_reg(&(sc->common), AUDIO_REG_DEEMPH1_B1, 0x0000399A);
   cx88_write_reg(&(sc->common), AUDIO_REG_POLYPH80SCALEFAC, 3);
   return 0;
}

static int cx88_audio_btsc(cx88_audio_softc *sc, uint32_t *ctl_reg)
{
   uint32_t src_sel = 0;
   if (sc->second_lang)
   {
      src_sel = AUD_SRC_BTSC_FORCE_SAP;
   }
   else
   {
      switch (sc->stereo_format)
      {
         case CX88_AUDIO_STEREO_OFF:
            src_sel = AUD_SRC_BTSC_FORCE_MONO;
            break;
         case CX88_AUDIO_STEREO_ON:
            src_sel = AUD_SRC_BTSC_FORCE_STEREO;
            break;
         case CX88_AUDIO_STEREO_AUTO:
            src_sel = AUD_SRC_BTSC_AUTO_STEREO;
            break;
      }
   }
   *ctl_reg = cx88_update_field(*ctl_reg, AUD_CTL_SRC_SEL, src_sel);
   *ctl_reg |= AUD_CTL_FMRADIO_RDS;
   static const cx88_reg regs_btsc[] =
   {
      {AUDIO_REG_OUT1_SEL, 0X00000013},
      {AUDIO_REG_OUT1_SHIFT, 0X00000000},
      {AUDIO_REG_POLY0_DDS_CONSTANT, 0X0012010C},
      {AUDIO_REG_DMD_RA_DDS, 0X00C3E7AA},
      {AUDIO_REG_DBX_IN_GAIN, 0X00004734},
      {AUDIO_REG_DBX_WBE_GAIN, 0X00004640},
      {AUDIO_REG_DBX_SE_GAIN, 0X00008D31},
      {AUDIO_REG_DCOC_0_SRC, 0X0000001A},
      {AUDIO_REG_IIR1_4_SEL, 0x00000021},
      {AUDIO_REG_DCOC_PASS_IN, 0x00000003},
      {AUDIO_REG_DCOC_0_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_0_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_DCOC_1_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_1_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_DN0_FREQ, 0x0000283B},
      {AUDIO_REG_DN2_SRC_SEL, 0x00000008},
      {AUDIO_REG_DN2_FREQ, 0x00003000},
      {AUDIO_REG_DN2_AFC, 0x00000002},
      {AUDIO_REG_DN2_SHFT, 0x00000000},
      {AUDIO_REG_IIR2_2_SEL, 0x00000020},
      {AUDIO_REG_IIR2_2_SHIFT, 0x00000000},
      {AUDIO_REG_IIR2_3_SEL, 0x0000001F},
      {AUDIO_REG_IIR2_3_SHIFT, 0x00000000},
      {AUDIO_REG_CRDC1_SRC_SEL, 0x000003CE},
      {AUDIO_REG_CRDC1_SHIFT, 0x00000000},
      {AUDIO_REG_CORDIC_SHIFT_1, 0x00000007},
      {AUDIO_REG_DCOC_1_SRC, 0x0000001B},
      {AUDIO_REG_DCOC1_SHIFT, 0x00000000},
      {AUDIO_REG_RDSI_SEL, 0X00000008},
      {AUDIO_REG_RDSQ_SEL, 0X00000008},
      {AUDIO_REG_RDSI_SHIFT, 0X00000000},
      {AUDIO_REG_RDSQ_SHIFT, 0X00000000},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000003}
   };
   static const cx88_reg regs_btsc_sap[] = 
   {
      {AUDIO_REG_DBX_IN_GAIN, 0x00007200},
      {AUDIO_REG_DBX_WBE_GAIN, 0x00006200},
      {AUDIO_REG_DBX_SE_GAIN, 0x00006200},
      {AUDIO_REG_IIR1_1_SEL, 0x00000000},
      {AUDIO_REG_IIR1_3_SEL, 0x00000001},
      {AUDIO_REG_DN1_SRC_SEL, 0x00000007},
      {AUDIO_REG_IIR1_4_SHIFT, 0x00000006},
      {AUDIO_REG_IIR2_1_SHIFT, 0x00000000},
      {AUDIO_REG_IIR2_2_SHIFT, 0x00000000},
      {AUDIO_REG_IIR3_0_SHIFT, 0x00000000},
      {AUDIO_REG_IIR3_1_SHIFT, 0x00000000},
      {AUDIO_REG_IIR3_0_SEL, 0x0000000D},
      {AUDIO_REG_IIR3_1_SEL, 0x0000000E},
      {AUDIO_REG_DEEMPH1_SRC_SEL, 0x00000014},
      {AUDIO_REG_DEEMPH1_SHIFT, 0x00000000},
      {AUDIO_REG_DEEMPH1_G0, 0x00004000},
      {AUDIO_REG_DEEMPH1_A0, 0x00000000},
      {AUDIO_REG_DEEMPH1_B0, 0x00000000},
      {AUDIO_REG_DEEMPH1_A1, 0x00000000},
      {AUDIO_REG_DEEMPH1_B1, 0x00000000},
      {AUDIO_REG_OUT0_SEL, 0x0000003F},
      {AUDIO_REG_OUT1_SEL, 0x0000003F},
      {AUDIO_REG_DN1_AFC, 0x00000002},
      {AUDIO_REG_DCOC_0_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_0_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_DCOC_1_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_1_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_IIR1_0_SEL, 0x0000001D},
      {AUDIO_REG_IIR1_2_SEL, 0x0000001E},
      {AUDIO_REG_IIR2_1_SEL, 0x00000002},
      {AUDIO_REG_IIR2_2_SEL, 0x00000004},
      {AUDIO_REG_IIR3_2_SEL, 0x0000000F},
      {AUDIO_REG_DCOC2_SHIFT, 0x00000001},
      {AUDIO_REG_IIR3_2_SHIFT, 0x00000001},
      {AUDIO_REG_DEEMPH0_SRC_SEL, 0x00000014},
      {AUDIO_REG_CORDIC_SHIFT_1, 0x00000006},
      {AUDIO_REG_POLY0_DDS_CONSTANT, 0x000E4DB2},
      {AUDIO_REG_DMD_RA_DDS, 0x00F696E6},
      {AUDIO_REG_IIR2_3_SEL, 0x00000025},
      {AUDIO_REG_IIR1_4_SEL, 0x00000021},
      {AUDIO_REG_DN1_FREQ, 0x0000C965},
      {AUDIO_REG_DCOC_PASS_IN, 0x00000003},
      {AUDIO_REG_DCOC_0_SRC, 0x0000001A},
      {AUDIO_REG_DCOC_1_SRC, 0x0000001B},
      {AUDIO_REG_DCOC1_SHIFT, 0x00000000},
      {AUDIO_REG_RDSI_SEL, 0x00000009},
      {AUDIO_REG_RDSQ_SEL, 0x00000009},
      {AUDIO_REG_RDSI_SHIFT, 0x00000000},
      {AUDIO_REG_RDSQ_SHIFT, 0x00000000},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000003}
   };
   if (sc->second_lang)
   {
      cx88_write_regs(&(sc->common), regs_btsc_sap,
         sizeof(regs_btsc_sap) / sizeof(cx88_reg));  
   }
   else
   {
      cx88_write_regs(&(sc->common), regs_btsc,
         sizeof(regs_btsc) / sizeof(cx88_reg));  
   }
   return 0;
}

static int cx88_audio_eiaj(cx88_audio_softc *sc, uint32_t *ctl_reg)
{
   /* TODO: Linux driver doesn't have EIAJ register values, so these
    * are taken from EIAJ example in CX2388x datasheet.  For BTSC at
    * least, the datasheet example values will work, but the cribbed
    * Linux driver values give better quality.  So hopefully these will
    * work at least as a stopgap. */
   uint32_t src_sel = 0;
   if (sc->second_lang)
   {
      src_sel = AUD_SRC_EIAJ_FORCE_MONO2;
   }
   else
   {
      switch (sc->stereo_format)
      {
         case CX88_AUDIO_STEREO_OFF:
            src_sel = AUD_SRC_EIAJ_FORCE_MONO1;
            break;
         case CX88_AUDIO_STEREO_ON:
            src_sel = AUD_SRC_EIAJ_FORCE_STEREO;
            break;
         case CX88_AUDIO_STEREO_AUTO:
            src_sel = AUD_SRC_EIAJ_AUTO_STEREO;
            break;
      }
   }
   *ctl_reg = cx88_update_field(*ctl_reg, AUD_CTL_SRC_SEL, src_sel);
   static const cx88_reg regs_eiaj[] = 
   {
      {AUDIO_REG_RATE_ADJ1, 0x00000100},
      {AUDIO_REG_RATE_ADJ2, 0x00000200},
      {AUDIO_REG_RATE_ADJ3, 0x00000300},
      {AUDIO_REG_RATE_ADJ4, 0x00000400},
      {AUDIO_REG_RATE_ADJ5, 0x00000500},
      {AUDIO_REG_THR_FR, 0x00000000},
      {AUDIO_REG_PILOT_BQD_1_K0, 0x00008000},
      {AUDIO_REG_PILOT_BQD_1_K1, 0x00000000},
      {AUDIO_REG_PILOT_BQD_1_K2, 0x00000000},
      {AUDIO_REG_PILOT_BQD_1_K3, 0x00000000},
      {AUDIO_REG_PILOT_BQD_1_K4, 0x00000000},
      {AUDIO_REG_PILOT_BQD_2_K0, 0x00C00000},
      {AUDIO_REG_PILOT_BQD_2_K1, 0x00000000},
      {AUDIO_REG_PILOT_BQD_2_K2, 0x00000000},
      {AUDIO_REG_PILOT_BQD_2_K3, 0x00000000},
      {AUDIO_REG_PILOT_BQD_2_K4, 0x00000000},
      {AUDIO_REG_C2_UP_THR, 0x00001000},
      {AUDIO_REG_C2_LO_THR, 0x00000400},
      {AUDIO_REG_C1_UP_THR, 0x00001800},
      {AUDIO_REG_C1_LO_THR, 0x00001000},
      {AUDIO_REG_MODE_CHG_TIMER, 0x00000080},
      {AUDIO_REG_START_TIMER, 0x00000200},
      {AUDIO_REG_DEEMPH0_G0, 0x000023F8},
      {AUDIO_REG_PHASE_FIX_CTL, 0x00000009},
      {AUDIO_REG_CORDIC_SHIFT_0, 0x00000006},
      {AUDIO_REG_CORDIC_SHIFT_1, 0x00000006}
   };
   cx88_write_regs(&(sc->common), regs_eiaj, sizeof(regs_eiaj) / sizeof(cx88_reg));
   return 0;  
}

static int cx88_audio_a2(cx88_audio_softc *sc, uint32_t *ctl_reg)
{   
   uint32_t src_sel = 0;
   if (sc->second_lang)
   {
      src_sel = AUD_SRC_A2_FORCE_MONO2;
   }
   else
   {
      switch (sc->stereo_format)
      {
         case CX88_AUDIO_STEREO_OFF:
            src_sel = AUD_SRC_A2_FORCE_MONO1;
            break;
         case CX88_AUDIO_STEREO_ON:
            src_sel = AUD_SRC_A2_FORCE_STEREO;
            break;
         case CX88_AUDIO_STEREO_AUTO:
            src_sel = AUD_SRC_A2_AUTO_STEREO;
            break;
      }
   }
   *ctl_reg = cx88_update_field(*ctl_reg, AUD_CTL_SRC_SEL, src_sel);
   *ctl_reg |= AUD_CTL_FMRADIO_RDS;
   static const cx88_reg regs_a2_common[] = 
   {
      {AUDIO_REG_ERRLOGPERIOD_R, 0x00000064},
      {AUDIO_REG_ERRINTRPTTHSHLD1_R, 0x00000FFF},
      {AUDIO_REG_ERRINTRPTTHSHLD2_R, 0x0000001F},
      {AUDIO_REG_ERRINTRPTTHSHLD3_R, 0x0000000F},
      {AUDIO_REG_PDF_DDS_CNST, 0x12820600},
      {AUDIO_REG_QAM_MODE, 0x00000005},
      {AUDIO_REG_PHACC_FREQ, 0x4C340000},
      {AUDIO_REG_RATE_ADJ1, 0x00000100},
      {AUDIO_REG_RATE_ADJ2, 0x00000200},
      {AUDIO_REG_RATE_ADJ3, 0x00000300},
      {AUDIO_REG_RATE_ADJ4, 0x00000400},
      {AUDIO_REG_RATE_ADJ5, 0x00000500},
      {AUDIO_REG_THR_FR, 0x00000000},
      {AUDIO_REG_AGC_HYST, 0x0000001A},
      {AUDIO_REG_PILOT_BQD_1_K0, 0x0000755B},
      {AUDIO_REG_PILOT_BQD_1_K1, 0x00551340},
      {AUDIO_REG_PILOT_BQD_1_K2, 0x006D30BE},
      {AUDIO_REG_PILOT_BQD_1_K3, 0xFFD394AF},
      {AUDIO_REG_PILOT_BQD_1_K4, 0x00400000},
      {AUDIO_REG_PILOT_BQD_2_K0, 0x00040000},
      {AUDIO_REG_PILOT_BQD_2_K1, 0x002A4841},
      {AUDIO_REG_PILOT_BQD_2_K2, 0x00400000},
      {AUDIO_REG_PILOT_BQD_2_K3, 0x00000000},
      {AUDIO_REG_PILOT_BQD_2_K4, 0x00000000},
      {AUDIO_REG_MODE_CHG_TIMER, 0x00000040},
      {AUDIO_REG_CORDIC_SHIFT_0, 0x00000007},
      {AUDIO_REG_CORDIC_SHIFT_1, 0x00000007},
      {AUDIO_REG_DEEMPH0_G0, 0x00000380},
      {AUDIO_REG_DEEMPH1_G0, 0x00000380},
      {AUDIO_REG_DCOC_0_SRC, 0x0000001A},
      {AUDIO_REG_DCOC0_SHIFT, 0x00000000},
      {AUDIO_REG_DCOC_0_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_0_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_DCOC_PASS_IN, 0x00000003},
      {AUDIO_REG_IIR3_0_SEL, 0x00000021},
      {AUDIO_REG_DN2_AFC, 0x00000002},
      {AUDIO_REG_DCOC_1_SRC, 0x0000001B},
      {AUDIO_REG_DCOC1_SHIFT, 0x00000000},
      {AUDIO_REG_DCOC_1_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_1_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_IIR3_1_SEL, 0x00000023},
      {AUDIO_REG_RDSI_SEL, 0x00000017},
      {AUDIO_REG_RDSI_SHIFT, 0x00000000},
      {AUDIO_REG_RDSQ_SEL, 0x00000017},
      {AUDIO_REG_RDSQ_SHIFT, 0x00000000},
      {AUDIO_REG_PLL_INT, 0x0000001E},
      {AUDIO_REG_PLL_DDS, 0x00000000},
      {AUDIO_REG_PLL_FRAC, 0x0000E542},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000001},
      {AUDIO_REG_START_TIMER, 0x00000000}
   };
   static const cx88_reg regs_a2_bg[] =
   {
      {AUDIO_REG_DMD_RA_DDS, 0x002A4F2F},
      {AUDIO_REG_C1_UP_THR, 0x00007000},
      {AUDIO_REG_C1_LO_THR, 0x00005400},
      {AUDIO_REG_C2_UP_THR, 0x00005400},
      {AUDIO_REG_C2_LO_THR, 0x00003000}
   };
   static const cx88_reg regs_a2_dk[] =
   {
      {AUDIO_REG_DMD_RA_DDS, 0x002A4F2F},
      {AUDIO_REG_C1_UP_THR, 0x00007000},
      {AUDIO_REG_C1_LO_THR, 0x00005400},
      {AUDIO_REG_C2_UP_THR, 0x00005400},
      {AUDIO_REG_C2_LO_THR, 0x00003000},
      {AUDIO_REG_DN0_FREQ, 0x00003A1C},
      {AUDIO_REG_DN2_FREQ, 0x0000D2E0}
   };
   static const cx88_reg regs_deemph_50[] = 
   {
      {AUDIO_REG_DEEMPH0_G0, 0x00000380},
      {AUDIO_REG_DEEMPH1_G0, 0x00000380},
      {AUDIO_REG_DEEMPHGAIN_R, 0x000011E1},
      {AUDIO_REG_DEEMPHNUMER1_R, 0x0002A7BC},
      {AUDIO_REG_DEEMPHNUMER2_R, 0x0003023C}
   };
   static const cx88_reg regs_a1_i[] = 
   {
      {AUDIO_REG_ERRLOGPERIOD_R, 0x00000064},
      {AUDIO_REG_ERRINTRPTTHSHLD1_R, 0x00000FFF},
      {AUDIO_REG_ERRINTRPTTHSHLD2_R, 0x0000001F},
      {AUDIO_REG_ERRINTRPTTHSHLD3_R, 0x0000000F},
      {AUDIO_REG_PDF_DDS_CNST, 0x12820600},
      {AUDIO_REG_QAM_MODE, 0x00000005},
      {AUDIO_REG_PHACC_FREQ, 0x933A0000},
      {AUDIO_REG_DMD_RA_DDS, 0x002A4F2F},
      {AUDIO_REG_PLL_INT, 0x0000001E},
      {AUDIO_REG_PLL_DDS, 0x00000004},
      {AUDIO_REG_PLL_FRAC, 0x0000E542},
      {AUDIO_REG_RATE_ADJ1, 0x00000100},
      {AUDIO_REG_RATE_ADJ2, 0x00000200},
      {AUDIO_REG_RATE_ADJ3, 0x00000300},
      {AUDIO_REG_RATE_ADJ4, 0x00000400},
      {AUDIO_REG_RATE_ADJ5, 0x00000500},
      {AUDIO_REG_THR_FR, 0x00000000},
      {AUDIO_REG_PILOT_BQD_1_K0, 0x0000755B},
      {AUDIO_REG_PILOT_BQD_1_K1, 0x00551340},
      {AUDIO_REG_PILOT_BQD_1_K2, 0x006D30BE},
      {AUDIO_REG_PILOT_BQD_1_K3, 0xFFD394AF},
      {AUDIO_REG_PILOT_BQD_1_K4, 0x00400000},
      {AUDIO_REG_PILOT_BQD_2_K0, 0x00040000},
      {AUDIO_REG_PILOT_BQD_2_K1, 0x002A4841},
      {AUDIO_REG_PILOT_BQD_2_K2, 0x00400000},
      {AUDIO_REG_PILOT_BQD_2_K3, 0x00000000},
      {AUDIO_REG_PILOT_BQD_2_K4, 0x00000000},
      {AUDIO_REG_MODE_CHG_TIMER, 0x00000060},
      {AUDIO_REG_AGC_HYST, 0x0000000A},
      {AUDIO_REG_CORDIC_SHIFT_0, 0x00000007},
      {AUDIO_REG_CORDIC_SHIFT_1, 0x00000007},
      {AUDIO_REG_C1_UP_THR, 0x00007000},
      {AUDIO_REG_C1_LO_THR, 0x00005400},
      {AUDIO_REG_C2_UP_THR, 0x00005400},
      {AUDIO_REG_C2_LO_THR, 0x00003000},
      {AUDIO_REG_DCOC_0_SRC, 0x0000001A},
      {AUDIO_REG_DCOC0_SHIFT, 0x00000000},
      {AUDIO_REG_DCOC_0_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_0_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_DCOC_PASS_IN, 0x00000003},
      {AUDIO_REG_IIR3_0_SEL, 0x00000021},
      {AUDIO_REG_DN2_AFC, 0x00000002},
      {AUDIO_REG_DCOC_1_SRC, 0x0000001B},
      {AUDIO_REG_DCOC1_SHIFT, 0x00000000},
      {AUDIO_REG_DCOC_1_SHIFT_IN0, 0x0000000A},
      {AUDIO_REG_DCOC_1_SHIFT_IN1, 0x00000008},
      {AUDIO_REG_IIR3_1_SEL, 0x00000023},
      {AUDIO_REG_DN0_FREQ, 0x000035A3},
      {AUDIO_REG_DN2_FREQ, 0x000029C7},
      {AUDIO_REG_CRDC0_SRC_SEL, 0x00000511},
      {AUDIO_REG_IIR1_0_SEL, 0x00000001},
      {AUDIO_REG_IIR1_1_SEL, 0x00000000},
      {AUDIO_REG_IIR3_2_SEL, 0x00000003},
      {AUDIO_REG_IIR3_2_SHIFT, 0x00000000},
      {AUDIO_REG_IIR3_0_SEL, 0x00000002},
      {AUDIO_REG_IIR2_0_SEL, 0x00000021},
      {AUDIO_REG_IIR2_0_SHIFT, 0x00000002},
      {AUDIO_REG_DEEMPH0_SRC_SEL, 0x0000000B},
      {AUDIO_REG_DEEMPH1_SRC_SEL, 0x0000000B},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000001},
      {AUDIO_REG_START_TIMER, 0x00000000}
   };
   static const cx88_reg regs_am_l[] = 
   {
      {AUDIO_REG_ERRLOGPERIOD_R, 0x00000064},
      {AUDIO_REG_ERRINTRPTTHSHLD1_R, 0x00000FFF},
      {AUDIO_REG_ERRINTRPTTHSHLD2_R, 0x0000001F},
      {AUDIO_REG_ERRINTRPTTHSHLD3_R, 0x0000000F},
      {AUDIO_REG_PDF_DDS_CNST, 0xF53D4800},
      {AUDIO_REG_QAM_MODE, 0x00000000},
      {AUDIO_REG_PHACC_FREQ, 0x4A3A0000},
      {AUDIO_REG_DEEMPHGAIN_R, 0x00006680},
      {AUDIO_REG_DEEMPHNUMER1_R, 0x000353DE},
      {AUDIO_REG_DEEMPHNUMER2_R, 0x000001B1},
      {AUDIO_REG_DEEMPHDENOM1_R, 0x0000F3D0},
      {AUDIO_REG_DEEMPHDENOM2_R, 0x00000000},
      {AUDIO_REG_FM_MODE_ENABLE, 0x00000007},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000003},
      {AUDIO_REG_AGC_GAIN, 0x00000000},
      {AUDIO_REG_AGC_HYST, 0x00000018},
      {AUDIO_REG_AGC_DEF, 0x00000020},
      {AUDIO_REG_DN0_FREQ, 0x00000000},
      {AUDIO_REG_POLY0_DDS_CONSTANT, 0x000E4DB2},
      {AUDIO_REG_DCOC_0_SRC, 0x00000021},
      {AUDIO_REG_IIR1_0_SEL, 0x00000000},
      {AUDIO_REG_IIR1_0_SHIFT, 0x00000007},
      {AUDIO_REG_IIR1_1_SEL, 0x00000002},
      {AUDIO_REG_IIR1_1_SHIFT, 0x00000000},
      {AUDIO_REG_DCOC_1_SRC, 0x00000003},
      {AUDIO_REG_DCOC1_SHIFT, 0x00000000},
      {AUDIO_REG_DCOC_PASS_IN, 0x00000000},
      {AUDIO_REG_IIR1_2_SEL, 0x00000023},
      {AUDIO_REG_IIR1_2_SHIFT, 0x00000000},
      {AUDIO_REG_IIR1_3_SEL, 0x00000004},
      {AUDIO_REG_IIR1_3_SHIFT, 0x00000007},
      {AUDIO_REG_IIR1_4_SEL, 0x00000005},
      {AUDIO_REG_IIR1_4_SHIFT, 0x00000007},
      {AUDIO_REG_IIR3_0_SEL, 0x00000007},
      {AUDIO_REG_IIR3_0_SHIFT, 0x00000000},
      {AUDIO_REG_DEEMPH0_SRC_SEL, 0x00000011},
      {AUDIO_REG_DEEMPH0_SHIFT, 0x00000000},
      {AUDIO_REG_DEEMPH0_G0, 0x00007000},
      {AUDIO_REG_DEEMPH0_A0, 0x00000000},
      {AUDIO_REG_DEEMPH0_B0, 0x00000000},
      {AUDIO_REG_DEEMPH0_A1, 0x00000000},
      {AUDIO_REG_DEEMPH0_B1, 0x00000000},
      {AUDIO_REG_DEEMPH1_SRC_SEL, 0x00000011},
      {AUDIO_REG_DEEMPH1_SHIFT, 0x00000000},
      {AUDIO_REG_DEEMPH1_G0, 0x00007000},
      {AUDIO_REG_DEEMPH1_A0, 0x00000000},
      {AUDIO_REG_DEEMPH1_B0, 0x00000000},
      {AUDIO_REG_DEEMPH1_A1, 0x00000000},
      {AUDIO_REG_DEEMPH1_B1, 0x00000000},
      {AUDIO_REG_OUT0_SEL, 0x0000003F},
      {AUDIO_REG_OUT1_SEL, 0x0000003F},
      {AUDIO_REG_DMD_RA_DDS, 0x00F5C285},
      {AUDIO_REG_PLL_INT, 0x0000001E},
      {AUDIO_REG_PLL_DDS, 0x00000000},
      {AUDIO_REG_PLL_FRAC, 0x0000E542},
      {AUDIO_REG_RATE_ADJ1, 0x00000100},
      {AUDIO_REG_RATE_ADJ2, 0x00000200},
      {AUDIO_REG_RATE_ADJ3, 0x00000300},
      {AUDIO_REG_RATE_ADJ4, 0x00000400},
      {AUDIO_REG_RATE_ADJ5, 0x00000500},
      {AUDIO_REG_RATE_THRES_DMD, 0x000000C0}
   };
   switch (sc->broadcast_format)
   {
      case CX88_AUDIO_FMT_AM_L:
         cx88_write_regs(&(sc->common), regs_am_l, sizeof(regs_am_l) / sizeof(cx88_reg));
         break;
      case CX88_AUDIO_FMT_A2_BG:
         cx88_write_regs(&(sc->common), regs_a2_common, sizeof(regs_a2_common) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_a2_bg, sizeof(regs_a2_bg) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_deemph_50, sizeof(regs_deemph_50) / sizeof(cx88_reg));
         break;
      case CX88_AUDIO_FMT_A2_DK:
         cx88_write_regs(&(sc->common), regs_a2_common, sizeof(regs_a2_common) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_a2_dk, sizeof(regs_a2_dk) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_deemph_50, sizeof(regs_deemph_50) / sizeof(cx88_reg));
         break;
      case CX88_AUDIO_FMT_A1_I:
         cx88_write_regs(&(sc->common), regs_a1_i, sizeof(regs_a1_i) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_deemph_50, sizeof(regs_deemph_50) / sizeof(cx88_reg));
         break;
   }
   return 0;
}

static int cx88_audio_nicam(cx88_audio_softc *sc, uint32_t *ctl_reg)
{
   uint32_t src_sel = 0;
   if (sc->second_lang)
   {
      src_sel = AUD_SRC_NICAM_FORCE_MONO2;
   }
   else
   {
      switch (sc->stereo_format)
      {
         case CX88_AUDIO_STEREO_OFF:
            src_sel = AUD_SRC_NICAM_FORCE_MONO1;
            break;
         case CX88_AUDIO_STEREO_ON:
            src_sel = AUD_SRC_NICAM_FORCE_STEREO;
            break;
         case CX88_AUDIO_STEREO_AUTO:
            src_sel = AUD_SRC_NICAM_AUTO_STEREO;
            break;
      }
   }
   *ctl_reg = cx88_update_field(*ctl_reg, AUD_CTL_SRC_SEL, src_sel);
   *ctl_reg = cx88_update_field(*ctl_reg, AUD_CTL_DEMATRIX, AUD_DEMATRIX_LR);
   *ctl_reg |= AUD_CTL_DEMATRIX_BYPASS;
   static const cx88_reg regs_nicam_l[] =
   {
      {AUDIO_REG_RATE_ADJ1, 0x00000060},
      {AUDIO_REG_RATE_ADJ2, 0x000000F9},
      {AUDIO_REG_RATE_ADJ3, 0x000001CC},
      {AUDIO_REG_RATE_ADJ4, 0x000002B3},
      {AUDIO_REG_RATE_ADJ5, 0x00000726},
      {AUDIO_REG_DEEMPHDENOM1_R, 0x0000F3D0},
      {AUDIO_REG_DEEMPHDENOM2_R, 0x00000000},
      {AUDIO_REG_ERRLOGPERIOD_R, 0x00000064},
      {AUDIO_REG_ERRINTRPTTHSHLD1_R, 0x00000FFF},
      {AUDIO_REG_ERRINTRPTTHSHLD2_R, 0x0000001F},
      {AUDIO_REG_ERRINTRPTTHSHLD3_R, 0x0000000F},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000003},
      {AUDIO_REG_DMD_RA_DDS, 0x00C00000},
      {AUDIO_REG_PLL_INT, 0x0000001E},
      {AUDIO_REG_PLL_DDS, 0x00000000},
      {AUDIO_REG_PLL_FRAC, 0x0000E542},
      {AUDIO_REG_START_TIMER, 0x00000000},
      {AUDIO_REG_DEEMPHNUMER1_R, 0x000353DE},
      {AUDIO_REG_DEEMPHNUMER2_R, 0x000001B1},
      {AUDIO_REG_PDF_DDS_CNST, 0x12820600},
      {AUDIO_REG_QAM_MODE, 0x00000005},
      {AUDIO_REG_PHACC_FREQ, 0x4C340000},
      {AUDIO_REG_DEEMPHGAIN_R, 0x00006680},
      {AUDIO_REG_RATE_THRES_DMD, 0x000000C0}   
   };
   static const cx88_reg regs_nicam_bgdki[] =
   {
      {AUDIO_REG_RATE_ADJ1, 0x00000010},
      {AUDIO_REG_RATE_ADJ2, 0x00000040},
      {AUDIO_REG_RATE_ADJ3, 0x00000100},
      {AUDIO_REG_RATE_ADJ4, 0x00000400},
      {AUDIO_REG_RATE_ADJ5, 0x00001000},
      {AUDIO_REG_ERRLOGPERIOD_R, 0x00000FFF},
      {AUDIO_REG_ERRINTRPTTHSHLD1_R, 0x000003FF},
      {AUDIO_REG_ERRINTRPTTHSHLD2_R, 0x000000FF},
      {AUDIO_REG_ERRINTRPTTHSHLD3_R, 0x0000003F},
      {AUDIO_REG_POLYPH80SCALEFAC, 0x00000003},
      {AUDIO_REG_DEEMPHGAIN_R, 0x000023C2},
      {AUDIO_REG_DEEMPHNUMER1_R, 0x0002A7BC},
      {AUDIO_REG_DEEMPHNUMER2_R, 0x0003023E},
      {AUDIO_REG_DEEMPHDENOM1_R, 0x0000F3D0},
      {AUDIO_REG_DEEMPHDENOM2_R, 0x00000000},
      {AUDIO_REG_QAM_MODE, 0x00000005}
   };
   static const cx88_reg regs_nicam_i[] =
   { 
      {AUDIO_REG_PDF_DDS_CNST, 0x12820600},
      {AUDIO_REG_PHACC_FREQ, 0x933A0000}
   };
   static const cx88_reg regs_nicam_bgdk[] =
   {
      {AUDIO_REG_PDF_DDS_CNST, 0x16820600},
      {AUDIO_REG_PHACC_FREQ, 0x4C340000}
   };
   switch (sc->broadcast_format)
   {
      case CX88_AUDIO_FMT_NICAM_BGDK:
         cx88_write_regs(&(sc->common), regs_nicam_bgdki, sizeof(regs_nicam_bgdki) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_nicam_bgdk, sizeof(regs_nicam_bgdk) / sizeof(cx88_reg));
         break;
      case CX88_AUDIO_FMT_NICAM_I:
         cx88_write_regs(&(sc->common), regs_nicam_bgdki, sizeof(regs_nicam_bgdki) / sizeof(cx88_reg));
         cx88_write_regs(&(sc->common), regs_nicam_i, sizeof(regs_nicam_i) / sizeof(cx88_reg));
         break;
      case CX88_AUDIO_FMT_NICAM_L:
         cx88_write_regs(&(sc->common), regs_nicam_l, sizeof(regs_nicam_l) / sizeof(cx88_reg));
         break;
   }
   return 0;
}

static int cx88_audio_load_format(cx88_audio_softc *sc)
{
   int error = 0;
   uint32_t ctrl_reg = 0;
   uint32_t i2s_ctrl_reg = 0;
   if (sc->input == CX88_AUDIO_INPUT_I2S)
   {
      ctrl_reg |= AUD_CTL_I2SIN_ENABLE;
   }
   if (sc->i2s_config & CX88_AUDIO_I2S_TO_DAC)
   {
      ctrl_reg |= (AUD_CTL_I2SIN_ENABLE | AUD_CTL_I2SIN_ENABLE | AUD_CTL_I2STODAC);
   }
   switch (sc->i2s_config & 0x70)
   {
      case CX88_AUDIO_I2S_MODE_OSRDAC:  
         ctrl_reg &= (~AUD_CTL_I2SIN_ENABLE);
         ctrl_reg |= AUD_CTL_I2SOUT_ENABLE;
         i2s_ctrl_reg |= I2S_CTL_OS_INSCLK;
         break;
      case CX88_AUDIO_I2S_MODE_OSRADC:
         ctrl_reg &= (~AUD_CTL_I2SOUT_ENABLE);
         ctrl_reg |= AUD_CTL_I2SIN_ENABLE;
         i2s_ctrl_reg |= I2S_CTL_OS_OUTSCLK;
         break;
      case CX88_AUDIO_I2S_MODE_PASS:
         ctrl_reg &= (~AUD_CTL_I2SIN_ENABLE);
         ctrl_reg |= AUD_CTL_I2SOUT_ENABLE;
         i2s_ctrl_reg |= I2S_CTL_PASSTHRU;
         break;
      case CX88_AUDIO_I2S_MODE_FULL:
         ctrl_reg |= (AUD_CTL_I2SIN_ENABLE | AUD_CTL_I2SOUT_ENABLE);
         i2s_ctrl_reg |= (I2S_CTL_SLAVE_INSCLK | I2S_CTL_OS_INSCLK);
         break;
   }
   if (sc->input == CX88_AUDIO_INPUT_BCAST)
   {
      cx88_audio_broadcast_format *fmt = &(broadcast_formats[sc->broadcast_format]);
      cx88_lock_board(&(sc->common));
      cx88_write_reg(&(sc->common), CX88_REG_ADC_CONFIG,
         cx88_read_reg(&(sc->common), CX88_REG_ADC_CONFIG) & (~CX88_CHROMA_VIDEO));
      cx88_unlock_board(&(sc->common));
      cx88_write_reg(&(sc->common), AUDIO_REG_AFE_12DB, 1);
      if ((fmt->init_reg == AUDIO_INIT_BTSC) && sc->second_lang)
      {
         cx88_write_reg(&(sc->common), AUDIO_REG_INIT, AUDIO_INIT_BTSC_SAP);
      }
      else
      {
         cx88_write_reg(&(sc->common), AUDIO_REG_INIT, fmt->init_reg);
      }
      cx88_write_reg(&(sc->common), AUDIO_REG_INIT_LOAD, 1);
      cx88_write_reg(&(sc->common), AUDIO_REG_SOFT_RESET, 1);
      error = fmt->set_format(sc, &ctrl_reg);
   }
   else
   { 
      cx88_write_reg(&(sc->common), AUDIO_REG_SOFT_RESET, 1);
   }
   cx88_write_reg(&(sc->common), AUDIO_REG_RATE_THRES_DMD, 0xC0);
   cx88_write_reg(&(sc->common), AUDIO_REG_I2S_CTL, i2s_ctrl_reg);
   cx88_write_reg(&(sc->common), AUDIO_REG_I2SINPUT, 
      sc->i2s_config & (CX88_AUDIO_I2S_IN_MASTER | CX88_AUDIO_I2S_IN_SONY));
   cx88_write_reg(&(sc->common), AUDIO_REG_I2SOUTPUT,
      (sc->i2s_config & (CX88_AUDIO_I2S_OUT_SONY)) >> 2);
   cx88_write_reg(&(sc->common), AUDIO_REG_AUD_CTL, ctrl_reg);
   cx88_write_reg(&(sc->common), AUDIO_REG_SOFT_RESET, 0);
   return error;
}

static int cx88_audio_start_capture(cx88_audio_softc *sc)
{
   cx88_audio_reset(sc, sc->state);
   if (sc->state != CX88_AUDIO_CONFIGURED)
   {
      return ENXIO;
   }
   int error = cx88_audio_load_format(sc);
   if (error)
   {
      return error;
   }
   cx88_write_reg(&(sc->common), AUDIO_REG_IN_LINE_LENGTH, sc->channel.line_size);
   uint32_t dma_ctrl = AUDIO_IN_RISC_ENABLE | AUDIO_IN_FIFO_ENABLE;
   sc->buffer_available = 0;
   sc->isr_status = CX88_AUDIO_CAP_NOERR;
   sc->state = CX88_AUDIO_CAPTURING;
   /* Protect access to the shared PCI interrupt register. */
   cx88_lock_board(&(sc->common));
   cx88_write_reg(&(sc->common), CX88_REG_PCI_INTMASK,
      (cx88_read_reg(&(sc->common), CX88_REG_PCI_INTMASK) | CX88_PCI_INT_AUDIO));
   cx88_unlock_board(&(sc->common));
   /* Acknowledge any outstanding interrupts */
   cx88_write_reg(&(sc->common), AUDIO_REG_INT_STAT,
      cx88_read_reg(&(sc->common), AUDIO_REG_INT_STAT));
   cx88_write_reg(&(sc->common), AUDIO_REG_INT_MASK, 
      (AUDIO_INT_IN_SYNC | AUDIO_INT_OPCODE_ERR | AUDIO_INT_PCI_ABORT | 
       AUDIO_INT_PAR_ERR | AUDIO_INT_RIP_ERR | AUDIO_INT_IN_RISCI1));
   cx88_write_reg(&(sc->common), CX88_REG_DEVCNTL2, CX88_RUN_RISC);
   cx88_write_reg(&(sc->common), AUDIO_REG_DMA_CONTROL, dma_ctrl);
   cx88_write_reg(&(sc->common), AUDIO_REG_BAL_CTL, sc->balance_reg);
   if (!sc->mute)
   {
      cx88_write_reg(&(sc->common), AUDIO_REG_VOL_CTL, sc->volume_reg);
   }
   return 0;
}

/* Configures channel, buffers, and I2C devices */
static int cx88_audio_configure(
   cx88_audio_softc *sc,
   cx88_audio_config *config,
   uint32_t blocksize,
   struct thread *context)
{
   int error = 0;
   if (sc->state == CX88_AUDIO_PCM)
   {
      return EBUSY;
   }
   if (sc->state < CX88_AUDIO_UNCONFIGURED)
   {
      return ENXIO;
   }
   cx88_audio_reset(sc, CX88_AUDIO_UNCONFIGURED); /* Ensure there is no outstanding capture */
   uint32_t samples_per_buffer, samples_last_buffer;
   uint8_t num_buffers = 0;
   /* Validate ioctl parameters */
   if ((config->in_num_samples < 2) ||
      ((config->in_buftype == CX88_AUDIO_BUF_USER) && (config->bufconfig.in_userbuf == NULL)) ||
      ((config->in_buftype != CX88_AUDIO_BUF_USER) && (config->in_buftype != CX88_AUDIO_BUF_KERNEL)))
   {
      return EINVAL;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Configuring audio input channel\n"))
   if (!error)
   {
      /* Allocate the channel */
      error = cx88_configure_channel(
         &(sc->channel),
         0,
         CX88_AUDIO_SAMPLE_SIZE,
         MAX_AUDIO_LINE_SIZE,
         config->in_num_samples);
   }
   if (!error)
   {
      samples_per_buffer = blocksize / CX88_AUDIO_SAMPLE_SIZE;
      if (samples_per_buffer == 0)
      {
         samples_per_buffer = config->in_num_samples / 2;
         if (samples_per_buffer > (sc->pcm_bufsize / CX88_AUDIO_SAMPLE_SIZE / 2))
         {
            samples_per_buffer = sc->pcm_bufsize / CX88_AUDIO_SAMPLE_SIZE / 2;
         }
         uint32_t bufsize = samples_per_buffer * CX88_AUDIO_SAMPLE_SIZE;
         if (bufsize < PAGE_SIZE)
         {
            device_printf(sc->common.dev, "Unable to configure buffers; Number of samples is too small\n");
            return EINVAL;
         }
         samples_per_buffer = round_page(bufsize - (PAGE_SIZE / 2)) / CX88_AUDIO_SAMPLE_SIZE;
         num_buffers = config->in_num_samples / samples_per_buffer;
         if (((config->in_num_samples % samples_per_buffer) * CX88_AUDIO_SAMPLE_SIZE) >= PAGE_SIZE)
         {
            ++num_buffers;
         }
      }
      else
      {
         num_buffers = config->in_num_samples / samples_per_buffer;
      }
      sc->pcm_blocksize = samples_per_buffer * CX88_AUDIO_SAMPLE_SIZE;
      /* Number of samples in last buffer will be less than samples_per_buffer
       * if num_samples is not a multiple of num_buffers */
      samples_last_buffer = config->in_num_samples - ((num_buffers - 1) * samples_per_buffer);
      config->out_num_buffers = num_buffers;
      config->out_samples_per_buffer = samples_per_buffer;
      if (config->in_buftype == CX88_AUDIO_BUF_KERNEL)
      {
         /* Calculate the total amount of memory the application will need to mmap(), along
          * with the buffer size the application will need to use in calculating the offsets
          * of individual buffers within that mmap'ed region */
         config->bufconfig.out_kernelbuf.out_buffer_size = round_page(sc->pcm_blocksize);
         config->bufconfig.out_kernelbuf.out_mmap_size = 
            ((num_buffers - 1) * config->bufconfig.out_kernelbuf.out_buffer_size) + 
               (samples_last_buffer * CX88_AUDIO_SAMPLE_SIZE);
      }
      /* Request the buffer structures */
      error = cx88_request_buffers (&(sc->channel), num_buffers, 0);
   }
   if (!error)
   {
      uint32_t i;
      for (i = 0; i < num_buffers; ++i)
      {
         void *bufptr = NULL;
         struct thread *bufcontext = NULL;
         uint32_t num_samples = ((i == (num_buffers - 1)) ? samples_last_buffer : samples_per_buffer);
         uint32_t byte_offset = i * sc->pcm_blocksize;
         size_t buffer_size = num_samples * CX88_AUDIO_SAMPLE_SIZE;
         if (config->in_buftype == CX88_AUDIO_BUF_USER)
         {
            /* Each cx88 buffer is a segment of the user-allocated buffer */
            bufptr = (uint8_t *)(config->bufconfig.in_userbuf) + byte_offset;
            bufcontext = context;
         }
         else if ((byte_offset + buffer_size) <= sc->pcm_bufsize)
         {
            bufptr = (uint8_t *)(sc->pcm_databuf) + byte_offset;
         }
         
         /* Configure each buffer */
         error = cx88_configure_buffer(
            &(sc->channel.buffers[i]),
            bufptr,
            bufcontext,
            buffer_size,
            ((i == (num_buffers - 1)) ? 0 : (i + 1)),
            cx88_audio_dmaprog,
            CX88_REPROGRAM_ON_BUFFER_CHANGE);
         if (error)
         {
            break;
         }
      }
   }
   if (!error)
   {
      /* Now commit all changes we've made to the channel configuration */
      error = cx88_commit_channel(&(sc->channel));
   }
   if (!error)
   {
      sc->state = CX88_AUDIO_CONFIGURED;
   }
   else
   {
      cx88_release_channel(&(sc->channel));
   }
   return error;
}

/* Open file operation */
static int cx88_audio_open(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   cx88_audio_softc *sc = devclass_get_softc(cx88_audio_devclass, minor(dev));
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Opening device\n"))
   sx_xlock(&(sc->sxlock));
   if (sc->in_use)
   {
      sx_xunlock(&(sc->sxlock));
      return EBUSY;
   }
   if (sc->state >= CX88_AUDIO_CONFIGURED)
   {
      cx88_i2c_request_devices(&(sc->i2c));
   }
   sc->in_use = TRUE;
   sx_xunlock(&sc->sxlock);
   return 0;
}

static void cx88_audio_close_session(cx88_audio_softc *sc)
{
   if (sc->reconfig_requested)
   {
      cx88_audio_reconfigure(sc);
      sc->reconfig_requested = FALSE;
   }
   /* We can persist our configuration, so just release ownership
    * of I2C devices */
   cx88_i2c_release_devices(&(sc->i2c));
}

/* Close file operation */
static int cx88_audio_close(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{   
   cx88_audio_softc *sc = devclass_get_softc(cx88_audio_devclass, minor(dev));
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Closing device\n"))
   sx_xlock(&(sc->sxlock));
   if (sc->state != CX88_AUDIO_PCM)
   {
      cx88_audio_reset(sc, sc->state);
      /* If we're using user-allocated buffers, release them as
      * they can't be persisted across open() and close() in
      * different processes */
      cx88_free_userbufs(&(sc->channel), NULL);
      cx88_audio_close_session(sc);
   }
   sc->in_use = FALSE;
   sc->user_configured = FALSE;
   sx_xunlock(&(sc->sxlock));
   return 0;
}

/* Mmap file operation */
static int cx88_audio_mmap(
   struct cdev *dev,
   vm_offset_t offset,
   vm_paddr_t *paddr,
   int nprot)
{
   cx88_audio_softc *sc = devclass_get_softc(cx88_audio_devclass, minor(dev));
   sx_slock(&(sc->sxlock));
   DIAGNOSTIC(device_printf(sc->common.dev, "Attempting to mmap device memory\n"))
   if (nprot & (PROT_EXEC | PROT_WRITE))
   {
      /* Audio is output-only */
      DIAGNOSTIC(device_printf(sc->common.dev, "cx88_audio_mmap: Invalid device memory permissions\n"))
      return EINVAL;
   }
   int error = cx88_common_mmap(&(sc->channel), offset, paddr, nprot);
   sx_sunlock(&(sc->sxlock));
   return error;
}

/* Establish broadcast and stereo format */
static int cx88_audio_setformat(cx88_audio_softc *sc, cx88_audio_format *fmt)
{
   if (sc->state < CX88_AUDIO_UNCONFIGURED)
   {
      return ENXIO;
   }
   if ((fmt->in_audio_format >= (sizeof(broadcast_formats) / sizeof(cx88_audio_broadcast_format))) || 
       (fmt->in_stereo_format > CX88_AUDIO_STEREO_AUTO) ||
       (fmt->in_input == 0))
   {
      return EINVAL; 
   }
   sc->broadcast_format = fmt->in_audio_format;
   sc->stereo_format = fmt->in_stereo_format;
   sc->second_lang = fmt->in_second_lang;
   sc->input = fmt->in_input;
   sc->i2s_config = fmt->in_i2s_config;
   if (sc->state >= CX88_AUDIO_CAPTURING)
   {
      return cx88_audio_load_format(sc);  
   }
   return 0;
}

/* Ioctl file operation */
static int cx88_audio_ioctl(struct cdev *dev, u_long cmd, caddr_t data, int flags, struct thread *td)
{
   cx88_audio_softc *sc = devclass_get_softc(cx88_audio_devclass, minor(dev));
   int error = 0;
   if (sc == NULL)
   {
      return ENXIO;
   }
   switch (cmd)
   {
      case CX88_COMMON_CARD_ID: /* Retrieve vendor-specific identifiers */
         *((uint32_t*)data) = cx88_get_card(sc->common.dev);
         break;
      case CX88_AUDIO_FORMAT: /* Establish broadcast/stereo format */
         sx_xlock(&(sc->sxlock));
         error = cx88_audio_setformat(sc, (cx88_audio_format*)data);
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_AUDIO_CONFIG: /* Configure buffers */
         sx_xlock(&(sc->sxlock));
         error = cx88_audio_configure(sc, (cx88_audio_config*)data, 0, td);
         if (!error)
         {
            sc->user_configured = TRUE;  
         }
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_COMMON_I2C_CONFIG: /* Configure I2C devices */
         sx_xlock(&(sc->sxlock));
         if (sc->state < CX88_AUDIO_UNCONFIGURED)
         {
            error = ENXIO;
         }
         else
         {
            error = cx88_i2c_update_devices(&(sc->i2c), (cx88_common_i2c_config*)data, IICDEV_NAME);
            if (!error)
            {
               error = cx88_i2c_request_devices(&(sc->i2c));
            }
         }
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_AUDIO_CAPTURE: /* Start capturing data */
         sx_xlock(&(sc->sxlock));
         error = cx88_audio_start_capture(sc);
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_AUDIO_STOP: /* Stop capturing data */
         /* Wait and stop are slock'ed so that a stop can be issued with a pending wait */
         sx_slock(&(sc->sxlock));
         cx88_audio_reset(sc, sc->state);
         sx_sunlock(&(sc->sxlock));
         break;
      case CX88_AUDIO_WAIT: /* Wait for data to become available */
         sx_slock(&(sc->sxlock));
         error = cx88_audio_transfer(sc, (cx88_audio_wait*)data, ((flags & O_NONBLOCK) ? TRUE : FALSE));
         sx_sunlock(&(sc->sxlock));
         break;
      case CX88_COMMON_GPIO: /* Modify GPIO pin state */
         cx88_gpioctl(&(sc->common), (cx88_common_gpio*)data);
         break;
      case CX88_AUDIO_VOLUME:
         error = cx88_audio_set_volume(sc, (cx88_audio_volume*)data);
         break;
      case CX88_AUDIO_MUTE:
         cx88_audio_mute(sc);
         break;
      case CX88_AUDIO_GET_STATUS:
         cx88_audio_get_status(sc, (cx88_audio_status*)data);
         break;
      default:
         error = ENOTTY;
         break;
   }
   return error;
}

static int cx88_audio_set_volume(cx88_audio_softc *sc, cx88_audio_volume *vol)
{
   if ((vol->in_left > 100) || (vol->in_right > 100))
   {
      return EINVAL;
   }
   uint16_t temp;
   uint8_t greater = vol->in_right;
   uint8_t lesser = vol->in_left;
   uint8_t balance_reg = 0;
   if (vol->in_left > vol->in_right)
   {
      greater = vol->in_left;
      lesser = vol->in_right;
      balance_reg = BAL_CTL_RIGHT;
   }
   if (greater == 0)
   {
      mtx_lock(&(sc->isrlock));
      cx88_write_reg(&(sc->common), 
         AUDIO_REG_VOL_CTL, VOL_CTL_SRC_MUTE | VOL_CTL_I2S_MUTE | VOL_CTL_DAC_MUTE);
      sc->volume = *vol;
      mtx_unlock(&(sc->isrlock));
      return 0;  
   }
   temp = ((uint16_t)lesser * 100) / greater;  
   temp = 0x3F - ((temp * 0x3F) / 100);
   balance_reg |= (uint8_t)temp;
   temp = 0x3F - ((greater * 0x3F) / 100);
   mtx_lock(&(sc->isrlock));
   sc->mute = FALSE;
   sc->balance_reg = balance_reg;
   sc->volume_reg = (uint8_t)temp;
   cx88_write_reg(&(sc->common), AUDIO_REG_BAL_CTL, balance_reg);
   if (sc->state >= CX88_AUDIO_CAPTURING)
   {
      cx88_write_reg(&(sc->common), AUDIO_REG_VOL_CTL, (uint32_t)(sc->volume_reg));
   }
   sc->volume = *vol;
   mtx_unlock(&(sc->isrlock));
   return 0;
}

static void cx88_audio_mute(cx88_audio_softc *sc)
{
   mtx_lock(&(sc->isrlock));
   if (sc->mute && (sc->state >= CX88_AUDIO_CAPTURING))
   {
      cx88_write_reg(&(sc->common), AUDIO_REG_VOL_CTL, sc->volume_reg);
   }
   else
   {
      cx88_write_reg(&(sc->common), 
         AUDIO_REG_VOL_CTL, VOL_CTL_SRC_MUTE | VOL_CTL_I2S_MUTE | VOL_CTL_DAC_MUTE);
   }
   sc->mute = !sc->mute;
   mtx_unlock(&(sc->isrlock));
}

static void cx88_audio_get_status(cx88_audio_softc *sc, cx88_audio_status *status)
{
   status->out_vol_left = sc->volume.in_left;
   status->out_vol_right = sc->volume.in_right;
   status->out_mute = (sc->mute ? 1 : 0);
   uint32_t status_reg = cx88_read_reg(&(sc->common), AUDIO_REG_STATUS);
   status->out_carrier = cx88_read_field(status_reg, AUDIO_STATUS_PILOT);
   status->out_signal = cx88_read_field(status_reg, AUDIO_STATUS_STEREO_MODE);
}

/* Select/poll file operation */
static int cx88_audio_poll(struct cdev *dev, int events, struct thread *td)
{   
   cx88_audio_softc *sc = devclass_get_softc(cx88_audio_devclass, minor(dev));
   if (sc == NULL)
   {
      return ENXIO;
   }
   int revents = 0;
   sx_xlock(&(sc->sxlock));
   mtx_lock(&(sc->isrlock));
   if ((sc->state == CX88_AUDIO_CAPTURING) && (events & (POLLIN | POLLRDNORM)))
   {
      if (sc->data_available)
      {
         /* Notify the application that data is available */
         revents = events & (POLLIN | POLLRDNORM);
      }
      else
      {
         selrecord(td, &(sc->selectors));
      }
   }
   mtx_unlock(&(sc->isrlock));
   sx_xunlock(&(sc->sxlock));
   return revents;
}

/* Interrupt service routine */
static void cx88_audio_intr(void *arg)
{   
   cx88_audio_softc *sc = arg;
   mtx_lock(&(sc->isrlock));
   /* Ensure we actually have an interrupt--use MSTAT register since we
    * only care about interrupts we've enabled */
   uint32_t intrs = cx88_read_reg(&(sc->common), AUDIO_REG_INT_MSTAT);
   boolean_t pcm_intr = FALSE;
   if (intrs)
   {
      if (sc->state < CX88_AUDIO_CAPTURING)
      {
         mtx_unlock(&(sc->isrlock));
         return;  
      }
      if (intrs & AUDIO_INT_IN_SYNC)
      {
         sc->isr_status |= CX88_AUDIO_SYNC_ERROR;
      }
      /* Check for fatal errors */
      if (intrs & AUDIO_INT_OPCODE_ERR)
      {
         sc->isr_status |= CX88_AUDIO_DMA_PROGRAM;
      }
      if (intrs & (AUDIO_INT_PCI_ABORT | AUDIO_INT_PAR_ERR | AUDIO_INT_RIP_ERR))
      {
         sc->isr_status |= CX88_AUDIO_BUS_ERROR;
      }
      if (sc->isr_status & CX88_AUDIO_CAP_FATAL)
      {
         /* We should stop the transfer as soon as possible if we've encountered a fatal error */
         cx88_write_reg(&(sc->common), AUDIO_REG_DMA_CONTROL, 0);
      }
      if (intrs & AUDIO_INT_IN_RISCI1)
      {
         /* The counter indicates the most recently completed buffer */
         uint32_t bufindex = cx88_read_reg(&(sc->common), AUDIO_REG_IN_GP_COUNT);
         if (bufindex < sc->channel.num_buffers)
         {
            /* Perform any necessary cache/bounce buffer synchronization.
             * Do this in a loop to account for the unlikely case where
             * we've missed an interrupt and multiple buffers need to be 
             * synchronized */
            uint8_t sync_bufindex = sc->buffer_available;
            do
            {
               if (sync_bufindex >= sc->channel.num_buffers)
               {
                  sync_bufindex = 0;
               }
               cx88_buffer *buffer = &(sc->channel.buffers[sync_bufindex]);
               /*Synchronize the buffer */
               cx88_mark_transfer_complete(buffer);
               cx88_prepare_for_transfer(buffer);
            }
            while ((sync_bufindex++) != bufindex);
            /* Available index is 1-based */
            sc->buffer_available = bufindex + 1;
            sc->data_available = TRUE;
            if (sc->state == CX88_AUDIO_PCM)
            {
               pcm_intr = TRUE;  
            }
         }
      }
      /* Acknowledge the interrupt */
      cx88_write_reg(&(sc->common), AUDIO_REG_INT_STAT, intrs);
      if (!pcm_intr)
      {
         wakeup(&(sc->data_available));
         selwakeup(&(sc->selectors));
      }
   }
   mtx_unlock(&(sc->isrlock));
   if (pcm_intr)
   {
      chn_intr(sc->pcm_chan);
   }
}

static int cx88_audio_dmaprog(cx88_buffer *buffer, uint8_t bufindex)
{
   int error = 0;
   DIAGNOSTIC(device_printf(buffer->channel->sc->dev, "Generating DMA program\n"))
   if (bufindex == 0)
   {
      /* Initial SYNC at beginning of capture */
      cx88_risc_sync(buffer, 0, TRUE, CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
   }
   /* Transfer the entire buffer */
   error = cx88_risc_transfer_dmabuf (buffer, 0, (uint32_t)(buffer->data_buffer_size), CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
   if (!error)
   {
      /* Jump to next buffer, adjust the counter to reflect the index of the completed
       * buffer, and trigger RISC interrupt 1 so that the buffer can be processed */
      error = cx88_risc_next_buffer(
         buffer,
         (bufindex == (buffer->channel->num_buffers - 1) ? CX88_RISC_SYNC_SIZE : 0),
         CX88_TRIGGER_RISCI1,
         ((bufindex == 0) ? CX88_RESET_COUNT : CX88_INCREMENT_COUNT));
   }
   return error;
}

static int cx88_audio_probe(device_t dev) 
{
   if ((pci_get_vendor(dev) == CX88_VID) && (pci_get_device(dev) == 0x8811))
   {
      device_set_desc(dev, cx88_audio_client.description);
      return BUS_PROBE_DEFAULT;  
   }
   return cx88_common_probe(dev, &cx88_audio_client);
}

static int cx88_audio_attach(device_t dev) 
{
   return cx88_common_attach(dev, &cx88_audio_client);
}

/* I2C bus accessor DEVMETHOD */
static cx88_i2c_bus *cx88_audio_get_i2c_bus(device_t dev)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx88_audio_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return cx88_i2c_get_bus(&(sc->i2c));
}

/* I2C device accessor DEVMETHOD */
static cx88_i2c_dev *cx88_audio_get_i2c_dev(device_t dev, uint8_t addr)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx88_audio_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return cx88_i2c_get_dev(&(sc->i2c), addr);
}

static void cx88_audio_reconfigure(cx88_audio_softc *sc)
{
   cx88_free_sram(&(sc->channel));
   if (sc->state > CX88_AUDIO_UNCONFIGURED)
   {
      sc->state = CX88_AUDIO_UNCONFIGURED;  
   }
}

static int cx88_audio_request_reconfig(device_t dev)
{   
   int error = 0;
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx88_audio_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return ENODEV;
   }
   /* Since this may be called from a sibling driver in a different thread
    * than any thread that is doing I/O with this driver, use trylock
    * a) because blocking on the lock could deadlock and
    * b) because the sibling wouldn't want to wait an indeterminate amount
    *    of time for I/O on this driver to complete */
   if (sx_try_xlock(&(sc->sxlock)))
   {
      if ((sc->in_use) || (sc->state == CX88_AUDIO_PCM))
      {
         /* If someone has a handle open to our device,
          * don't disrupt their session.  We'll just free our resources
          * when the handle is closed that they can be re-allocated
          * the next time a handle is opened. */
         sc->reconfig_requested = TRUE;
         error = EINPROGRESS;
      }
      else
      {
         cx88_audio_reconfigure(sc);  
      }
      sx_xunlock(&(sc->sxlock));
   }
   else
   {
      sc->reconfig_requested = TRUE;
      error = EINPROGRESS;
   }
   return error; 
}

/* Shutdown DEVMETHOD */
static int cx88_audio_shutdown(device_t dev)
{
   cx88_audio_softc *sc = device_get_softc(dev);
   cx88_audio_reset(sc, CX88_AUDIO_INVALID);
   return 0;
}

/* Suspend DEVMETHOD */
static int cx88_audio_suspend(device_t dev)
{
   cx88_audio_softc *sc = device_get_softc(dev);
   sx_slock(&(sc->sxlock));
   cx88_audio_reset(sc, CX88_AUDIO_UNCONFIGURED);
   sx_sunlock(&(sc->sxlock));
   sx_xlock(&(sc->sxlock));
   cx88_release_channel(&(sc->channel));
   sx_xunlock(&(sc->sxlock));
   return 0;
}

static device_method_t cx88_audio_methods[] = 
{
   /* Device interface */
   DEVMETHOD(device_probe,             cx88_audio_probe),
   DEVMETHOD(device_attach,            cx88_audio_attach),
   DEVMETHOD(device_detach,            cx88_common_detach),
   DEVMETHOD(device_shutdown,          cx88_audio_shutdown),
   DEVMETHOD(device_suspend,           cx88_audio_suspend),
   DEVMETHOD(device_resume,            cx88_common_resume),

   /* cx88 interface */
   CX88_METHODS(cx88_audio_request_reconfig),

   /* i2c host methods */
   CX88_I2C_METHODS(cx88_audio_get_i2c_bus, cx88_audio_get_i2c_dev),

   {NULL, NULL}
};

static void *cx88_pcmchan_init(
   kobj_t obj,
   void *devinfo,
   struct snd_dbuf *b,
   struct pcm_channel *c,
   int dir)
{
   cx88_audio_softc *sc = devinfo;
   sc->pcm_buf = b;
   sc->pcm_chan = c;
   sc->pcm_bufsize = pcm_getbuffersize(sc->pcm_dev,
      MIN_AUDIO_BUFFER_SIZE, DEFAULT_AUDIO_BUFFER_SIZE, MAX_AUDIO_BUFFER_SIZE);
   sc->pcm_databuf = malloc(sc->pcm_bufsize, M_DEVBUF, M_NOWAIT);
   if (sc->pcm_databuf == NULL)
   {
      sc->pcm_bufsize = 0;
      return NULL;
   }
   if (sndbuf_setup(b, sc->pcm_databuf, sc->pcm_bufsize))
   {
      sc->pcm_bufsize = 0;
      free(sc->pcm_databuf, M_DEVBUF);
      sc->pcm_databuf = NULL;
      return NULL;
   }
   return sc;
}

static int cx88_pcmchan_free(kobj_t obj, void *data)
{
   cx88_audio_softc *sc = data;
   sc->pcm_bufsize = 0;
   free(sc->pcm_databuf, M_DEVBUF);  
   sc->pcm_databuf = NULL;
   return 0;
}

static int cx88_pcmchan_setformat(kobj_t obj, void *data, uint32_t format)
{   
   int error = 0;
   cx88_audio_softc *sc = data;
   if (!(sx_try_xlock(&(sc->sxlock))))
   {
      return EBUSY;  
   }
   if (sc->user_configured)
   {
      error = EBUSY;  
   }
   else if (format & (~(AFMT_STEREO | AFMT_S16_LE)))
   {
      error = ENOTSUP;
   }
   else
   {
      sc->stereo_format = ((format & AFMT_STEREO) ? CX88_AUDIO_STEREO_ON : CX88_AUDIO_STEREO_OFF);
   }
   sx_xunlock(&(sc->sxlock));
   return error;
}
   
static uint32_t cx88_pcmchan_setspeed (kobj_t obj, void *data, uint32_t speed)
{
   return CX88_AUDIO_SPEED;  
}
      
static uint32_t cx88_pcmchan_setblocksize(kobj_t obj, void *data, uint32_t blocksize)
{
   int error = 0;
   cx88_audio_softc *sc = data;
   sx_xlock(&(sc->sxlock));
   if (sc->user_configured)
   {
      error = EBUSY;
   }
   else if (sc->pcm_buf == NULL)
   {
      error = ENXIO;
   }
   else
   {
      cx88_audio_config config;
      config.in_num_samples = sc->pcm_bufsize / CX88_AUDIO_SAMPLE_SIZE;
      config.in_buftype = CX88_AUDIO_BUF_KERNEL;
      error = cx88_audio_configure(sc, &config, blocksize, NULL);
   }
   if (error)
   {
      sc->pcm_blocksize = 0;
   }
   sx_xunlock(&(sc->sxlock));
   return sc->pcm_blocksize;
}

static int cx88_pcmchan_trigger (kobj_t obj, void *data, int go)
{
   int error = 0;
   cx88_audio_softc *sc = data;
   if (!(sx_try_xlock(&(sc->sxlock))))
   {
      return EBUSY;  
   }
   switch (go)
   {
      case PCMTRIG_START:
         if (sc->user_configured)
         {
            error = EBUSY;
         }
         else
         {
            error = cx88_i2c_request_devices(&(sc->i2c));
            if (!error)
            {
               error = cx88_audio_start_capture(sc);
            }
            if (!error)
            {
               sc->state = CX88_AUDIO_PCM;
            }
         }
         break;
      case PCMTRIG_ABORT:
      case PCMTRIG_STOP:
         if (sc->state == CX88_AUDIO_PCM)
         {
            cx88_audio_reset(sc, sc->state);
            if (!(sc->in_use))
            {
               cx88_audio_close_session(sc);
            }
         }
         break;
      default:
         break;
   }
   sx_xunlock(&(sc->sxlock));
   return error;
}

static uint32_t cx88_pcmchan_getptr(kobj_t obj, void *data)
{
   cx88_audio_softc *sc = data;
   if (sc->buffer_available <= sc->channel.num_buffers)
   {
      if (sc->buffer_available == sc->channel.num_buffers)
      {
         return sc->pcm_bufsize;  
      }
      else
      {
         return (sc->buffer_available * sc->pcm_blocksize);  
      }
   }
   return 0;
}

static struct pcmchan_caps * cx88_pcmchan_getcaps(kobj_t obj, void *devinfo)
{
   return (&cx88_audio_caps);
}

static kobj_method_t cx88_pcmchan_methods[] = 
{
   KOBJMETHOD(channel_init,            cx88_pcmchan_init),
   KOBJMETHOD(channel_free,            cx88_pcmchan_free),
   KOBJMETHOD(channel_setformat,       cx88_pcmchan_setformat),
   KOBJMETHOD(channel_setspeed,        cx88_pcmchan_setspeed),
   KOBJMETHOD(channel_setblocksize,    cx88_pcmchan_setblocksize),
   KOBJMETHOD(channel_trigger,         cx88_pcmchan_trigger),
   KOBJMETHOD(channel_getptr,          cx88_pcmchan_getptr),
   KOBJMETHOD(channel_getcaps,         cx88_pcmchan_getcaps),
   {NULL, NULL}
};
CHANNEL_DECLARE(cx88_pcmchan);

static int cx88_mixer_init(struct snd_mixer *m)
{
   mix_setdevs(m, SOUND_MASK_LINE);
   mix_setrecdevs(m, SOUND_MASK_LINE);
   return 0;  
}

static int cx88_mixer_uninit(struct snd_mixer *m)
{
   cx88_audio_softc *sc = mix_getdevinfo(m);
   cx88_audio_mute(sc);
   return 0;  
}

static int cx88_mixer_set(
   struct snd_mixer *m,
   unsigned int dev,
   unsigned int left,
   unsigned int right)
{
   cx88_audio_softc *sc = mix_getdevinfo(m);
   cx88_audio_volume vol;
   vol.in_left = (uint8_t)left;
   vol.in_right = (uint8_t)right;
   cx88_audio_set_volume(sc, &vol);
   return (left | (right << 8)); 
}

static int cx88_mixer_setrecsrc(struct snd_mixer *m, uint32_t src)
{
   return SOUND_MASK_LINE;  
}

static kobj_method_t cx88_mixer_methods[] = 
{
   KOBJMETHOD(mixer_init,          cx88_mixer_init),
   KOBJMETHOD(mixer_uninit,        cx88_mixer_uninit),
   KOBJMETHOD(mixer_set,           cx88_mixer_set),
   KOBJMETHOD(mixer_setrecsrc,     cx88_mixer_setrecsrc),
   {NULL, NULL}
};
MIXER_DECLARE(cx88_mixer);

static int cx88_pcm_probe(device_t dev)
{
   device_set_desc(dev, "CX2388x PCM interface");
   return BUS_PROBE_SPECIFIC;
}

static int cx88_pcm_attach(device_t dev)
{
   int error = 0;
   char status[SND_STATUSLEN];
   cx88_audio_softc *sc = device_get_softc(device_get_parent(dev));
   error = mixer_init(dev, &cx88_mixer_class, sc);
   if (!error)
   {
      pcm_setflags(dev, pcm_getflags(dev) | SD_F_MPSAFE); 
      error = pcm_register(dev, sc, 0, 1);
   }
   if (!error)
   {
      error = pcm_addchan(dev, PCMDIR_REC, &cx88_pcmchan_class, sc);
   }
   if (!error)
   {
      snprintf(status, SND_STATUSLEN, "on %s", device_get_nameunit(device_get_parent(dev)));
      pcm_setstatus(dev, status);
   }
   return error;
}

static int cx88_pcm_detach(device_t dev)
{
   return pcm_unregister(dev);
}

static device_method_t cx88_pcm_methods[] = 
{
   DEVMETHOD(device_probe,             cx88_pcm_probe),
   DEVMETHOD(device_attach,            cx88_pcm_attach),
   DEVMETHOD(device_detach,            cx88_pcm_detach),
   {NULL, NULL}
};

static driver_t cx88_audio_driver =
{
   "cx88audio",
   cx88_audio_methods,
   0
};

DRIVER_MODULE(cx88audio, pci, cx88_audio_driver, cx88_audio_devclass, NULL, NULL);
DRIVER_MODULE(cx88audio, cardbus, cx88_audio_driver, cx88_audio_devclass, NULL, NULL);
CX88_DRIVER_MODULE(cx88audio);
CX88_I2C_DRIVER_MODULE(cx88audio);
MODULE_DEPEND(cx88audio, iicdev, IICDEV_MINVER, IICDEV_MODVER, IICDEV_MAXVER);
MODULE_VERSION(cx88audio, CX88_AUDIO_MODVER);

static driver_t cx88_pcm_driver = 
{
   "pcm",
   cx88_pcm_methods,
   PCM_SOFTC_SIZE
};

DRIVER_MODULE(cx88pcm, cx88audio, cx88_pcm_driver, pcm_devclass, NULL, NULL);
MODULE_DEPEND(cx88pcm, cx88audio, CX88_AUDIO_MINVER, CX88_AUDIO_MODVER, CX88_AUDIO_MAXVER); MODULE_DEPEND(cx88pcm, sound, SOUND_MINVER, SOUND_PREFVER, SOUND_MAXVER);
MODULE_VERSION(cx88pcm, CX88_AUDIO_MODVER);
