/*-
 * Copyright 2008 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <strings.h>
#include <time.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/mman.h>

#include <sstream>
using namespace std;

#include "cx88_video_capture_config.h"
#include "cx88_analog_driver.h"
               
#define CX88_ANALOG_DRIVER_VIDEO_FRAMES   8
#define CX88_ANALOG_DRIVER_PAL_COLS       768
#define CX88_ANALOG_DRIVER_PAL_ROWS       576
#define CX88_ANALOG_DRIVER_NTSC_COLS      640
#define CX88_ANALOG_DRIVER_NTSC_ROWS      480
#define CX88_ANALOG_DRIVER_MAX_COLS       768
#define CX88_ANALOG_DRIVER_MAX_ROWS       576

cx88_analog_driver::cx88_analog_driver(
   cx88_xml_config &cx88conf,
   tuner_config &tunerconf,
   const char *devnode,
   capvec &captures,
   cx88_analog_mode mode,
   cx88_analog_driver *slave,
   cx88_analog_driver *pip_driver,
   int &error)
   : cx88_driver(cx88conf, tunerconf, devnode, captures, error),
     m_mode(mode),
     m_slave(slave),
     m_pip_driver(pip_driver),
     m_slave_mode(false),
     m_pip_mode(false),
     m_buffer(NULL)
{
   if (error)
   {
      return;
   }
   if (slave != NULL)
   {
      if ((slave->m_mode == mode) && (slave->m_slave != NULL))
      {
         CX88ERR << "Analog slave driver cannot have slave or same mode as master" << endl;
         error = EINVAL;
         return;  
      }  
      slave->m_slave_mode = true;
   }
   if (pip_driver != NULL)
   {
      if ((mode != CX88_ANALOG_MODE_VIDEO) && (pip_driver->m_mode != CX88_ANALOG_MODE_VIDEO))
      {
         CX88ERR << "Picture-in-Picture drivers must both be in video mode" << endl;
         error = EINVAL;
         return;  
      }
      pip_driver->m_pip_mode = true;
   }
}

cx88_analog_driver::~cx88_analog_driver(void)
{}

int cx88_analog_driver::capture(const string &channel_string, cx88_capture_params &params)
{
   int error = 0; 
   cx88_common_i2c_config i2config;
   i2config.in_num_i2c_devices = 0;
   if ((m_mode == CX88_ANALOG_MODE_VIDEO) && !m_pip_mode)
   {
      m_buffer = (uint8_t*)(malloc(CX88_ANALOG_DRIVER_VIDEO_FRAMES * CX88_ANALOG_DRIVER_MAX_COLS * CX88_ANALOG_DRIVER_MAX_ROWS * 3));
      if (NULL == m_buffer)
      {
         CX88ERR << "Unable to allocate video capture buffer" << endl;
         return ENOMEM;  
      }
      if (m_pip_driver != NULL)
      {
         m_pip_driver->m_buffer = m_buffer;  
      }
   }
   if ((m_fd = open(m_devnode, O_RDWR)) < 0)
   {
      CX88ERR << "Unable to open device node " << m_devnode << ": " << strerror(errno) << endl;
      error = errno; 
   }
   if (m_slave != NULL)
   {
      int fd = open(m_slave->m_devnode, O_RDWR);
      cx88_common_i2c_config slave_i2config;
      slave_i2config.in_num_i2c_devices = 0;
      if (!error && (ioctl(fd, CX88_COMMON_I2C_CONFIG, &slave_i2config) != 0))
      {
         CX88ERR << "Error issuing I2C CONFIG ioctl to reset slave device: " << strerror(errno) << endl;
         error = errno;  
      }
      close(fd);
   }
   if (!m_slave_mode)
   {
      error = (error ? error : i2c_setup(i2config));
   }
   if (!error && ((m_slave != NULL) || (m_pip_driver != NULL)))
   {
      cx88_capture_params slave_pip_params = params;
      slave_pip_params.completion = NULL;
      slave_pip_params.channel_event = NULL;
      if (m_slave != NULL)
      {
         error = m_slave->start(channel_string, slave_pip_params);
      }
      if (m_pip_driver != NULL)
      {
         error = m_pip_driver->start(channel_string, slave_pip_params);
      }
   }
   if (m_slave_mode)
   {
      wait_for_event(0);
   }
   if (m_mode == CX88_ANALOG_MODE_VIDEO)
   {
      error = (error ? error : capture_video(i2config, params));
   }
   else
   {
      error = (error ? error : capture_audio(i2config, params));      
   }
   if (!m_slave_mode)
   {
      i2c_teardown(i2config);
   }
   if (m_fd >= 0)
   {
      close(m_fd);
   }
   if (m_pip_driver != NULL)
   {
      m_pip_driver->stop();
   }
   if (m_slave != NULL)
   {
      m_slave->stop();
   }
   if ((m_mode == CX88_ANALOG_MODE_VIDEO) && !m_pip_mode)
   {
      free(m_buffer);
      m_buffer = NULL;
   }
   if (params.completion != NULL)
   {
      params.completion(params.completion_arg, error);  
   }
   return error;
}

int cx88_analog_driver::change_channel(cx88_common_i2c_config &config, cx88_capture_params &params)
{
   cx88_driver_lock lock(this);
   m_channel_changed = false;
   int error = 0;
   if (m_pip_driver != NULL)
   {
      string::size_type chan_begin = m_channel_name.find_first_of(';', 0);
      if (chan_begin != string::npos)
      {
         string pip_channel = m_channel_name.substr(
            chan_begin + 1, m_channel_name.length() - chan_begin - 1);
         error = m_pip_driver->set_channel(pip_channel);
         if (chan_begin == 0)
         {
            return error;
         }
         string primary_channel = m_channel_name.substr(0, chan_begin);
         m_channel_name = primary_channel;
      }
   }
   if (!error && (m_slave != NULL))
   {
      error = m_slave->set_channel(m_channel_name);
   }
   if (!error)
   {
      error = m_channel.select_avb_channel(m_channel_name, m_avb_channel);
   }
   if (!error)
   {
      error = set_channel_options (m_avb_channel);
   }
   for (uint8_t i = 0; !error && (i < config.in_num_i2c_devices); ++i)
   {
      avb_driver *driver = dynamic_cast<avb_driver*>(m_i2c_devs[i].driver);
      if (driver != NULL)
      {   
         if ((error = driver->set_channel(m_avb_channel)))
         {
            CX88ERR << "Error setting channel on peripheral device" << endl;
         }
      }
   }
   for (uint8_t i = 0; i < config.in_num_i2c_devices; ++i)
   {
      if (m_i2c_devs[i].driver != NULL)
      {   
         if ((error = m_i2c_devs[i].driver->start(params.lock_timeout_ms)))
         {
            CX88ERR << "Error starting peripheral device" << endl;
         }
      }
   }
   return error;
}

int cx88_analog_driver::set_video_format(cx88_video_format &format)
{
   int error = 0;
   format.in_pixel_format = CX88_VIDEO_YCC422;
   format.in_enable_vbi = 0;
   format.in_even_field.enable = 1;
   format.in_even_field.swap = 0;
   format.in_odd_field.enable = 1;
   format.in_odd_field.swap = 0;
   format.in_x_offset = 0;
   format.in_y_offset = 0;
   format.in_interlace = 1;
   switch (m_avb_channel.video_format)
   {
      case AVB_VIDEO_FMT_NTSC_M:
      case AVB_VIDEO_FMT_NTSC_J:
      case AVB_VIDEO_FMT_NTSC_443:
      case AVB_VIDEO_FMT_PAL_M:
      case AVB_VIDEO_FMT_PAL_60:
         format.in_cols = CX88_ANALOG_DRIVER_NTSC_COLS;
         format.in_rows = CX88_ANALOG_DRIVER_NTSC_ROWS;
         break;
      default:
         format.in_cols = CX88_ANALOG_DRIVER_PAL_COLS;
         format.in_rows = CX88_ANALOG_DRIVER_PAL_ROWS;
         break;
   }
   if (AVB_FORMAT_IS_PAL(m_avb_channel.video_format))
   {
      format.in_video_format = CX88_VIDEO_FMT_PALBDGHI;
   }
   else if (AVB_FORMAT_IS_SECAM(m_avb_channel.video_format))
   {
      format.in_video_format = CX88_VIDEO_FMT_SECAM;
   }
   switch (m_avb_channel.video_format)
   {
      case AVB_VIDEO_FMT_NTSC_M:
         format.in_video_format = CX88_VIDEO_FMT_NTSCM;
         break;
      case AVB_VIDEO_FMT_NTSC_J:
         format.in_video_format = CX88_VIDEO_FMT_NTSCJ;
         break;
      case AVB_VIDEO_FMT_NTSC_443:
         format.in_video_format = CX88_VIDEO_FMT_NTSC443;
         break;
      case AVB_VIDEO_FMT_PAL_M:
         format.in_video_format = CX88_VIDEO_FMT_PALM;
         break;
      case AVB_VIDEO_FMT_PAL_N:
         format.in_video_format = CX88_VIDEO_FMT_PALN;
         break;
      case AVB_VIDEO_FMT_PAL_NC:
         format.in_video_format = CX88_VIDEO_FMT_PALNC;
         break;
      case AVB_VIDEO_FMT_PAL_60:
         format.in_video_format = CX88_VIDEO_FMT_PAL60;
         break;
      default:
         break;
   }
   if (m_pip_mode)
   {
      format.in_cols /= 4;
      format.in_rows /= 4;  
   }
   format.in_x_size = format.in_cols;
   format.in_y_size = format.in_rows;
   error = set_format_options(format);
   if (!error)
   {
      if (ioctl(m_fd, CX88_VIDEO_FORMAT, &format) != 0)
      {
         CX88ERR << "Error issuing FORMAT ioctl on video device: " << strerror(errno) << endl;
         error = errno;
      }
   }
   return error;
}

int cx88_analog_driver::capture_video(cx88_common_i2c_config &i2config, cx88_capture_params &params)
{
   int error = 0;
   time_t start_time;
   avb_video_fmt_t standard = AVB_VIDEO_FMT_NONE;
   cx88_video_clip clip;
   bzero(&clip, sizeof(clip));
   if (ioctl(m_fd, CX88_VIDEO_CLIP, &clip) != 0)
   {
      CX88ERR << "Error issuing CLIP ioctl on video device: " << strerror(errno) << endl;
      return errno;
   }
   cx88_video_format format;
   cx88_video_config config;
   config.in_num_frames = CX88_ANALOG_DRIVER_VIDEO_FRAMES;
   config.in_one_shot = 0;
   for (uint8_t i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      config.bufconfig[i].in_buftype = CX88_VIDEO_BUF_USER;
      config.bufconfig[i].in_x_header = 0;
      config.bufconfig[i].in_x_trailer = 0;
      config.bufconfig[i].in_y_header = 0;
      config.bufconfig[i].in_y_trailer = 0;
   }   
   cx88_video_capture capture;
   capture.in_input_source = CX88_VIDEO_INPUT_MUX1;
   capture.in_input_type = CX88_VIDEO_INPUT_COMPOSITE;
   capture.in_input_options = 0;
   capture.in_lock_timeout_ms = 1000;
   capture.in_lock_delay_ms = 100;
   capture.in_no_frame_intrs = 0;
   while (!m_stop)
   {
      error = change_channel(i2config, params);
      if (m_avb_channel.video_format == AVB_VIDEO_FMT_NONE)
      {
         CX88ERR << "Video format cannot be 'NONE'" << endl;
         error = EINVAL;  
      }
      if (!error && (m_avb_channel.video_format != standard))
      {
         error = set_video_format(format);
         standard = m_avb_channel.video_format;
      }
      if (params.channel_event != NULL)
      {
         params.channel_event(params.channel_arg, error);
      }
      if (m_channel_changed)
      {
         continue;
      }
      if (error)
      {
         error = wait_for_event(params.duration_s);
         if (error)
         {
            return error;
         }
         continue;
      }
      if (m_pip_mode)
      {
         for (uint8_t i = 0; i <= CX88_VIDEO_CHANNEL_V; ++i)
         {
            config.bufconfig[i].in_y_trailer = 3 * format.out_geoms[i].num_lines;
            config.bufconfig[i].in_x_trailer = 3 * format.out_geoms[i].line_size;
         }
      }
      else if (m_pip_driver != NULL)
      {
         static const size_t pattern_size = sizeof(cx88_video_clip_pattern) + sizeof(uint16_t);
         uint8_t clipbuf[(CX88_VIDEO_CHANNEL_V + 1) * pattern_size];
         for (uint8_t i = 0; i <= CX88_VIDEO_CHANNEL_V; ++i)
         {
            cx88_video_clip_pattern *pattern = 
               reinterpret_cast<cx88_video_clip_pattern*> (&(clipbuf[i * pattern_size]));
            pattern->in_start_line = 0;
            pattern->in_end_line = (format.out_geoms[i].num_lines / 4) - 1;
            pattern->in_clip_type = CX88_CLIP_EXCLUDE;
            pattern->in_num_points = 1;
            pattern->in_points[0] = format.out_geoms[i].line_size / 4;
            clip.channels[i].in_patterns = pattern;
            clip.channels[i].in_pattern_size = pattern_size;
            if (ioctl(m_fd, CX88_VIDEO_CLIP, &clip) != 0)
            {
               CX88ERR << "Error issuing CLIP ioctl on video device: " << strerror(errno) << endl;
               error = errno;
               goto cleanup;
            }
         }
      }
      size_t bufsize = 0;
      for (uint8_t i = 0; i <= CX88_VIDEO_CHANNEL_V; ++i)
      {
         bufsize += (
            (config.bufconfig[i].in_x_header + format.out_geoms[i].line_size +         
               config.bufconfig[i].in_x_trailer) *
            (config.bufconfig[i].in_y_header + format.out_geoms[i].num_lines +
               config.bufconfig[i].in_y_trailer));
      }
      uint8_t *temp = m_buffer;
      for (uint8_t i = 0; i <= CX88_VIDEO_CHANNEL_V; ++i)
      {
         for (uint8_t j = 0; j < config.in_num_frames; ++j)
         {
            config.bufconfig[i].u.in_userbuf.buffers[j] = temp + (j * bufsize);  
         }
         temp += (
            (config.bufconfig[i].in_x_header + format.out_geoms[i].line_size +         
               config.bufconfig[i].in_x_trailer) *
            (config.bufconfig[i].in_y_header + format.out_geoms[i].num_lines +
               config.bufconfig[i].in_y_trailer));
      }
      if ((error = set_config_options(config)) != 0)
      {
         goto cleanup;
      }
      if (ioctl(m_fd, CX88_VIDEO_CONFIG, &config) != 0)
      {
         CX88ERR << "Error issuing CONFIG ioctl on video device: " << strerror(errno) << endl;
         error = errno;
         goto cleanup;  
      }
      cx88_video_capture_config capconf(format, config);
      if ((error = start_captures(&capconf)) != 0)
      {
         goto cleanup;
      }
      if ((error = set_capture_options(capture)) != 0)
      {
         goto cleanup;
      }
      start_time = time(NULL);
capture:
      if (ioctl(m_fd, CX88_VIDEO_CAPTURE, &capture) != 0)
      {
         CX88ERR << "Error issuing CAPTURE ioctl on video device: " << strerror(errno) << endl;
         error = errno;
         goto cleanup;
      }
      uint8_t last_buffer = 0;
      cx88_video_wait wait;
      wait.in_timeout_ms = 1000;
      wait.in_channel_mask = CX88_VIDEO_FIELD_EVEN(CX88_VIDEO_CHANNEL_Y);
      while (!m_stop && !m_channel_changed)
      {  
         if (ioctl(m_fd, CX88_VIDEO_WAIT, &wait) != 0)
         {
            CX88ERR << "Error issuing WAIT ioctl on video device: " << strerror(errno) << endl;
            error = errno;
            goto cleanup;
         }
         if ((wait.out_status & CX88_VIDEO_SYNC_ERROR) == CX88_VIDEO_SYNC_ERROR)
         {
            CX88ERR << "Warning: DMA sync error; resetting capture" << endl;
            goto capture;
         }        
         if ((wait.out_status & CX88_VIDEO_BUS_ERROR) == CX88_VIDEO_BUS_ERROR)
         {
            CX88ERR << "Fatal error: PCI bus error" << endl;
         }
         if ((wait.out_status & CX88_VIDEO_DMA_PROGRAM) == CX88_VIDEO_DMA_PROGRAM)
         {
            CX88ERR << "Fatal error: DMA program error" << endl;
         }
         if ((wait.out_status & CX88_VIDEO_WAIT_ABORTED) == CX88_VIDEO_WAIT_ABORTED)
         {
            CX88ERR << "Fatal error: Wait aborted" << endl;
         }
         if ((wait.out_status & CX88_VIDEO_CAP_TIMEOUT) == CX88_VIDEO_CAP_TIMEOUT)
         {
            CX88ERR << "Nonfatal error: Capture timed out" << endl;
         }
         if (wait.out_status & CX88_VIDEO_CAP_FATAL)
         {
            error = EIO;
            goto cleanup;
         }
         if (wait.channels[CX88_VIDEO_CHANNEL_Y].out_even_frame_available == 0 ||
            (wait.channels[CX88_VIDEO_CHANNEL_Y].out_even_frame_available > config.in_num_frames))
         {
            error = EIO;
            goto cleanup;
         }
         while (last_buffer != wait.channels[CX88_VIDEO_CHANNEL_Y].out_even_frame_available)
         {
            ++last_buffer;
            if (last_buffer > config.in_num_frames)
            {
               last_buffer = 1;
            }
            submit_buffer(m_buffer + ((last_buffer - 1) * bufsize), bufsize);
         }
         if ((params.duration_s > 0) && ((unsigned int)(time(NULL) - start_time) >= params.duration_s))
         {
            goto cleanup;
         }
      }
      stop_captures();
      ioctl(m_fd, CX88_VIDEO_STOP);
   }
cleanup:
   stop_captures();
   ioctl(m_fd, CX88_VIDEO_STOP);
   return error;
}

int cx88_analog_driver::set_audio_format(void)
{
   int error = 0;
   cx88_audio_format format;
   format.in_second_lang = 0;
   format.in_stereo_format = CX88_AUDIO_STEREO_AUTO;
   format.in_input = CX88_AUDIO_INPUT_BCAST;
   format.in_i2s_config = 0;
   switch (m_avb_channel.audio_format)
   {
      case AVB_AUDIO_FMT_BTSC_SAP:
         format.in_second_lang = 1;
      case AVB_AUDIO_FMT_BTSC:
         format.in_audio_format = CX88_AUDIO_FMT_BTSC;
         break;
      case AVB_AUDIO_FMT_EIAJ_SAP:
         format.in_second_lang = 1;
      case AVB_AUDIO_FMT_EIAJ:
         format.in_audio_format = CX88_AUDIO_FMT_EIAJ;
         break;
      case AVB_AUDIO_FMT_FM_MONO:
      case AVB_AUDIO_FMT_FM_STEREO:
         format.in_audio_format = CX88_AUDIO_FMT_FM_RADIO;
         break;
      case AVB_AUDIO_FMT_FM_MONO_NON_USA:
      case AVB_AUDIO_FMT_FM_STEREO_NON_USA:
         format.in_audio_format = CX88_AUDIO_FMT_FM_RADIO_50;
         break;
      case AVB_AUDIO_FMT_FM_MONO_USA:
      case AVB_AUDIO_FMT_FM_STEREO_USA:
         format.in_audio_format = CX88_AUDIO_FMT_FM_RADIO_75;
         break;
      case AVB_AUDIO_FMT_AM_MONO:
      case AVB_AUDIO_FMT_AM_STEREO:
         format.in_audio_format = CX88_AUDIO_FMT_AM_L;
         break;
      case AVB_AUDIO_FMT_A2_SAP:
         format.in_second_lang = 1;
      case AVB_AUDIO_FMT_A2:
      {
         switch (m_avb_channel.video_format)
         {
            case AVB_VIDEO_FMT_PAL_B:
            case AVB_VIDEO_FMT_PAL_G:
            case AVB_VIDEO_FMT_SECAM_B:
            case AVB_VIDEO_FMT_SECAM_G:
               format.in_audio_format = CX88_AUDIO_FMT_A2_BG;
               break;
            case AVB_VIDEO_FMT_PAL_D:
            case AVB_VIDEO_FMT_PAL_D1:
            case AVB_VIDEO_FMT_SECAM_D:
            case AVB_VIDEO_FMT_PAL_K:
            case AVB_VIDEO_FMT_SECAM_K:
            case AVB_VIDEO_FMT_SECAM_K1:
               format.in_audio_format = CX88_AUDIO_FMT_A2_DK;
               break;   
            default:
               format.in_audio_format = CX88_AUDIO_FMT_A1_I;
               break;
         }
         break;
      }
      case AVB_AUDIO_FMT_NICAM_SAP:
         format.in_second_lang = 1;
      case AVB_AUDIO_FMT_NICAM:
      {
         switch (m_avb_channel.video_format)
         {
            case AVB_VIDEO_FMT_PAL_I:
               format.in_audio_format = CX88_AUDIO_FMT_NICAM_I;
               break;
            case AVB_VIDEO_FMT_SECAM_L:
            case AVB_VIDEO_FMT_SECAM_LC:
               format.in_audio_format = CX88_AUDIO_FMT_NICAM_L;
               break;
            default:
               format.in_audio_format = CX88_AUDIO_FMT_NICAM_BGDK;
               break;
         }
         break;  
      }
      default:
         CX88ERR << "Unrecognized audio format" << endl;
         return EINVAL;
   }
   error = set_format_options(format);
   if (!error)
   {
      if (ioctl(m_fd, CX88_AUDIO_FORMAT, &format) != 0)
      {
         CX88ERR << "Error issuing FORMAT ioctl on audio device: " << strerror(errno) << endl;
         error = errno;
      }
   }
   return error;
}

int cx88_analog_driver::capture_audio(cx88_common_i2c_config &i2config, cx88_capture_params &params)
{
   int error = 0;
   time_t start_time;
   avb_audio_fmt_t standard = AVB_AUDIO_FMT_NONE;
   cx88_audio_config config;
   m_buffer = (uint8_t*)MAP_FAILED;
   if (m_captures.size() > 0)
   {
      config.in_buftype = CX88_AUDIO_BUF_KERNEL;
      config.in_num_samples = 8000;
      if (ioctl(m_fd, CX88_AUDIO_CONFIG, &config) != 0)
      {
         CX88ERR << "Error issuing CONFIG ioctl on audio device: " << strerror(errno) << endl;
         return errno;
      }
      if ((error = set_config_options(config)) != 0)
      {
         return error;
      }
      m_buffer = (uint8_t*)(mmap(NULL, config.bufconfig.out_kernelbuf.out_mmap_size, PROT_READ, MAP_SHARED, m_fd, 0));
      if (m_buffer == (uint8_t*)MAP_FAILED)
      {
         CX88ERR << "Unable to map buffer memory from audio device: " << strerror(errno) << endl;
         return errno;  
      }
      if ((error = start_captures()))
      {
         return error;
      }
   }
   while (!m_stop)
   {
      error = change_channel(i2config, params);
      if (m_avb_channel.audio_format == AVB_AUDIO_FMT_NONE)
      {
         CX88ERR << "Audio format cannot be 'NONE'" << endl;
         error = EINVAL;  
      }
      if (m_avb_channel.audio_format == AVB_AUDIO_FMT_AUTO)
      {
         CX88ERR << "'AUTO' audio format not currently supported" << endl;
         error = EINVAL;
      }
      if (!error && (m_avb_channel.audio_format != standard))
      {
         error = set_audio_format();
         standard = m_avb_channel.audio_format;
      }
      if (params.channel_event != NULL)
      {
         params.channel_event(params.channel_arg, error);
      }
      if (error || (m_captures.size() == 0))
      {
         error = wait_for_event(params.duration_s);
         if (error)
         {
            return error;
         }
         continue;
      }
      start_time = time(NULL);
      if (ioctl(m_fd, CX88_AUDIO_CAPTURE) != 0)
      {
         CX88ERR << "Error issuing CAPTURE ioctl on audio device: " << strerror(errno) << endl;
         error = errno;
         goto cleanup;
      }
      uint8_t last_buffer = 0;
      cx88_audio_wait wait;
      wait.in_timeout_ms = 1000;
      while (!m_stop && !m_channel_changed)
      {        
         if (ioctl(m_fd, CX88_AUDIO_WAIT, &wait) != 0)
         {
            CX88ERR << "Error issuing WAIT ioctl on audio device: " << strerror(errno) << endl;
            error = errno;
            goto cleanup;
         }
         if ((wait.out_status & CX88_AUDIO_SYNC_ERROR) == CX88_AUDIO_SYNC_ERROR)
         {
            CX88ERR << "Fatal error: DMA sync error" << endl;
         }        
         if ((wait.out_status & CX88_AUDIO_BUS_ERROR) == CX88_AUDIO_BUS_ERROR)
         {
            CX88ERR << "Fatal error: PCI bus error" << endl;
         }
         if ((wait.out_status & CX88_AUDIO_DMA_PROGRAM) == CX88_AUDIO_DMA_PROGRAM)
         {
            CX88ERR << "Fatal error: DMA program error" << endl;
         }
         if ((wait.out_status & CX88_AUDIO_WAIT_ABORTED) == CX88_AUDIO_WAIT_ABORTED)
         {
            CX88ERR << "Fatal error: Wait aborted" << endl;
         }
         if ((wait.out_status & CX88_AUDIO_CAP_TIMEOUT) == CX88_AUDIO_CAP_TIMEOUT)
         {
            CX88ERR << "Nonfatal error: Capture timed out" << endl;
         }
         if ((wait.out_status & CX88_AUDIO_BERTHRESH) == CX88_AUDIO_BERTHRESH)
         {
            CX88ERR << "Nonfatal error: BER threshold exceeded" << endl;
         }
         if ((wait.out_status & CX88_AUDIO_OVERFLOW) == CX88_AUDIO_OVERFLOW)
         {
            CX88ERR << "Nonfatal error: overflow detected" << endl;
         }
         if (wait.out_status & CX88_AUDIO_CAP_FATAL)
         {
            error = EIO;
            goto cleanup;
         }
         if (wait.out_buffer_available == 0 || wait.out_buffer_available > config.out_num_buffers)
         {
            error = EIO;
            goto cleanup;
         }
         while (last_buffer != wait.out_buffer_available)
         {
            ++last_buffer;
            if (last_buffer > config.out_num_buffers)
            {
               last_buffer = 1;
            }
            size_t bufoffset = (last_buffer - 1) * config.bufconfig.out_kernelbuf.out_buffer_size;
            size_t bufsize = config.bufconfig.out_kernelbuf.out_mmap_size - bufoffset;
            void *bufptr = m_buffer + bufoffset;
            if (bufsize > (config.out_samples_per_buffer * CX88_AUDIO_SAMPLE_SIZE))
            {
               bufsize = config.out_samples_per_buffer * CX88_AUDIO_SAMPLE_SIZE;
            }
            submit_buffer(bufptr, bufsize, CX88_AUDIO_SAMPLE_SIZE);
         }
         if ((params.duration_s > 0) && ((unsigned int)(time(NULL) - start_time) >= params.duration_s))
         {
            goto cleanup;
         }
      }
      ioctl(m_fd, CX88_AUDIO_STOP);
   }
cleanup:
   stop_captures();
   ioctl(m_fd, CX88_AUDIO_STOP);
   if (m_buffer != MAP_FAILED)
   {
      munmap(m_buffer, config.bufconfig.out_kernelbuf.out_mmap_size);
      m_buffer = (uint8_t*)MAP_FAILED;
   }
   return error;
}

int cx88_analog_driver::get_status(cx88_audio_status &status)
{ 
   status.out_carrier = CX88_AUDIO_NO_CARRIER;
   if (ioctl(m_fd, CX88_AUDIO_GET_STATUS, &status) != 0)
   {
      CX88ERR << "Unable to retrieve audio signal status: " << strerror(errno) << endl;
      return errno;
   }
   return 0;
}

int cx88_analog_driver::get_status(cx88_video_status &status)
{
   status.out_video_present = 0;
   if (ioctl(m_fd, CX88_VIDEO_GET_STATUS, &status) != 0)
   {
      CX88ERR << "Unable to retrieve video signal status: " << strerror(errno) << endl;
      return errno;
   }
   return 0;
}

double cx88_analog_driver::get_signal_strength(void)
{
   cx88_driver_lock lock(this);
   if (m_mode != CX88_ANALOG_MODE_VIDEO)
   {
      cx88_audio_status status;
      get_status(status);
      if (status.out_carrier != CX88_AUDIO_NO_CARRIER)
      {
         return 100.0;  
      }
      return 0.0;
   }
   else
   {
      cx88_video_status status;
      get_status(status);
      if (status.out_video_present)
      {
         return 100.0;
      }
      return 0.0;
   }
}

int cx88_analog_driver::get_info(string &info)
{
   cx88_driver_lock lock(this);
   int error = 0;
   try
   {
      ostringstream stream;
      string channel;
      if (!m_slave_mode)
      {
         m_channel.get_channel_name(channel);
         stream << "Tuned to channel " << channel << endl;
      }
      if (m_mode == CX88_ANALOG_MODE_VIDEO)
      {
         cx88_video_status status;
         if ((error = get_status(status)))
         {
            return error;
         }
         stream << "\tVideo:" << endl;
         if (status.out_video_present == 0)
         {
            stream << "\t\tNo signal" << endl;
         }
         else
         {
            if (status.out_play_mode == CX88_VIDEO_MODE_SP)
            {
               stream << "\t\tVCR special play mode" << endl;
            }
            else
            {
               stream << "\t\tNormal play mode" << endl;  
            }
            if (status.out_horizontal_lock == 0)
            {
               stream << "\t\tNo horizontal lock" << endl;
            }
            if (status.out_system_type == CX88_VIDEO_SYSTEM_625)
            {
               stream << "\t\tUsing 625-line broadcast system" << endl;
            }
            else
            {
               stream << "\t\tUsing 525-line broadcast system" << endl;
            }
         }
      }
      else
      {
         cx88_audio_status status;
         if ((error = get_status(status)))
         {
            return error;
         }
         stream << "\tAudio:" << endl;
         if (status.out_carrier == CX88_AUDIO_NO_CARRIER)
         {
            stream << "\t\tNo signal" << endl;
         }
         else
         {
            if (status.out_carrier & CX88_AUDIO_CARRIER_1)
            {
               stream << "\t\tPrimary carrier detected" << endl;
            }
            if (status.out_carrier & CX88_AUDIO_CARRIER_2)
            {
               stream << "\t\tSecondary carrier detected" << endl;
            }
            switch (status.out_signal)
            {
               case CX88_AUDIO_SIGNAL_STEREO:
                  stream << "\t\tStereo" << endl;
                  break;
               case CX88_AUDIO_SIGNAL_DUAL_MONO:
                  stream << "\t\tDual mono" << endl;
                  break;
               case CX88_AUDIO_SIGNAL_MONO:
                  stream << "\t\tMono" << endl;
                  break;
               case CX88_AUDIO_SIGNAL_SAP:
                  stream << "\t\tSAP" << endl;
                  break;  
            }
            if (status.out_mute)
            {
               stream << "\t\tMuted" << endl;
            }
         }
      }
      info = stream.str();
      if (!error && (m_slave != NULL))
      {
         string slaveinfo;
         m_slave->get_info(slaveinfo);  
         info += slaveinfo;
      }
   }
   catch(...)
   {
      CX88ERR << "Insufficient memory to generate info string" << endl;
      error = ENOMEM;
   }
   return 0;
}
