/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <sys/errno.h>
#include "cx88_driver.h"

cx88_driver::cx88_driver(
   cx88_xml_config &cx88conf,
   tuner_config &tunerconf,
   const char *devnode,
   capvec &captures,
   int &error)
   : m_channel(cx88conf, tunerconf, error),
     m_tunerconf(tunerconf),
     m_devnode(devnode),
     m_stop(true),
     m_channel_changed(false),
     m_fd(-1),
     m_captures(captures),
     m_mutex_created(false),
     m_signal_created(false)
{
   if ((error = pthread_mutex_init(&m_lock, NULL)))
   {
      CX88ERR << "Unable to initialize driver lock" << endl;
      return;  
   }
   m_mutex_created = true;
   if ((error = pthread_cond_init(&m_signal, NULL)))
   {
      CX88ERR << "Unable to initialize driver event signal" << endl;
      return;  
   }
   m_signal_created = true;
   for (int i = 0; i < CX88_COMMON_MAX_I2C_DEVICES; ++i)
   {
      m_i2c_devs[i].dev = NULL;
      m_i2c_devs[i].driver = NULL;
      m_i2c_devs[i].addr = 0;  
   }
}

cx88_driver::~cx88_driver(void)
{
   if (m_signal_created)
   {
      pthread_cond_destroy(&m_signal);
      m_signal_created = false;  
   }
   if (m_mutex_created)
   {
      pthread_mutex_destroy(&m_lock);
      m_mutex_created = false;
   }
}

int cx88_driver::set_channel(const string &channel_string)
{
   cx88_driver_lock lock(this);
   if (!m_stop)
   {
      m_channel_changed = true;
      m_channel_name = channel_string;
      int error = change_channel(channel_string);
      pthread_cond_signal(&m_signal);
      return error;
   }
   else
   {
      CX88ERR << "Unable to set channel: driver has not been started" << endl;
      return ENXIO;
   }
}

int cx88_driver::start(const string &channel_string, cx88_capture_params &params)
{
   cx88_driver_lock lock(this);
   int error = 0;
   if (m_stop)
   {
      m_params = params;
      m_channel_name = channel_string;
      m_stop = false;
      if ((error = pthread_create(&m_thread, NULL, thread_routine, this)))
      {
         CX88ERR << "Unable to create driver thread: " << strerror(error) << endl;
         m_stop = true;
         return error;
      }
   }
   return error;
}

int cx88_driver::stop(void)
{
   void *retval = NULL;
   {
      cx88_driver_lock lock(this);
      if (!m_stop)
      {
         m_stop = true;
         pthread_cond_signal(&m_signal);
      }
   }
   pthread_join(m_thread, &retval);
   return (reinterpret_cast <size_t> (retval));
}

void *cx88_driver::thread_routine(void *arg)
{
   cx88_driver *driver = reinterpret_cast<cx88_driver*>(arg);
   int error = driver->capture(driver->m_channel_name, driver->m_params);
   return ((void*)error);
}

int cx88_driver::start_captures(cx88_capture_config *cfg)
{
   int error = 0;
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      if ((error = m_captures[i]->start(cfg)))
      {
         CX88ERR << "Error starting captures: " << strerror(error) << endl;
         break;
      }
   }
   return error;
}

int cx88_driver::submit_buffer(void *buffer, size_t num_bytes, size_t packet_size)
{
   int error = 0;
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      if ((error = m_captures[i]->submit_buffer(buffer, num_bytes, packet_size)))
      {
         CX88ERR << "Error submitting buffer to captures: " << strerror(error) << endl;
         break;
      }
   }
   return error;
}

void cx88_driver::stop_captures(void)
{
   for (capvec::size_type i = 0; i < m_captures.size(); ++i)
   {
      m_captures[i]->stop();
   }
}

int cx88_driver::wait_for_event(uint32_t timeout_s)
{
   cx88_driver_lock lock(this);
   if (!m_stop && !m_channel_changed)
   {
      if (timeout_s == 0)
      {
         return pthread_cond_wait(&m_signal, &m_lock);
      }
      else
      {
         struct timespec systime;
         systime.tv_sec = time(NULL) + timeout_s;
         systime.tv_nsec = 0;
         return pthread_cond_timedwait(&m_signal, &m_lock, &systime);  
      }
   }
   return 0;
}

int cx88_driver::i2c_setup(cx88_common_i2c_config &config)
{
   int error = create_i2c_devices(config);
   if (!error)
   {
      error = ioctl(m_fd, CX88_COMMON_I2C_CONFIG, &config);
   }
   if (error)
   {
      CX88ERR << "Error issuing I2C CONFIG ioctl: " << strerror(errno) << endl;
      return errno;  
   }
   for (uint8_t i = 0; i < config.in_num_i2c_devices; ++i)
   {
      char i2c_devnode[CX88_COMMON_MAX_DEVNODE_NAME + 5];
      snprintf(i2c_devnode, CX88_COMMON_MAX_DEVNODE_NAME, "/dev/%s", config.out_i2c_devnodes[i]);
      m_i2c_devs[i].dev = new tuner_devnode_device(m_tunerconf, i2c_devnode, error);
      if (m_i2c_devs[i].dev == NULL) 
      {
         error = ENOMEM;
      }
      if (error)
      {
         CX88ERR << "Unable to create control instance for peripheral device" << endl;
         break;  
      }
      m_i2c_devs[i].addr = config.in_i2c_addrs[i];
   }
   if (!error)
   {
      error = create_i2c_drivers();  
   }
   return error;
}

void cx88_driver::i2c_teardown(cx88_common_i2c_config &config)
{
   destroy_i2c_drivers();
   for (uint8_t i = 0; i < config.in_num_i2c_devices; ++i)
   {
      if (m_i2c_devs[i].dev != NULL)
      {
         delete m_i2c_devs[i].dev;
         m_i2c_devs[i].dev = NULL;
         m_i2c_devs[i].driver = NULL;
      }
   }
}
