#include <sys/errno.h>
#include "cx88_mpegps_filter.h"

#define MPEG_START_BYTE 0x47

cx88_mpegps_filter::cx88_mpegps_filter(capvec &captures, int &error)
   : cx88_capture(error),
     m_captures(captures),
     m_stream(0),
     m_handle(NULL),
     m_state(CX88_MPEGPS_IDLE)
{
   if (error)
   {
      return;
   }
   try
   {
      m_pids.reserve(10);
   }
   catch (...)
   {
      CX88ERR << "Unable to reserve space for stream PIDs" << endl;
      error = ENOMEM;
   }
}

int cx88_mpegps_filter::initialize(void)
{
   if (m_stream > 0)
   {
      m_handle = dvbpsi_AttachPAT(pat_complete, this);
      if (m_handle == NULL)
      {
         return ENOMEM;  
      }
   }
   m_state = CX88_MPEGPS_STARTED;
   return 0;
}

void cx88_mpegps_filter::flush(void)
{
   if (m_state == CX88_MPEGPS_PAT_FOUND)
   {
      dvbpsi_DetachPMT(m_handle);
   }
   else if (m_state == CX88_MPEGPS_STARTED)
   {
      dvbpsi_DetachPAT(m_handle);
   }
   m_state = CX88_MPEGPS_IDLE;
}

void cx88_mpegps_filter::process_buffer(void *data, size_t num_bytes, size_t packet_size)
{
   if (m_stream == 0)
   {
      int error = 0;
      for (capvec::size_type i = 0; i < m_captures.size(); ++i)
      {
         if ((error = m_captures[i]->submit_buffer(data, num_bytes, packet_size)))
         {
            CX88ERR << "<MPEG PS filter> Error submitting buffer to captures: " << strerror(error) << endl;
            break;
         }
      }
   }
   else
   {
      size_t i = 0;
      uint8_t *bytebuffer = reinterpret_cast <uint8_t*> (data);
      while ((m_state != CX88_MPEGPS_IDLE) && (m_state != CX88_MPEGPS_PMT_FOUND) && (i <= (num_bytes - packet_size)))
      {
         if (m_handle == NULL)
         {
            CX88ERR << "<MPEG PS filter> Invalid demultiplexer handle" << endl;
            return;
         }
         dvbpsi_PushPacket(m_handle, bytebuffer + i);
         i += packet_size;
      }
      if (i > 0)
      {
         for (capvec::size_type j = 0; j < m_captures.size(); ++j)
         {
            m_captures[j]->submit_buffer(bytebuffer, i, packet_size);
         }
      }
      size_t valid_packet_start = num_bytes;
      while (i <= num_bytes)
      {
         bool pid_in_list = false;
         if (i <= (num_bytes - packet_size))
         {
            uint16_t pid = (bytebuffer[i + 1] & 0x1F) << 8;
            pid |= bytebuffer[i + 2];
            pid_in_list = (find(m_pids.begin(), m_pids.end(), pid) != m_pids.end());
         }
         if ((valid_packet_start == num_bytes) && pid_in_list)
         {
            valid_packet_start = i;
         }
         else if ((valid_packet_start != num_bytes) && !pid_in_list)
         {
            for (capvec::size_type j = 0; j < m_captures.size(); ++j)
            {
               m_captures[j]->submit_buffer(bytebuffer + valid_packet_start, i - valid_packet_start, packet_size);
            }
            valid_packet_start = num_bytes;
         }
         i += packet_size;
      }
   }
}

void cx88_mpegps_filter::pat_complete(void *context, dvbpsi_pat_t *pat)
{
   cx88_mpegps_filter *filter = reinterpret_cast <cx88_mpegps_filter*> (context);
   int i = 1;
   dvbpsi_pat_program_t *program = pat->p_first_program;
   DIAGNOSTIC(CX88LOG << "<MPEG PS filter> PAT complete" << endl)
   while ((i < filter->m_stream) && (program != NULL))
   {
      program = program->p_next;
      ++i;
   }
   if (program != NULL)
   {
      //dvbpsi_DetachPAT(filter->m_handle);
      filter->m_handle = dvbpsi_AttachPMT(program->i_number, cx88_mpegps_filter::pmt_complete, filter);
      filter->m_state = CX88_MPEGPS_PAT_FOUND;
   }
   else
   {
      CX88ERR << "<MPEG PS filter> Invalid program ID" << endl;
      filter->m_state = CX88_MPEGPS_IDLE; 
   }
}

void cx88_mpegps_filter::pmt_complete(void *context, dvbpsi_pmt_t *pmt)
{
   cx88_mpegps_filter *filter = reinterpret_cast <cx88_mpegps_filter*> (context);
   dvbpsi_pmt_es_t *stream = pmt->p_first_es;
   DIAGNOSTIC(CX88LOG << "<MPEG PS filter> PMT complete" << endl)
   while (stream != NULL)
   {
      filter->m_pids.push_back(stream->i_pid);
      stream = stream->p_next;  
   }
   //dvbpsi_DetachPMT(filter->m_handle);
   filter->m_handle = NULL;
   filter->m_state = CX88_MPEGPS_PMT_FOUND;
}
