/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef __CX88_XML_CONFIG_H__
#define __CX88_XML_CONFIG_H__

#include <iostream>
#include <string>
#include <map>
#include <list>
using namespace std;

#ifdef _DIAGNOSTIC
#define DIAGNOSTIC(stmt) stmt;
#else
#define DIAGNOSTIC(stmt)
#endif

#define CX88ERR (cerr << "[cx88] ")
#define CX88LOG (clog << "[cx88] ")
#include <ezxml.h>
#include "dvb_driver.h"
#include "avb_driver.h"

#define CX88_XML_CONFIG_NAME "config"
#define CX88_XML_NAME_NAME "name"
#define CX88_XML_CHANNEL_NAME "channel"
#define CX88_XML_LIBTUNER_NAME "libtuner"
#define CX88_XML_ALIAS_NAME "alias"

#define CX88_XML_PROFILE_NAME "profile"
#define CX88_XML_MODULATION_NAME "modulation"
#define CX88_XML_ANALOG_VIDEO_NAME "analog_video"
#define CX88_XML_ANALOG_AUDIO_NAME "analog_audio"
#define CX88_XML_RANGE_NAME "range"
#define CX88_XML_BANDWIDTH_NAME "bandwidth"
#define CX88_XML_START_NAME "start"
#define CX88_XML_END_NAME "end"
#define CX88_XML_FREQ_NAME "freq"
#define CX88_XML_INVERSION "inversion"
#define CX88_XML_SYMBOL_RATE "symbol_rate"
#define CX88_XML_CODE_RATE "code_rate"
#define CX88_XML_CODE_RATE_HP "hp_code_rate"
#define CX88_XML_CODE_RATE_LP "lp_code_rate"
#define CX88_XML_SUBMODULATION "sub_modulation"
#define CX88_XML_SUBCARRIERS "subcarriers"
#define CX88_XML_GUARD_INTERVAL "guard_interval"
#define CX88_XML_HIERARCHY "hierarchy"
#define CX88_XML_CHANNEL_UP "up"
#define CX88_XML_CHANNEL_DOWN "down"
#define CX88_XML_CHANNEL_FIRST "first"
#define CX88_XML_CHANNEL_LAST "last"
#define CX88_XML_CHANNEL_DEFAULT "default"
               
class cx88_xml_config
{
   public:
      
      cx88_xml_config(const char *filename, int &error);
      
      virtual ~cx88_xml_config(void);
            
   private:
      
      friend class cx88_xml_channel;
      
      ezxml_t m_file;
      
      typedef map<string, string> strmap;
      strmap m_aliases;
      
      string m_default_channel;
};

class cx88_xml_channel
{  
   public:
      
      cx88_xml_channel(cx88_xml_config &config, tuner_config &tunerconf, int &error);
      
      virtual ~cx88_xml_channel(void) {}
            
      virtual int select_profile(const char *name);
      
      virtual const char *get_profile_name(void);
      
      virtual int select_channel(unsigned long channel);
      
      virtual int select_channel(const string &name);
      
      virtual unsigned long get_channel(void)
      {
         return m_channel;
      }
      
      virtual void get_channel_name(string &name);
      
      virtual int channel_up(void);
      
      virtual int channel_down(void);
      
      virtual int select_first_channel(void);
      
      virtual int select_last_channel(void);
      
      virtual bool first_channel(void)
      {
         return ((m_range == m_ranges.begin()) && (m_channel == m_range->start));  
      }
      
      virtual bool last_channel(void)
      {
         rangelist::iterator iter = m_range;
         return (((++iter) == m_ranges.end()) && (m_channel == m_range->end));
      }
      
      virtual const char *get_property(const char *property);
      
      virtual int select_dvb_channel(const string &channel, dvb_channel &dvbchan);
      
      virtual int select_avb_channel(const string &channel, avb_channel &avbchan);
      
      template <typename payload>
      struct table_entry
      {
         const char *name;
         payload value;
      };
      
      template <typename payload>
      static size_t lookup(struct table_entry<payload> *table, size_t table_size, const char *value);
   
   private:

      typedef ezxml_t cx88_profile_t;

      typedef struct 
      {
         ezxml_t elem;
         tuner_config tunerconf;
         unsigned long start;
         unsigned long end;
      } cx88_range_t;
      
      cx88_xml_config &m_config;

      cx88_profile_t m_profile;
      
      typedef list<cx88_range_t> rangelist;
      rangelist m_ranges;
      rangelist::iterator m_range;
            
      cx88_xml_config::strmap::iterator m_alias;
      
      unsigned long m_channel;
      
      tuner_config &m_tunerconf;
      
      static int parse_libtuner(ezxml_t parent, tuner_config &conf);
      
      static int parse_modulation(const char *mod_string, dvb_modulation_t &modulation);
      
      static int parse_fec(const char *fec_string, dvb_fec_t &fec);
      
      int parse_qam(dvb_qam_settings &qam);
      
      int parse_ofdm(dvb_ofdm_settings &qam);
      
      int parse_analog_video(const char *video_string, avb_video_fmt_t &fmt);
      
      int parse_analog_audio(const char *audio_string, avb_audio_fmt_t &fmt);
};

#endif
