/*-
 * Copyright 2008 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <strings.h>
#include "cx88_cards.h"

cx88_card_id get_card_id(tuner_config &conf, const char *devnode)
{
   int fd = open(devnode, O_RDWR);
   if (fd < 0)
   {
      CX88ERR << "Unable to open device " << devnode << ": " << strerror(errno) << endl;
      return CX88_CARD_ID_UNKNOWN;  
   }
   uint32_t id;
   int error = ioctl(fd, CX88_COMMON_CARD_ID, &id);
   close(fd);
   if (error)
   {
      CX88ERR << "Unable to issue CARD_ID ioctl: " << strerror(errno) << endl;
      return CX88_CARD_ID_UNKNOWN;
   }
   switch (id)
   {
      case 0x70633000:
         return CX88_CARD_ID_PCHDTV_HD3000;
      case 0x70635500:
         return CX88_CARD_ID_PCHDTV_HD5500;
      case 0x00709000:
      case 0x00709001:
      case 0x00709002:
         return CX88_CARD_ID_WINTV_NOVA_T;
      /* DViCO FusionHDTV 5 Gold, PCI SubVen ID: 0x18ac, PCI SubDev ID: 0xd500 */
      case 0x18acd500:
         return CX88_CARD_ID_FUSION_5_GOLD;
      /* Club3D Zap TV2100, PCI SubVen ID: 0x12ab, PCI SubDev ID: 0x2300             */
      /* Conexant Reference DVB-T card, PCI SubVen ID: 0x14f1, PCI SubDev ID: 0x0187 */
      /* Kworld VStream Xpert DVB-T, PCI SubVen ID: 0x17de, PCI SubDev ID: 0x08a1    */
      /* Terratec Cinergy 1400 DVB-T, PCI SubVen ID: 0x153b, PCI SubDev ID: 0x1166   */
      /* Leadtek WinFast DTV1000T, PCI SubVen ID: 0x107d, PCI SubDev ID: 0x665f      */
      case 0x12ab2300:
         return CX88_CARD_ID_CLUB3D_ZAP_TV2100;
      case 0x14f10187:
         return CX88_CARD_ID_REF_DVBT;
      case 0x17de08a1:
         return CX88_CARD_ID_KWORLD_VSTREAM_XPERT;
      case 0x153b1166:
         return CX88_CARD_ID_TERRATEC_CINERGY_1400;
      case 0x107d665f:
         return CX88_CARD_ID_LEADTEK_WINFAST_DTV1000T;
      /* Hauppauge WinTV-HVR-1100 (DVB-T, analog)                         */
      /* PCI SubVen ID: 0x0070, PCI SubDev ID: 0x9400 or 0x9401 or 0x9402 */
      /* Hauppauge WinTV-HVR-1100 LP (DVB-T, analog)            */
      /* PCI SubVen ID: 0x0070, PCI SubDev ID: 0x9800 or 0x9802 */
      /* Leadtek WinFast DTV2000H                     */
      /* PCI SubVen ID: 0x107d, PCI SubDev ID: 0x665e */
      case 0x00709400:
      case 0x00709401:
      case 0x00709402:
      case 0x00709800:
      case 0x00709802:
         return CX88_CARD_ID_WINTV_HVR1100;
      case 0x107d665e:
         return CX88_CARD_ID_LEADTEK_WINFAST_DTV2000H;
      /* Hauppauge WinTV-HVR-1300 (DVB-T, analog)                         */
      /* PCI SubVen ID: 0x0070, PCI SubDev ID: 0x9600 or 0x9601 or 0x9602 */
      case 0x00709600:
      case 0x00709601:
      case 0x00709602:
         return CX88_CARD_ID_WINTV_HVR1300;
      /* Hauppauge WinTV-HVR-3000 (DVB-T, DVB-S, analog)                            */
      /* PCI SubVen ID: 0x0070, PCI SubDev ID: 0x1400 or 0x1401 or 0x1402 or 0x1404 */
      case 0x00701400:
      case 0x00701401:
      case 0x00701402:
      case 0x00701404:
         return CX88_CARD_ID_WINTV_HVR3000;
      /* Hauppauge WinTV-HVR-4000 (DVB-T, DVB-S, DVB-S2, analog)                    */
      /* PCI SubVen ID: 0x0070, PCI SubDev ID: 0x6900 or 0x6901 or 0x6902 or 0x6904 */
      case 0x00706900: 
      case 0x00706901:
      case 0x00706902:
      case 0x00706904:
         return CX88_CARD_ID_WINTV_HVR4000;
      case 0x0:
      {
         const char *card_name = conf.get_string("card");
         if (card_name != NULL)
         {
            if (strcasecmp(card_name, "xcapture") == 0)
            {
               return CX88_CARD_ID_PIXELVIEW_XCAPTURE;    
            }
         }
      }
      default:
         CX88ERR << "Unrecognized card ID: 0x" << hex << id << endl;
         return CX88_CARD_ID_UNKNOWN;
   }

}
