/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <strings.h>
#include <time.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/mman.h>

#include <sstream>
using namespace std;

#include "cx88_mpeg_driver.h"
               
#define PACKET_DELAY_OPTION   "packet_delay"
#define DEFAULT_PACKET_DELAY  100

cx88_mpeg_driver::cx88_mpeg_driver(
   cx88_xml_config &cx88conf,
   tuner_config &tunerconf,
   const char *devnode,
   capvec &captures,
   int &error)
   : cx88_driver(cx88conf, tunerconf, devnode, captures, error)
{}

cx88_mpeg_driver::~cx88_mpeg_driver(void)
{
}

int cx88_mpeg_driver::capture(const string &channel_string, cx88_capture_params &params)
{
   int error = 0;
   cx88_mpeg_config config;
   cx88_common_i2c_config i2config;
   i2config.in_num_i2c_devices = 0;
   if ((m_fd = open(m_devnode, O_RDWR)) < 0)
   {
      CX88ERR << "Unable to open device node " << m_devnode << ": " << strerror(errno) << endl;
      error = errno; 
   }
   if (!error)
   {
      config.in_buftype = CX88_MPEG_BUF_KERNEL;
      config.in_packet_size = CX88_MPEG_DEFAULT_PACKET_SIZE;
      config.inout_num_packets = 10000;
      error = set_config_options(config);
   }
   if (!error && (error = ioctl(m_fd, CX88_MPEG_CONFIG, &config)))
   {
      CX88ERR << "Error issuing CONFIG ioctl: " << strerror(errno) << endl;
   }
   void *buf = MAP_FAILED;
   if (!error)
   {
      buf = mmap(NULL, config.bufconfig.out_kernelbuf.out_mmap_size, PROT_READ, MAP_SHARED, m_fd, 0);
      if (buf == MAP_FAILED)
      {
         CX88ERR << "Unable to map buffer memory from device: " << strerror(errno) << endl;
         error = errno;
      }
   }
   error = (error ? error : start_captures());
   error = (error ? error : i2c_setup(i2config));
   error = (error ? error : capture(buf, i2config, config, params));
   i2c_teardown(i2config);
   stop_captures();
   if (buf != MAP_FAILED)
   {
      munmap(buf, config.bufconfig.out_kernelbuf.out_mmap_size);
   }
   if (m_fd >= 0)
   {
      close(m_fd);
   }
   if (params.completion != NULL)
   {
      params.completion(params.completion_arg, error);  
   }
   return error;
}

int cx88_mpeg_driver::change_channel(cx88_common_i2c_config &config, dvb_interface &ifc, cx88_capture_params &params)
{
   cx88_driver_lock lock(this);
   dvb_channel channel;
   m_channel_changed = false; 
   int error = m_channel.select_dvb_channel(m_channel_name, channel);
   if (error)
   {
      return error;
   }
   error = set_channel_options (channel);  
   if (error)
   {
      return error;
   }
   for (uint8_t i = 0; i < config.in_num_i2c_devices; ++i)
   {
      dvb_driver *driver = dynamic_cast <dvb_driver*> (m_i2c_devs[i].driver);
      if (driver != NULL)
      {   
         if ((error = driver->set_channel(channel, ifc)))
         {
            CX88ERR << "Error setting channel on peripheral device" << endl;
            return error;
         }
      }
   }
   for (uint8_t i = 0; i < config.in_num_i2c_devices; ++i)
   {
      if (m_i2c_devs[i].driver != NULL)
      {   
         if ((error = m_i2c_devs[i].driver->start(params.lock_timeout_ms)))
         {
            CX88ERR << "Error starting peripheral device" << endl;
            return error;
         }
      }
   }
   return error;
}

int cx88_mpeg_driver::capture(void *buffer, cx88_common_i2c_config &i2config, cx88_mpeg_config &config, cx88_capture_params &params)
{
   cx88_mpeg_capture capture;
   capture.in_start_byte = CX88_MPEG_DEFAULT_START_BYTE;
   capture.in_start_byte_count = CX88_MPEG_DEFAULT_START_BYTE_COUNT;
   capture.in_options = 0;
   dvb_interface ifc;
   cx88_mpeg_wait wait;
   wait.in_timeout_s = 10;
   int error = 0;
   time_t start_time;
   uint64_t packet_count;
   uint32_t packet_delay = m_tunerconf.get_number<uint32_t>(PACKET_DELAY_OPTION, DEFAULT_PACKET_DELAY);
   while (!m_stop)
   {
      error = change_channel(i2config, ifc, params);
      if (params.channel_event != NULL)
      {
         params.channel_event(params.channel_arg, error);
      }
      if (error)
      {
         error = wait_for_event(params.duration_s);
         if (error)
         {
            return error;
         }
         continue;
      }
      if (ifc.input_width_bits == DVB_INPUT_SERIAL)
      {
         capture.in_options |= CX88_MPEG_SERIAL_INPUT;
      }
      if (ifc.clock == DVB_IFC_PUNC_CLCK)
      {
         capture.in_options |= CX88_MPEG_PUNCTURED_CLOCK;
      }
      if (ifc.polarity == DVB_IFC_POS_POL)
      {
         capture.in_options |= CX88_MPEG_POS_POLARITY;
      }
      if (ifc.bit_endianness == DVB_IFC_BIT_LE)
      {
         capture.in_options |= CX88_MPEG_LITTLE_ENDIAN;
      }
      if ((error = set_capture_options(capture)))
      {
         return error;
      }
      packet_count = 0;
      start_time = time(NULL);
      if ((error = ioctl(m_fd, CX88_MPEG_CAPTURE, &capture)))
      {
         CX88ERR << "Error issuing CAPTURE ioctl on device: " << strerror(errno) << endl;
         return error;
      }
      uint8_t last_buffer = 0;
      while (!m_stop && !m_channel_changed)
      {
         if ((error = ioctl(m_fd, CX88_MPEG_WAIT, &wait)))
         {
            CX88ERR << "Error issuing WAIT ioctl on device: " << strerror(errno) << endl;
            ioctl(m_fd, CX88_MPEG_STOP);
            return error;
         }
         if ((wait.out_status & CX88_MPEG_BUS_ERROR) == CX88_MPEG_BUS_ERROR)
         {
            CX88ERR << "Fatal error: PCI bus error" << endl;
         }
         if ((wait.out_status & CX88_MPEG_DMA_PROGRAM) == CX88_MPEG_DMA_PROGRAM)
         {
            CX88ERR << "Fatal error: DMA program error" << endl;
         }
         if ((wait.out_status & CX88_MPEG_WAIT_ABORTED) == CX88_MPEG_WAIT_ABORTED)
         {
            CX88ERR << "Fatal error: Wait aborted" << endl;
         }
         if ((wait.out_status & CX88_MPEG_CAP_TIMEOUT) == CX88_MPEG_CAP_TIMEOUT)
         {
            CX88ERR << "Nonfatal error: Capture timed out" << endl;
         }
         if ((wait.out_status & CX88_MPEG_BAD_PACKET) == CX88_MPEG_BAD_PACKET)
         {
            CX88ERR << "Nonfatal error: " << wait.out_num_bad_packets << " bad packet(s)" << endl;
         }
         if ((wait.out_status & CX88_MPEG_HW_OVERFLOW) == CX88_MPEG_HW_OVERFLOW)
         {
            CX88ERR << "Nonfatal error: " << wait.out_num_overflows << " hardware buffer overflow(s)" << endl;
         }
         if ((wait.out_packet_count - packet_count) >= config.inout_num_packets)
         {
            CX88ERR << "Nonfatal error: Software buffer overflow" << endl;
         }
         if (wait.out_status & CX88_MPEG_CAP_FATAL)
         {
            ioctl(m_fd, CX88_MPEG_STOP);
            return EIO;
         }
         if (wait.out_buffer_available == 0 || wait.out_buffer_available > config.out_num_buffers)
         {
            ioctl(m_fd, CX88_MPEG_STOP);
            return EIO;
         }
         while (last_buffer != wait.out_buffer_available)
         {
            ++last_buffer;
            if (last_buffer > config.out_num_buffers)
            {
               last_buffer = 1;
            }
            size_t bufoffset = (last_buffer - 1) * config.bufconfig.out_kernelbuf.out_buffer_size;
            size_t bufsize = config.bufconfig.out_kernelbuf.out_mmap_size - bufoffset;
            void *bufptr = (char *)buffer + bufoffset;
            if (bufsize > (config.out_packets_per_buffer * config.in_packet_size))
            {
               bufsize = config.out_packets_per_buffer * config.in_packet_size;
            }
            if (packet_delay > packet_count)
            {
               size_t byte_delay = (packet_delay - packet_count) * config.in_packet_size;
               if (byte_delay >= bufsize)
               {
                  continue;
               }
               else
               {
                  bufsize -= byte_delay;
                  bufptr = (char *)bufptr + byte_delay;
               }
            }
            submit_buffer(bufptr, bufsize, config.in_packet_size);
         }
         if ((params.duration_s > 0) && ((unsigned int)(time(NULL) - start_time) >= params.duration_s))
         {
            ioctl(m_fd, CX88_MPEG_STOP);
            return error;  
         }
         packet_count = wait.out_packet_count;
      }
      ioctl(m_fd, CX88_MPEG_STOP);
   }
   return error;
}

int cx88_mpeg_driver::get_signal(dvb_signal &signal)
{
   int error = 0;
   signal.locked = false;
   for (uint8_t i = 0; i < CX88_COMMON_MAX_I2C_DEVICES; ++i)
   {
      dvb_driver *driver = dynamic_cast<dvb_driver *>(m_i2c_devs[i].driver);
      if (driver != NULL)
      {   
         if ((error = driver->get_signal(signal)))
         {
            return error;
         }
      }
   }
   return error;
}

double cx88_mpeg_driver::get_signal_strength(void)
{
   cx88_driver_lock lock(this);
   double strength = 0.0;  
   dvb_signal signal;
   get_signal(signal);
   if (signal.locked)
   {
      if (signal.strength == 0.0)
      {
         strength = 100.0;
      }
      else
      {
         strength = signal.strength;
      }  
   }
   return strength;
}

int cx88_mpeg_driver::get_info(string &info)
{
   cx88_driver_lock lock(this);
   int error = 0;
   try
   {
      dvb_signal signal;
      ostringstream stream;
      string channel;
      m_channel.get_channel_name(channel);
      stream << "Tuned to channel " << channel << endl;
      if ((error = get_signal(signal)))
      {
         return error;
      }
      if (!(signal.locked))
      {
         stream << "\tNo signal" << endl;
      }
      else
      {
         if (signal.strength > 0.0)
         {
            stream << "\tSignal strength: " << signal.strength << "%" << endl;
         }
         if (signal.snr > 0.0)
         {
            stream << "\tSignal-to-noise ratio: " << signal.snr << " dB" << endl;
         }
         if (signal.ber > 0)
         {
            stream << "\tBit-error rate: " << signal.ber << endl;
         }
         if (signal.uncorrected_blocks > 0)
         {
            stream << "\tUncorrected blocks: " << signal.uncorrected_blocks << endl;
         }
      }
      info = stream.str();
   }
   catch(...)
   {
      CX88ERR << "Insufficient memory to generate info string" << endl;
      error = ENOMEM;
   }
   return 0;
}

