/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 
#include <sys/types.h>
#include <sys/param.h>
#include <sys/bus.h>
#include <sys/systm.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/proc.h>
#include "cx88_regs.h"
#include "cx88_i2c.h"
#include "cx88_i2c_if.h"

/* Register constants */
#define I2C_REG_DATA_CONTROL       0x368000
#define I2C_DATA_CONTROL_SLAVEADDR 0x7F,25
#define I2C_DATA_CONTROL_READ      (1 << 24)
#define I2C_DATA_CONTROL_DATA0     0xFF,24
#define I2C_DATA_CONTROL_DATA1     0xFF,16
#define I2C_DATA_CONTROL_DATA2     0xFF,8
#define I2C_DATA_CONTROL_HWMODE    (1 << 7)
#define I2C_DATA_CONTROL_400KHZ    (1 << 6)
#define I2C_DATA_CONTROL_NOSTOP    (1 << 5)
#define I2C_DATA_CONTROL_NOSTART   (1 << 4)
#define I2C_DATA_CONTROL_SYNC      (1 << 3)
#define I2C_DATA_CONTROL_3BYTEWR   (1 << 2)
#define I2C_DATA_CONTROL_ACKRD     (1 << 2)
#define I2C_DATA_CONTROL_SCL       (1 << 1)
#define I2C_DATA_CONTROL_SDA       (1 << 0)

/* Default interrupt timeout--5s */
#define DEFAULT_TIMEOUT (5 * hz)

static void *cx88_i2c_get_sibling_data(device_t dev);
static int cx88_i2c_init_board_data(void *arg);
static int cx88_i2c_destroy_board_data(void *arg);
static int cx88_i2c_filter(void *arg);
static void cx88_i2c_set_intr(cx88_i2c_bus *bus, boolean_t enable);
static void cx88_i2c_terminate(cx88_i2c_bus *bus);
static int cx88_i2c_commit(cx88_i2c_bus *bus, uint32_t reg, int timeout);
static int cx88_i2c_transfer(cx88_i2c_bus *bus, cx88_i2c_transaction *transaction);

/* Accessor for sibling i2c shared data */
static void *cx88_i2c_get_sibling_data(device_t dev)
{
   cx88_i2c_bus *bus = CX88_I2C_GET_BUS(dev);
   if (bus != NULL)
   {
      return bus->board_data;
   }
   return NULL;
}

/* Initializer to be called when i2c shared data is allocated */
static int cx88_i2c_init_board_data(void *arg)
{
   DIAGNOSTIC(printf("CX88-I2C: Initializing board data\n"))
   cx88_i2c_board_data *data = arg;
   mtx_init(&(data->lock), "CX88 I2C board lock", NULL, MTX_DEF);
   data->bus_owned = FALSE;
   mtx_init(&(data->isr_lock), "CX88 I2C interrupt lock", NULL, MTX_SPIN);
   return 0;
}

/* Cleanup routine to be called when i2c shared data is about to be freed */
static int cx88_i2c_destroy_board_data(void *arg)
{
   DIAGNOSTIC(printf("CX88-I2C: Destroying board data\n"))
   cx88_i2c_board_data *data = arg;
   mtx_destroy(&(data->lock));
   mtx_destroy(&(data->isr_lock));
   return 0;
}

/* Create an i2c bus instance */
int cx88_i2c_create(cx88_i2c_bus *bus, cx88_common_softc *host_sc)
{
   bus->host_sc = host_sc;
   bus->dev = NULL;
   bus->board_data = NULL;
   bus->isr_error = 0;
   int error = 0;
   /* Allocate shared data structure or retrieve it from sibling driver */
   if ((error = cx88_add_board_data(host_sc, (void**)(&(bus->board_data)),
      sizeof(*(bus->board_data)), cx88_i2c_get_sibling_data, cx88_i2c_init_board_data)))
   {
      return error;
   }
   /* This is less than ideal, but much of the code that modifies the device tree
      appears to still require Giant (and doesn't acquire it internally).  Since
      cx88 is designed to operate without Giant, and i2c setup routines shouldn't 
      be forced to operate in contexts where Giant is known to be held (e.g. attach,
      detach), we acquire it here to be safe*/
   mtx_lock(&Giant);
   bus->dev = device_add_child(host_sc->dev, "iicbus", -1);
   if (bus->dev == NULL)
   {
      cx88_i2c_destroy(bus);
      mtx_unlock(&Giant);
      return ENXIO;
   }
   if ((error = device_probe_and_attach(bus->dev)))
   {
      cx88_i2c_destroy(bus);
      mtx_unlock(&Giant);
      return error;
   }
   /* Register i2c interrupt service routine */
   if ((error = bus_setup_intr(host_sc->dev, host_sc->irq_res, INTR_TYPE_AV | INTR_MPSAFE,
      cx88_i2c_filter, NULL, bus, &(bus->irq_cookie))))
   {
      cx88_i2c_destroy(bus);
   }
   mtx_unlock(&Giant);
   return error;
}

/* Destroy an i2c bus instance */
int cx88_i2c_destroy(cx88_i2c_bus *bus)
{
   int error = 0;
   mtx_lock(&Giant);
   if (bus->irq_cookie != NULL)
   {
      error = bus_teardown_intr(bus->host_sc->dev, bus->host_sc->irq_res, bus->irq_cookie);
      if (!error)
      {
         bus->irq_cookie = NULL;
      }
   }
   if (!error && (bus->dev != NULL))
   {
      error = device_detach(bus->dev);
      if (!error)
      {
         /* This is kind of ugly.  Newbus currently provides
          * no way for bus drivers to clean up instance 
          * variables for their children, so I have to do
          * it here.  This is NOT a permanent solution. */
         device_t *children = NULL;
         int num_children = 0, i;
         error = device_get_children(bus->dev, &children,  &num_children);
         if (!error)
         {
            for (i = 0; i < num_children; ++i)
            {
               DIAGNOSTIC(device_printf(bus->host_sc->dev, "Freeing iicbus ivar\n"))
               free(IICBUS_IVAR(children[i]), M_DEVBUF);
            }
            free(children, M_TEMP);
            error = device_delete_child(bus->host_sc->dev, bus->dev);
            if (!error)
            {
               bus->dev = NULL;
            }
         }
      }
   }
   mtx_unlock(&Giant);
   if (!error && (bus->board_data != NULL))
   {
      /* Release reference to shared data structure */
      error = cx88_remove_board_data(bus->host_sc, (void **)(&(bus->board_data)), cx88_i2c_destroy_board_data);
      if (!error)
      {
         bus->board_data = NULL;
      }
   }
   return error;
}

/* I2C interrupt filter */
static int cx88_i2c_filter(void *arg)
{
   cx88_i2c_bus *bus = arg;
   uint32_t reg;
   /* Ensure the interrupt is actually ours */
   if ((bus->transaction_state != CX88_I2C_INACTIVE) && 
      ((reg = cx88_read_reg(bus->host_sc, CX88_REG_PCI_INTSTAT)) & CX88_PCI_INT_I2C_DONE))
   {
      /* Acknowledge the interrupt */
      cx88_write_reg(bus->host_sc, CX88_REG_PCI_INTSTAT, (CX88_PCI_INT_I2C_DONE | CX88_PCI_INT_I2C_ACK));
      bus->isr_error = IIC_NOERR;
      if (!(reg & CX88_PCI_INT_I2C_ACK))
      {
         bus->isr_error = IIC_ENOACK;
      }
      mtx_lock_spin(&(bus->board_data->isr_lock));
      wakeup(bus);
      mtx_unlock_spin(&(bus->board_data->isr_lock));
      return FILTER_HANDLED;
   }
   return FILTER_STRAY;
}

/* Add a device to a bus instance */
int cx88_i2c_add_device(
   cx88_i2c_bus *bus,
   uint8_t addr,
   const char *name,
   cx88_i2c_dev *dev)
{
   int error = 0;
   if (addr & (1 << 7))
   {
      return EINVAL;
   }
   if (bus->dev == NULL)
   {
      return ENXIO;
   }
   cx88_i2c_init_device(dev);
   dev->bus = bus;
   if (name != NULL)
   {
      /* Again less than ideal, but we need to hold Giant while touching the device tree */
      mtx_lock(&Giant);
      dev->dev = BUS_ADD_CHILD(bus->dev, 0, name, -1);
      if (dev->dev == NULL)
      {
         mtx_unlock(&Giant);
         return ENXIO;
      }
      struct iicbus_ivar *devaddr = IICBUS_IVAR(dev->dev);
      if (devaddr == NULL)
      {
         mtx_unlock(&Giant);
         return ENXIO;
      }
      dev->name = device_get_nameunit(dev->dev);
      devaddr->addr = addr;
      error = device_probe_and_attach(dev->dev);
      mtx_unlock(&Giant);
   }
   if (!error)
   {
      /* Now initialize device state by checking siblings to see if they are
       * using the device address */
      int i;
      cx88_lock_board(bus->host_sc);
      dev->addr = addr;
      for (i = 0; i < bus->host_sc->num_siblings; ++i)
      {
         cx88_i2c_dev *i2cdev = CX88_I2C_GET_DEV(bus->host_sc->siblings[i], addr);
         if (i2cdev != NULL)
         {
            if (i2cdev->state == CX88_I2CDEV_OWNED)
            {
               dev->state = CX88_I2CDEV_REQUESTED;
               break;
            }
         }
      }
      cx88_unlock_board(bus->host_sc);
   }
   return error;
}

/* Remove device from a bus */
int cx88_i2c_remove_device(cx88_i2c_dev *dev)
{
   int error = 0;
   /* Release the device (if it is owned) */
   cx88_i2c_release(dev);
   if (dev->dev != NULL)
   {
      struct iicbus_ivar *addr = IICBUS_IVAR(dev->dev);
      mtx_lock(&Giant);
      /* This kind of sucks--there's no BUS_REMOVE_CHILD method to free ivars for us */
      error = device_detach(dev->dev);
      if (!error)
      {
         error = device_delete_child(dev->bus->dev, dev->dev);
      }
      mtx_unlock(&Giant);
      if (!error)
      {
         free(addr, M_DEVBUF);
         /* Reset device state to default values */
         cx88_i2c_init_device(dev);
      }
   }
   return error;
}

/* Request ownership of device address */
int cx88_i2c_request(cx88_i2c_dev *dev)
{
   int i, error = 0;
   cx88_i2c_bus *bus = dev->bus;
   if (dev->addr == CX88_I2C_INVALID_ADDR)
   {
      return EINVAL;
   }
   if (dev->state == CX88_I2CDEV_OWNED)
   {
      return 0;
   }
   /* Prevent siblings from messing with bus or detaching while we are requesting device */
   cx88_lock_board(bus->host_sc);
   if (dev->state == CX88_I2CDEV_REQUESTED)
   {
      /* Someone else already owns the device, so return EBUSY */
      cx88_unlock_board(bus->host_sc);
      return EBUSY;
   }
   for (i = 0; i < bus->host_sc->num_siblings; ++i)
   {
      /* Retrieve any sibling instances of this device and mark them as REQUESTED */
      cx88_i2c_dev *i2cdev = CX88_I2C_GET_DEV(bus->host_sc->siblings[i], dev->addr);
      if (i2cdev != NULL)
      {
         i2cdev->state = CX88_I2CDEV_REQUESTED;
      }
   }
   dev->state = CX88_I2CDEV_OWNED;
   cx88_unlock_board(bus->host_sc);
   return error;
}

/* Release ownership of device address */
void cx88_i2c_release(cx88_i2c_dev *dev)
{
   int i;
   cx88_i2c_bus *bus = dev->bus;
   if (dev->state != CX88_I2CDEV_OWNED)
   {
      return;
   }
   cx88_lock_board(bus->host_sc);
   for (i = 0; i < bus->host_sc->num_siblings; ++i)
   {
      /* Retrieve any sibling instances of this device and mark them as AVAILABLE */
      cx88_i2c_dev *i2cdev = CX88_I2C_GET_DEV(bus->host_sc->siblings[i], dev->addr);
      if (i2cdev != NULL)
      {
         i2cdev->state = CX88_I2CDEV_AVAIL;
      }
   }
   dev->state = CX88_I2CDEV_AVAIL;
   cx88_unlock_board(bus->host_sc);
}

/* iicbus bus ownership callback--child devices should use this around
 * transactions to prevent other drivers from placing the bus in 
 * an invalid state */
int cx88_i2c_callback(device_t dev, int index, caddr_t data)
{
   cx88_i2c_bus *bus = CX88_I2C_GET_BUS(dev);
   int how = IIC_INTR;
   int error = 0;
   int priority = PZERO;
   if (bus == NULL)
   {
      return ENXIO;  
   }
   switch (index)
   {
      case IIC_REQUEST_BUS:
         DIAGNOSTIC(device_printf(dev, "Requesting I2C bus\n"))
         if (data != NULL)
         {
            how = *data;
         }
         /* Atomically wait for the bus to become available and update the bus_owned flag */
         mtx_lock(&(bus->board_data->lock));
         if (bus->board_data->bus_owned)
         {
            if (how & IIC_WAIT)
            {
               if (how & IIC_INTR)
               {
                  /* Interruptible wait--catch signals */
                  priority |= PCATCH;
               }
               error = msleep(&(bus->board_data->bus_owned), &(bus->board_data->lock), priority, "cx88ic", 0);
            }
            else
            {
               error = EWOULDBLOCK;
            }
         }
         if (!error)
         {
            bus->board_data->bus_owned = TRUE;
         }
         mtx_unlock(&(bus->board_data->lock));
         break;
      case IIC_RELEASE_BUS:
         DIAGNOSTIC(device_printf(dev, "Releasing I2C bus\n"))
         mtx_lock(&(bus->board_data->lock));
         bus->board_data->bus_owned = FALSE;
         /* Only one waiter will get to acquire the bus */
         wakeup_one(&(bus->board_data->bus_owned));
         mtx_unlock(&(bus->board_data->lock));
         break;
      default:
         return EINVAL;
   }
   return error;
}

/* Helper function for setting the interrupt mask.  The enable
 * parameter indicates whether interrupts should be enabled or disabled. */
static void cx88_i2c_set_intr(cx88_i2c_bus *bus, boolean_t enable)
{
   /* Protect access to the shared PCI interrupt register. */
   cx88_lock_board(bus->host_sc);
   /* Acknowledge any outstanding interrupts */
   cx88_write_reg(bus->host_sc, CX88_REG_PCI_INTSTAT, (CX88_PCI_INT_I2C_DONE | CX88_PCI_INT_I2C_ACK));
   /* Enable/disable I2C interrupt */
   cx88_write_reg(
      bus->host_sc,
      CX88_REG_PCI_INTMASK,
      (enable ?
         (cx88_read_reg(bus->host_sc, CX88_REG_PCI_INTMASK) | CX88_PCI_INT_I2C_DONE) :
         (cx88_read_reg(bus->host_sc, CX88_REG_PCI_INTMASK) & ~CX88_PCI_INT_I2C_DONE)));
   cx88_unlock_board(bus->host_sc);
}

/* Helper function for terminating an i2c transfer */
static void cx88_i2c_terminate(cx88_i2c_bus *bus)
{
   if (bus->transaction_state != CX88_I2C_INACTIVE)
   {
      cx88_i2c_set_intr(bus, FALSE);
      bus->transaction_state = CX88_I2C_INACTIVE;
   }
}

/* Write a byte or sequence of bytes to the data/control register and
 * wait for the interrupt.  The reg parameter should contain the data
 * byte(s), R/W bit, and any required NOSTART/NOSTOP/ACK/NACK bits. */
static int cx88_i2c_commit(cx88_i2c_bus *bus, uint32_t reg, int timeout)
{
   int error = 0;
   mtx_lock_spin(&(bus->board_data->isr_lock));
   /* Add required HW I2C flags to the register */
   uint32_t i2c_control = reg | I2C_DATA_CONTROL_HWMODE | I2C_DATA_CONTROL_SYNC | I2C_DATA_CONTROL_SCL | I2C_DATA_CONTROL_SDA;
   cx88_write_reg(bus->host_sc, I2C_REG_DATA_CONTROL, i2c_control);
   error = msleep_spin(bus, &(bus->board_data->isr_lock), "cx88ic",
      (timeout ? timeout : DEFAULT_TIMEOUT));
   mtx_unlock_spin(&(bus->board_data->isr_lock));
   if (error)
   {
      bus->isr_error = IIC_ETIMEOUT;
   }
   if (bus->isr_error)
   {
      printf("ISR ERROR\n");
      /* There was an error handling the interrupt, so bail out. */
      cx88_i2c_terminate (bus);
      /* Kind of a hack, but make sure the bus is quiesced by sending out bogus data
       * with NOSTOP and ACK turned off */
      cx88_write_reg(bus->host_sc, I2C_REG_DATA_CONTROL, 
         (i2c_control | I2C_DATA_CONTROL_NOSTART) & (~(I2C_DATA_CONTROL_NOSTOP | I2C_DATA_CONTROL_ACKRD)));
      return bus->isr_error;
   }
   return error;
}

/* Conducts an I2C transaction. We cache writes and issue them with the next transaction
 * in order to properly handle the STOP condition (reads do not require this because
 * of the 'last' flag), so this function first commits any cached writes and then,
 * if transaction is non-NULL, commits transaction (if it is a read) or caches it. */
static int cx88_i2c_transfer(cx88_i2c_bus *bus, cx88_i2c_transaction *transaction)
{
   int error = IIC_NOERR;
   uint32_t reg = 0;
   /* Flush any pending write */
   if ((bus->transaction_state > CX88_I2C_ACTIVE) && bus->pending_write)
   {
      if (bus->transaction_state == CX88_I2C_START)
      {
         /* This is the first phase of the transaction, so transmit the
          * slave address */
         reg = cx88_write_field(I2C_DATA_CONTROL_SLAVEADDR, bus->slave_addr);
         reg |= cx88_write_field(I2C_DATA_CONTROL_DATA1, bus->last_write.byte);
      }
      else
      {
         reg = cx88_write_field(I2C_DATA_CONTROL_DATA0, bus->last_write.byte);
      }
      if ((transaction != NULL) || (bus->transaction_state == CX88_I2C_RESTART))
      {
         /* If transaction is non-NULL, we have another transaction after this one,
          * so we shouldn't generate the STOP condition.  Transaction will be NULL
          * if we are flushing transactions during a restart operation, but RESTART
          * also indicates there will be subsequent transactions */
         reg |= I2C_DATA_CONTROL_NOSTOP;
      }
      if (bus->transaction_state != CX88_I2C_START)
      {
         reg |= I2C_DATA_CONTROL_NOSTART;
      }
      error = cx88_i2c_commit(bus, reg, bus->last_write.timeout);
   }
   bus->pending_write = FALSE;
   /* Commit or cache the present transaction */
   if (!error && (transaction != NULL))
   {
      if (transaction->type == CX88_I2C_READ)
      {
         /* Commit the transaction if it is a read */
         reg = cx88_write_field(I2C_DATA_CONTROL_SLAVEADDR, bus->slave_addr);
         reg |= I2C_DATA_CONTROL_READ;
         if (!(transaction->last))
         {
            reg |= I2C_DATA_CONTROL_ACKRD;
            reg |= I2C_DATA_CONTROL_NOSTOP;
         }
         if (bus->transaction_state == CX88_I2C_ACTIVE)
         {
            /* If this read is the first transaction after a start, move
             * the state up to START since reads don't go through an initial
             * caching step */
            bus->transaction_state = CX88_I2C_START;
         }
         if (bus->transaction_state != CX88_I2C_START)
         {
            reg |= I2C_DATA_CONTROL_NOSTART;
         }
         error = cx88_i2c_commit(bus, reg, transaction->timeout);
         transaction->byte = cx88_read_field(
            cx88_read_reg(bus->host_sc, I2C_REG_DATA_CONTROL),
            I2C_DATA_CONTROL_DATA2);
      }
      else
      {
         /* If this is a write, just cache the transaction */
         bus->last_write = *transaction;
         bus->pending_write = TRUE;
      }
   }
   /* Now transition the state */
   switch (bus->transaction_state)
   {
      case CX88_I2C_ACTIVE:
         bus->transaction_state = CX88_I2C_START;
         break;
      case CX88_I2C_START:
         bus->transaction_state = CX88_I2C_STARTED;
         break;
      case CX88_I2C_RESTART:
         bus->transaction_state = CX88_I2C_ACTIVE;
         break;
      default:
         break;
   }
   return error;
}

/* iicbus start callback */
int cx88_i2c_start(device_t dev, u_char slave, int timeout)
{
   DIAGNOSTIC(device_printf(dev, "Starting I2C transaction for device 0x%x\n", (slave >> 1)))
   cx88_i2c_bus *bus = CX88_I2C_GET_BUS(dev);
   cx88_i2c_dev *i2c_dev = CX88_I2C_GET_DEV(dev, (slave >> 1));
   /* Ensure we own the device */
   if ((i2c_dev == NULL) || (i2c_dev->state != CX88_I2CDEV_OWNED)
      || (bus->transaction_state != CX88_I2C_INACTIVE))
   {
      return IIC_ESTATUS;
   }
   bus->transaction_state = CX88_I2C_ACTIVE;
   /* This is annoying--iicbus takes slave addresses as 8-bit values,
    * including the R/W bit--why? */
   bus->slave_addr = slave >> 1;
   cx88_i2c_set_intr(bus, TRUE); /* Enable i2c interrupt */
   return IIC_NOERR;
}

/* iicbus stop callback */
int cx88_i2c_stop(device_t dev)
{
   DIAGNOSTIC(device_printf(dev, "Stopping I2C transaction\n"))
   cx88_i2c_bus *bus = CX88_I2C_GET_BUS(dev);
   int error = IIC_NOERR;
   if ((bus == NULL) || (bus->transaction_state == CX88_I2C_INACTIVE))
   {
      return IIC_ESTATUS;
   }
   /* Flush any pending write */
   error = cx88_i2c_transfer(bus, NULL);
   cx88_i2c_terminate(bus);
   return error;
}

/* iicbus restart callback */
int cx88_i2c_restart(device_t dev, u_char slave, int timeout)
{
   DIAGNOSTIC(device_printf(dev, "Restarting I2C transaction\n"))
   cx88_i2c_bus *bus = CX88_I2C_GET_BUS(dev);
   int error = IIC_NOERR;
   if ((bus == NULL) || (bus->transaction_state == CX88_I2C_INACTIVE))
   {
      return IIC_ESTATUS;
   }
   bus->transaction_state = CX88_I2C_RESTART;
   /* Flush any pending write */
   error = cx88_i2c_transfer(bus, NULL);
   bus->slave_addr = slave >> 1;
   return error;
}

/* iicbus reset callback */
int cx88_i2c_reset(device_t dev, u_char speed, u_char addr, u_char * oldaddr)
{
   cx88_i2c_stop(dev);
   if (oldaddr != NULL)
   {
      *oldaddr = 0;
   }
   return IIC_NOERR;
}

/* iicbus write callback */
int cx88_i2c_write (
   device_t dev,
   const char *buf,
   int len,
   int *bytes,
   int timeout)
{
   DIAGNOSTIC(device_printf(dev, "I2C write of %d bytes\n", len))
   cx88_i2c_bus *bus = CX88_I2C_GET_BUS(dev);
   int i, error = IIC_NOERR;
   cx88_i2c_transaction transaction;
   if ((bus == NULL) || (bus->transaction_state == CX88_I2C_INACTIVE))
   {
      return IIC_ESTATUS;
   }
   transaction.type = CX88_I2C_WRITE;
   transaction.timeout = timeout;
   for (i = 0; i < len; ++i)
   {
      transaction.byte = buf[i];
      error = cx88_i2c_transfer(bus, &transaction);
      if (error)
      {
         break;
      }
      ++(*bytes); /* Increment byte counter */
   }
   return error;
}

/* iicbus read callback */
int cx88_i2c_read (
   device_t dev,
   char *buf,
   int len,
   int *bytes,
   int last,
   int delay)
{
   DIAGNOSTIC(device_printf(dev, "I2C read of %d bytes\n", len))
   cx88_i2c_bus *bus = CX88_I2C_GET_BUS(dev);
   int i, error = IIC_NOERR;
   cx88_i2c_transaction transaction;
   if ((bus == NULL) || (bus->transaction_state == CX88_I2C_INACTIVE))
   {
      return IIC_ESTATUS;
   }
   transaction.type = CX88_I2C_READ;
   transaction.timeout = delay;
   for (i = 0; i < len; ++i)
   {
      transaction.last = ((last == IIC_LAST_READ) && (i == (len - 1)));
      error = cx88_i2c_transfer(bus, &transaction);
      if (error)
      {
         break;
      }
      buf[i] = transaction.byte;
      ++(*bytes); /* Increment byte counter */
   }
   return error;
}

void cx88_i2c_init_device(cx88_i2c_dev *dev)
{
   dev->dev = NULL;
   dev->addr = CX88_I2C_INVALID_ADDR;
   dev->state = CX88_I2CDEV_AVAIL;
   dev->name = NULL;
}

/* Creates an i2c bus instance and initializes devices in the i2c softc */
int cx88_i2c_attach(cx88_i2c_softc *sc, cx88_common_softc *parent)
{
   uint8_t i;
   sc->num_i2c_devices = 0;
   for (i = 0; i < CX88_COMMON_MAX_I2C_DEVICES; ++i)
   {
      /* Initialize I2C device array to default (unconfigured) settings */
      cx88_i2c_init_device(&(sc->i2c_devices[i]));
   }
   return cx88_i2c_create(&(sc->bus), parent);
}

int cx88_i2c_detach(cx88_i2c_softc *sc)
{
   int error = 0;
   uint8_t i;
   for (i = 0; i < sc->num_i2c_devices; ++i)
   {
      /* Remove I2C device instances */
      if ((error = cx88_i2c_remove_device(&(sc->i2c_devices[i]))))
      {
         break;
      }  
   }
   if (!error)
   {
      error = cx88_i2c_destroy(&(sc->bus));
   }
   return error;
}

/* Adds/removes/modifies i2c device entries as specified by the addresses in conf */
int cx88_i2c_update_devices(cx88_i2c_softc *sc, cx88_common_i2c_config *conf, const char *driver)
{
   DIAGNOSTIC(device_printf(sc->bus->host_sc->dev, "Configuring i2c devices\n")); 
   int error = 0;
   uint8_t i;
   if (conf->in_num_i2c_devices >= CX88_COMMON_MAX_I2C_DEVICES)
   {
      return EINVAL;
   }
   /* If the new list of devices is larger than the previous one, 
    * remove the 'extra' devices */
   for (i = conf->in_num_i2c_devices; i < sc->num_i2c_devices; ++i)
   {
      if ((error = cx88_i2c_remove_device(&(sc->i2c_devices[i]))))
      {
         return error;
      }
   }
   for (i = 0; i < conf->in_num_i2c_devices; ++i)
   {  
      if (conf->in_i2c_addrs[i] != sc->i2c_devices[i].addr)
      {
         /* Only reconfigure the device if the address is different */
         error = cx88_i2c_remove_device(&(sc->i2c_devices[i]));
         if (!error)
         {
            error = cx88_i2c_add_device(
               &(sc->bus),
               conf->in_i2c_addrs[i],
               driver,
               &(sc->i2c_devices[i]));
         }
         if (error)
         {
            return error;
         }
      }
      if (sc->i2c_devices[i].name != NULL)
      {
         /* Return the devnode names to the application */
         if (strlcpy(conf->out_i2c_devnodes[i], sc->i2c_devices[i].name, 
             CX88_COMMON_MAX_DEVNODE_NAME) > CX88_COMMON_MAX_DEVNODE_NAME)
         {
            return ENOMEM;
         }
      }
   }
   sc->num_i2c_devices = conf->in_num_i2c_devices;
   return error;
}

/* Requests ownership of i2c devices */
int cx88_i2c_request_devices(cx88_i2c_softc *sc)
{
   int error = 0;
   uint8_t i;
   for (i = 0; i < sc->num_i2c_devices; ++i)
   {
      error = cx88_i2c_request(&(sc->i2c_devices[i]));
      if (error)
      {
         /* If we can't own all of the devices, then we shouldn't own any */
         cx88_i2c_release_devices(sc);
         return error;
      }
   }
   return error;   
}

/* Releases ownership of i2c devices */
void cx88_i2c_release_devices(cx88_i2c_softc *sc)
{
   uint8_t i;
   for (i = 0; i < sc->num_i2c_devices; ++i)
   {
      cx88_i2c_release(&(sc->i2c_devices[i]));
   }
}

cx88_i2c_dev *cx88_i2c_get_dev(cx88_i2c_softc *sc, uint8_t addr)
{
   uint8_t i;
   for (i = 0; i < sc->num_i2c_devices; ++i)
   {
      /* Determine whether we have a device with the given address */
      if (sc->i2c_devices[i].addr == addr)
      {
         return (&(sc->i2c_devices[i]));
      }  
   }
   return NULL;
}

/* Module declarations */
static int cx88_i2c_mod_handler(module_t mod, int what, void *arg)
{
   return 0;
}

static moduledata_t cx88_i2c_mod_data =
{
   "cx88i2c",
   cx88_i2c_mod_handler,
   NULL
};

DECLARE_MODULE(cx88_i2c, cx88_i2c_mod_data, SI_SUB_DRIVERS, SI_ORDER_ANY);
MODULE_DEPEND(cx88_i2c, cx88, CX88_MINVER, CX88_MODVER, CX88_MAXVER);
MODULE_DEPEND(cx88_i2c, iicbus, IICBUS_MINVER, IICBUS_MODVER, IICBUS_MAXVER);
MODULE_VERSION(cx88_i2c, CX88_I2C_MODVER);
