/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 
/* File: cx88_i2c.h
 * Purpose: Header file for cx88 I2C bus/peripheral control functions
 */

#ifndef __CX88_I2C_H__
#define __CX88_I2C_H__

#include <sys/types.h>
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/proc.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <dev/iicbus/iiconf.h>
#include <dev/iicbus/iicbus.h>
#include "iicbus_if.h"
#include "cx88_common.h"

#define CX88_I2C_MINVER 3
#define CX88_I2C_MODVER 3
#define CX88_I2C_MAXVER 3

#define CX88_I2C_INVALID_ADDR 0xFF

/* I2C data to be shared by sibling drivers on the same board
 * Used in conjunction with the *board_data functions exported by cx88.ko */
typedef struct
{
   struct mtx lock; /* Synchronize access to i2c bus */
   boolean_t bus_owned; /* Indicates whether any client currently has the bus */
   struct mtx isr_lock; /* Synchronize with i2c I/O interrupt */
} cx88_i2c_board_data;

/* Defines a transfer on the I2C bus */
typedef struct
{
   enum
   {
      CX88_I2C_READ,
      CX88_I2C_WRITE
   }                         type;
   uint8_t                   byte;
   boolean_t                 last; /* This is the last transaction in a sequence (useful for reads only) */
   int                       timeout; /* Timeout in HZ */
} cx88_i2c_transaction;

/* Bus structure that should be maintained in client driver's softc */
typedef struct
{
   cx88_common_softc        *host_sc; /* Common portion of client's softc */
   device_t                  dev; /* iicbus device */
   cx88_i2c_board_data      *board_data;
   volatile int              isr_error; /* Interrupt service routine status */
   void                     *irq_cookie;
   enum
   {
      CX88_I2C_INACTIVE, /* Bus is inactive */
      CX88_I2C_ACTIVE, /* Start/restart call has been made */
      CX88_I2C_START, /* Start condition is to be sent with this transfer */
      CX88_I2C_STARTED, /* Start condition has already been sent with previous transfer */
      CX88_I2C_RESTART /* Restart has been issued; flush pending transfer and return to ACTIVE state */
   }                         transaction_state;
   uint8_t                   slave_addr; /* 7-bit slave address for current transaction */
   cx88_i2c_transaction      last_write; /* Cached previous write transaction */
   boolean_t                 pending_write; /* Indicates a write has been cached */
} cx88_i2c_bus;

/* Device structure that should be maintained in client driver's softc */
typedef struct
{
   cx88_i2c_bus             *bus; /* Owning bus */
   device_t                  dev; /* device_t for this device's driver */
   uint8_t                   addr; /* 7-bit I2C address */
   enum
   {
      CX88_I2CDEV_AVAIL = 0, /* Device is not owned by this client or any siblings */
      CX88_I2CDEV_REQUESTED, /* Sibling owns this device */
      CX88_I2CDEV_OWNED /* This client owns the device */
   }                         state;
   const char               *name; /* Name of device node--may be NULL */
} cx88_i2c_dev;

typedef struct
{
   cx88_i2c_bus bus; /* I2C bus instance */
   cx88_i2c_dev i2c_devices[CX88_COMMON_MAX_I2C_DEVICES]; /* I2C device instances */
   uint8_t num_i2c_devices; /* Number of used element in the preceding array */
} cx88_i2c_softc;

/* Client function for creating a bus instance.
 * Returns 0 on success, appropriate error otherwise. */
int cx88_i2c_create(cx88_i2c_bus *bus, cx88_common_softc *host_sc);

/* Client function for destroying a bus instance.  Returns 0 
 * on success, appropriate error otherwise. */
int cx88_i2c_destroy(cx88_i2c_bus *bus);

/* Client function for adding a device to a bus instance.  Returns
 * 0 on success, appropriate error otherwise */
int cx88_i2c_add_device(
   cx88_i2c_bus *bus,
   uint8_t addr, /* 7-bit address */
   const char *name, /* Name of driver to use (NULL -> no driver) */
   cx88_i2c_dev *dev); /* Device structure to be loaded */
   
/* Client function for removing a device from a bus instance.
 * Will release ownership of the device if it is currently owned.
 * Returns 0 on success, appropriate error otherwise */
int cx88_i2c_remove_device(cx88_i2c_dev *dev);

/* Client function for requesting ownership of a device. 
 * Returns 0 on success, appropriate error if another sibling
 * already owns the device */
int cx88_i2c_request(cx88_i2c_dev *dev);

/* Client function for releasing ownership of a device */
void cx88_i2c_release(cx88_i2c_dev *dev);

/* These functions implement the iicbus interface and should not be 
 * called directly by the client driver */
int cx88_i2c_callback(device_t dev, int index, caddr_t data);
int cx88_i2c_start(device_t dev, u_char slave, int timeout);
int cx88_i2c_stop(device_t dev);
int cx88_i2c_restart(device_t dev, u_char slave, int timeout);
int cx88_i2c_reset(device_t dev, u_char speed, u_char addr, u_char * oldaddr);
int cx88_i2c_write (
   device_t dev,
   const char *buf,
   int len,
   int *bytes,
   int timeout);
int cx88_i2c_read (
   device_t dev,
   char *buf,
   int len,
   int *bytes,
   int last,
   int delay);

/* Convenience function for client to initialize/reset device state */
void cx88_i2c_init_device(cx88_i2c_dev *dev);

/* Creates an i2c bus instance and initializes devices in the i2c softc */
int cx88_i2c_attach(cx88_i2c_softc *sc, cx88_common_softc *parent);

/* Cleans up an i2c softc instance */
int cx88_i2c_detach(cx88_i2c_softc *sc);

/* Adds/removes/modifies i2c device entries as specified by the addresses in conf */
int cx88_i2c_update_devices(cx88_i2c_softc *sc, cx88_common_i2c_config *conf, const char *driver);

/* Requests ownership of i2c devices */
int cx88_i2c_request_devices(cx88_i2c_softc *sc);

/* Releases ownership of i2c devices */
void cx88_i2c_release_devices(cx88_i2c_softc *sc);

/* Returns bus instance associated with softc */
static inline cx88_i2c_bus *cx88_i2c_get_bus(cx88_i2c_softc *sc)
{
   return (&(sc->bus));
}
   
/* Finds device at given address, returns NULL if none found */
cx88_i2c_dev *cx88_i2c_get_dev(cx88_i2c_softc *sc, uint8_t addr);

/* DEVMETHODs all I2C client drivers should implement */
#define CX88_I2C_METHODS(bus_accessor, dev_accessor) \
   DEVMETHOD(cx88_i2c_get_bus, bus_accessor), \
   DEVMETHOD(cx88_i2c_get_dev, dev_accessor), \
   DEVMETHOD(iicbus_callback,  cx88_i2c_callback), \
   DEVMETHOD(iicbus_start,     cx88_i2c_start), \
   DEVMETHOD(iicbus_repeated_start, cx88_i2c_restart), \
   DEVMETHOD(iicbus_stop,      cx88_i2c_stop), \
   DEVMETHOD(iicbus_write,     cx88_i2c_write), \
   DEVMETHOD(iicbus_read,      cx88_i2c_read), \
   DEVMETHOD(iicbus_reset,     cx88_i2c_reset)

/* All I2C client drivers should declare dependency on this module */
#define CX88_I2C_DRIVER_MODULE(host) \
   DRIVER_MODULE(iicbus, host, iicbus_driver, iicbus_devclass, NULL, NULL); \
   MODULE_DEPEND(host, iicbus, IICBUS_MINVER, IICBUS_MODVER, IICBUS_MAXVER); \
   MODULE_DEPEND(host, cx88_i2c, CX88_I2C_MINVER, CX88_I2C_MODVER, CX88_I2C_MAXVER);

#endif
