/*-
 * Copyright 2007 Richard Airlie
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/module.h>
#include <sys/mman.h>
#include <sys/selinfo.h>
#include <sys/poll.h>
#include <sys/fcntl.h>
#include <sys/ioccom.h>
#include <sys/uio.h>
#include "cx88_common.h"
#include "cx88_if.h"
#include "cx88_regs.h"
#include "cx88_ir_regs.h"

#define MAX_SAMPLES 16

/* IR softc structure */
typedef struct
{
   cx88_common_softc common; /* All clients of the common layer MUST have this as the first element */
   struct cdev *devnode; /* Device file structure */
   struct mtx isrlock; /* Synchronize with interrupt service routine */
   boolean_t in_use; /* Indicates whether device is active */
   uint32_t samples[MAX_SAMPLES];
   uint8_t sample_count;
   uint8_t keycode;
   boolean_t data_avail;
} cx88_ir_softc;

static int cx88_ir_setup (void *arg);
static int cx88_ir_teardown (void *arg);
static void cx88_ir_intr (void *arg);
static int cx88_ir_attach(device_t dev);
static int cx88_ir_shutdown(device_t dev);
static int cx88_ir_suspend(device_t dev);
static int cx88_ir_stop(cx88_ir_softc *sc);
static int cx88_ir_request_reconfig(device_t dev);
static unsigned int cx88_ir_biphase_decode(uint32_t *samples, unsigned int num_samples);

static devclass_t cx88_ir_devclass;

static d_open_t cx88_ir_open;
static d_close_t cx88_ir_close;
static d_read_t cx88_ir_read;

/* Devnode descriptor */
static struct cdevsw cx88_ir_cdevsw =
{
   .d_version = D_VERSION,
   .d_flags = D_TRACKCLOSE,
   .d_name = "cx88ir",
   .d_open = cx88_ir_open,   /* open */
   .d_close = cx88_ir_close, /* close */
   .d_read = cx88_ir_read  /* read */
};

/* Client driver descriptor for cx88 common layer */
static const cx88_client_driver cx88_ir_client = 
{
   .softc_size               = sizeof(cx88_ir_softc),
   .function                 = CX88_FUNC_GPHOST,
   .setup                    = cx88_ir_setup,
   .teardown                 = cx88_ir_teardown,
   .interrupt_handler        = cx88_ir_intr,
   .pci_intmask              = CX88_PCI_INT_GPIO16_SAMP,
   .description              = "Conexant CX2388x IR"
};

/* Setup callback to be called by cx88 common layer during device attach */
static int cx88_ir_setup (void *arg)
{
   cx88_ir_softc *sc = arg; /* Argument will always be client driver's softc */
   int unit = device_get_unit(sc->common.dev);
   sc->in_use = FALSE;
   sc->keycode = 0;
   sc->data_avail = FALSE;
   mtx_init(&(sc->isrlock), "CX88 IR ISR lock", NULL, MTX_DEF);

   /* Create device file--this should be done as the last step in initialization */
   sc->devnode = make_dev(&cx88_ir_cdevsw, unit, UID_ROOT, GID_WHEEL, 0644, "cx88ir%d", unit);
   if (sc->devnode == NULL)
   {
      mtx_destroy(&(sc->isrlock));
      return ENOMEM;
   }
   return 0;
}

/* Cleanup callback to be called by cx88 common layer during device detach */
static int cx88_ir_teardown (void *arg)
{
   int error = 0;
   cx88_ir_softc *sc = arg; /* Argument will always be client driver's softc */

   mtx_lock(&(sc->isrlock));
   error = cx88_ir_stop(sc);
   mtx_unlock(&(sc->isrlock));

   if (!error && (sc->devnode != NULL))
   {
      /* Device file should be destroyed as the second step in cleanup*/
      destroy_dev(sc->devnode);
      sc->devnode = NULL;
   }
   if (!error)
   {
      mtx_destroy(&(sc->isrlock));
   }
   return error;
}

/* Open file operation */
static int cx88_ir_open(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   int error = 0;
   cx88_ir_softc *sc = devclass_get_softc(cx88_ir_devclass, minor(dev));
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Opening device\n"))
   mtx_lock(&(sc->isrlock));
   if (sc->in_use)
   {
      mtx_unlock(&(sc->isrlock));
      return EBUSY;
   }

   sc->sample_count = 0;

   /* 4khz sample rate */
   cx88_write_reg(&(sc->common), CX88_REG_DDS_IO, 0xa80a80);
   /* enable */
   cx88_write_reg(&(sc->common), CX88_REG_DDSCFG_IO, 0x5);
   /* Enable IR interrupt */
   cx88_lock_board(&(sc->common));
   cx88_write_reg(&(sc->common), CX88_REG_PCI_INTMASK,
       (cx88_read_reg(&(sc->common), CX88_REG_PCI_INTMASK) | CX88_PCI_INT_GPIO16_SAMP));
   cx88_unlock_board(&(sc->common));
   sc->in_use = TRUE;  
   mtx_unlock(&sc->isrlock);
   return error;
}

/* Close file operation */
static int cx88_ir_close(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   cx88_ir_softc *sc = devclass_get_softc(cx88_ir_devclass, minor(dev));
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Closing device\n"))
   mtx_lock(&(sc->isrlock));
   cx88_ir_stop(sc);
   sc->in_use = FALSE;
   mtx_unlock(&(sc->isrlock));
   return 0;
}

static int cx88_ir_stop(cx88_ir_softc *sc)
{
   int error = 0;
   cx88_lock_board(&(sc->common));
   /* Disable IR interrupt */
   cx88_write_reg(&(sc->common), CX88_REG_PCI_INTMASK,
       (cx88_read_reg(&(sc->common), CX88_REG_PCI_INTMASK) & ~CX88_PCI_INT_GPIO16_SAMP));
   cx88_unlock_board(&(sc->common));
   cx88_write_reg(&(sc->common), CX88_REG_DDSCFG_IO, 0x0);
   wakeup(&(sc->data_avail));
   return error;
}

static int cx88_ir_read(struct cdev *dev, struct uio *uio, int flag)
{
	char str[4];
	int ret;
	int len;

	cx88_ir_softc *sc = devclass_get_softc(cx88_ir_devclass, minor(dev));
	if (sc == NULL)
	{
	    return ENXIO;
	}

	mtx_lock(&(sc->isrlock));
	ret = 0;
	while(uio->uio_resid > 0 && ret==0){
	    if(sc->data_avail){
	        if((len = snprintf(str, 4, "%d\n", sc->keycode)) > 0){
	            ret = uiomove((caddr_t)str, len, uio);
	        }
	        sc->data_avail = FALSE;
	        mtx_unlock(&(sc->isrlock));
	        return ret;
	    }else{
	        ret = msleep(&(sc->data_avail), &(sc->isrlock),
					PZERO | PCATCH, "cx88ir", 10*hz);
	    }
	}
	mtx_unlock(&(sc->isrlock));
	return ret;
}

static unsigned int
cx88_ir_biphase_decode(uint32_t *samples, unsigned int num_samples){
	uint32_t mask;
	uint32_t sample;
	uint32_t value;
	uint8_t last;
	uint8_t got;
	int bank = 0;
	int changes = 0, len = 0, max = 7, min = 5;

	/* Find first set bit */
	for(mask = 0x80000000, sample = samples[0];
	    mask > 0 && !(mask&sample);
	    mask >>=1)
	    ;;

	for(value=1, last=1, len=0, min=5, max=7, changes=0;
	    mask > 0 && changes <= 1 && len <= max;
	    last=got){

	    got = (sample & mask) != 0;
	    if(got == last){
	        ++len;
	    }else if(len < min){
	        ++len;
	        ++changes;
	    }else{
	        value = (value << 1)|got;
	        changes = 0;
	        len = 1;
	    }
	    if(mask == 1 && bank < num_samples-1){
	        /* next sample */
	        sample = samples[++bank];
	        mask = 0x80000000;
	    }else{
	        mask >>= 1;
	    }
	}
	return value;
}

/* Interrupt service routine */
static void cx88_ir_intr(void *arg)
{
	cx88_ir_softc *sc = arg;
	uint32_t sample;
	uint32_t rc5;
	int i;
	uint32_t intrs = cx88_read_reg(&(sc->common), CX88_REG_PCI_INTSTAT);

	if(intrs & CX88_PCI_INT_GPIO16_SAMP){
	    mtx_lock(&(sc->isrlock));

	    sample = cx88_read_reg(&(sc->common), CX88_REG_SAMPLE_IO);
	    if(sample != 0 && sample != 0xffffffff){
	        /* more samples to come */
	        if(sc->sample_count < MAX_SAMPLES){
	            sc->samples[sc->sample_count++] = sample;
	        }
	    }else{
	        if(sc->sample_count > 0){
	            /* end of sample sequence, write last one */
	            if(sc->sample_count < MAX_SAMPLES){
	                sc->samples[sc->sample_count++] = sample;
	            }
	            DIAGNOSTIC(printf("Sample seq: "))
	            for(i=0; i<sc->sample_count; ++i){
	                sc->samples[i] = ~sc->samples[i];
	                DIAGNOSTIC(printf("%x,", sc->samples[i]))
	            }
	            DIAGNOSTIC(printf("\n"))

	            rc5 = cx88_ir_biphase_decode(sc->samples, sc->sample_count);
	            DIAGNOSTIC(device_printf(sc->common.dev, "RC5 = %x\n", rc5))
	            if((rc5 & 0xfffff000) == 0x3000){
	              /* start bits present, looks valid, get the 6 bit keycode */
	              sc->keycode = rc5 & 0x3f;
	              sc->data_avail = TRUE;
	            }
	            sc->sample_count = 0;
	        }
	    }

	    /* Acknowledge the interrupt */
	    cx88_write_reg(&(sc->common), CX88_REG_PCI_INTSTAT, intrs);
	    wakeup(&(sc->data_avail));
	    mtx_unlock(&(sc->isrlock));
	}
}


/* Probe DEVMETHOD */
static int cx88_ir_probe(device_t dev) 
{
   return cx88_common_probe(dev, &cx88_ir_client);
}

/* Attach DEVMETHOD */
static int cx88_ir_attach(device_t dev) 
{
   return cx88_common_attach(dev, &cx88_ir_client);
}

/* Shutdown DEVMETHOD */
static int cx88_ir_shutdown(device_t dev)
{
   DIAGNOSTIC(device_printf(dev, "Shutdown\n"))
   return 0;
}

/* Suspend DEVMETHOD */
static int cx88_ir_suspend(device_t dev)
{
   int error = 0;
   DIAGNOSTIC(device_printf(dev, "Suspend\n"))
   cx88_ir_softc *sc = device_get_softc(dev);
   mtx_lock(&(sc->isrlock));
   error = cx88_ir_stop(sc);
   mtx_unlock(&(sc->isrlock));
   return error;
}

static int cx88_ir_request_reconfig(device_t dev)
{
   return ENODEV;
}

/* DEVMETHODs */
static device_method_t cx88_ir_methods[] = 
{
   /* Device interface */
   DEVMETHOD(device_probe,             cx88_ir_probe),
   DEVMETHOD(device_attach,            cx88_ir_attach),
   DEVMETHOD(device_detach,            cx88_common_detach),
   DEVMETHOD(device_shutdown,          cx88_ir_shutdown),
   DEVMETHOD(device_suspend,           cx88_ir_suspend),
   DEVMETHOD(device_resume,            cx88_common_resume),
                
   /* cx88 interface */
   CX88_METHODS(cx88_ir_request_reconfig),
   {NULL, NULL}
};

/* Module declarations */
static driver_t cx88_ir_driver =
{
   "cx88ir",
   cx88_ir_methods,
   0
};

DRIVER_MODULE(cx88ir, pci, cx88_ir_driver, cx88_ir_devclass, NULL, NULL);
DRIVER_MODULE(cx88ir, cardbus, cx88_ir_driver, cx88_ir_devclass, NULL, NULL);
CX88_DRIVER_MODULE(cx88ir);
MODULE_VERSION(cx88ir, 1);
