/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/module.h>
#include <sys/mman.h>
#include <sys/selinfo.h>
#include <sys/poll.h>
#include <sys/fcntl.h>
#include "cx88_common.h"
#include "cx88_if.h"
#include "iicdev.h"
#include "cx88_i2c.h"
#include "cx88_i2c_if.h"
#include "cx88_regs.h"
#include "cx88_mpeg_regs.h"
#include "cx88_mpeg_ioctl.h"

/* Minimum and maximum data buffer sizes (in bytes) */
#define MAX_MPEG_BUFFER_SIZE (1 << 19)
#define MIN_MPEG_BUFFER_SIZE (1 << 15)

/* Maximum line size, in bytes (dictated by capacity of line length register) */
#define MAX_MPEG_LINE_SIZE 0xFFF

/* Driver states */
typedef enum
{
   CX88_MPEG_INVALID = 0, /* Uninitialized */
   CX88_MPEG_UNCONFIGURED, /* Channel/buffers/I2C devices not yet configured */
   CX88_MPEG_CONFIGURED, /* Configured but not capturing */
   CX88_MPEG_CAPTURING /* Configured and capturing */
} cx88_mpeg_state_t;

/* MPEG softc structure */
typedef struct
{
   cx88_common_softc common; /* All clients of the common layer MUST have this as the first element */
   cx88_channel channel; /* MPEG TS DMA channel */
   cx88_i2c_softc i2c;
   struct cdev *devnode; /* Device file structure */
   struct mtx isrlock; /* Synchronize with interrupt service routine */
   boolean_t data_avail; /* Indicates an interrupt has completed and data is available */
   struct sx sxlock; /* Driver entry point lock */
   struct selinfo selectors; /* Selectors polling for an interrupt */
   uint32_t gen_ctrl; /* TS_GEN_CNTRL register value */
   volatile uint8_t isr_status; /* Error status from interrupt service routine */
   volatile uint8_t buffer_available; /* 1-based index of most recent completed buffer */
   volatile uint64_t packet_count; /* Total packets transferred since current capture began */
   boolean_t in_use; /* Indicates whether device is active */
   uint16_t overflows; /* FIFO overflow counter */
   uint16_t bad_packets; /* Bad (!188-byte) packet counter */
   boolean_t reconfig_requested; /* Indicates a sibling driver has requested resource reallocation */
   cx88_mpeg_state_t state; /* Current driver state */
} cx88_mpeg_softc;

static int cx88_mpeg_setup (void *arg);
static int cx88_mpeg_teardown (void *arg);
static void cx88_mpeg_intr (void *arg);
static int cx88_mpeg_dmaprog (cx88_buffer *buffer, uint8_t bufindex);
static int cx88_mpeg_probe(device_t dev);
static int cx88_mpeg_attach(device_t dev);
static int cx88_mpeg_shutdown(device_t dev);
static int cx88_mpeg_suspend(device_t dev);
static cx88_i2c_bus *cx88_mpeg_get_i2c_bus(device_t dev);
static cx88_i2c_dev *cx88_mpeg_get_i2c_dev(device_t dev, uint8_t addr);
static int cx88_mpeg_configure(cx88_mpeg_softc *sc, cx88_mpeg_config *conf, struct thread *context);
static int cx88_mpeg_transfer(cx88_mpeg_softc *sc, cx88_mpeg_wait *wait, boolean_t nonblocking);
static void cx88_mpeg_reset(cx88_mpeg_softc *sc, cx88_mpeg_state_t state);
static int cx88_mpeg_start_capture(cx88_mpeg_softc *sc, cx88_mpeg_capture *capture);
static void cx88_mpeg_reconfigure(cx88_mpeg_softc *sc);
static int cx88_mpeg_request_reconfig(device_t dev);

static devclass_t cx88_mpeg_devclass;

static d_open_t cx88_mpeg_open;
static d_close_t cx88_mpeg_close;
static d_mmap_t cx88_mpeg_mmap;
static d_ioctl_t cx88_mpeg_ioctl;
static d_poll_t cx88_mpeg_poll;

/* Devnode descriptor */
static struct cdevsw cx88_mpeg_cdevsw =
{
   .d_version = D_VERSION,
   .d_flags = D_TRACKCLOSE,
   .d_name = "cx88mpeg",
   .d_open = cx88_mpeg_open,   /* open */
   .d_close = cx88_mpeg_close, /* close */
   .d_mmap = cx88_mpeg_mmap,   /* mmap */
   .d_ioctl = cx88_mpeg_ioctl, /* ioctl */
   .d_poll = cx88_mpeg_poll,   /* poll */
};

/* Client driver descriptor for cx88 common layer */
static const cx88_client_driver cx88_mpeg_client = 
{
   .softc_size               = sizeof(cx88_mpeg_softc),
   .function                 = CX88_FUNC_MPEGTS,
   .setup                    = cx88_mpeg_setup,
   .teardown                 = cx88_mpeg_teardown,
   .interrupt_handler        = cx88_mpeg_intr,
   .description              = "Conexant CX2388x MPEG Transport Stream"
};

/* Setup callback to be called by cx88 common layer during device attach */
static int cx88_mpeg_setup (void *arg)
{
   cx88_mpeg_softc *sc = arg; /* Argument will always be client driver's softc */
   int error = 0;
   sc->reconfig_requested = FALSE;
   sc->in_use = FALSE;
   mtx_init(&(sc->isrlock), "CX88 MPEG ISR lock", NULL, MTX_DEF);
   sx_init(&(sc->sxlock), "CX88 MPEG SX lock");
   if ((error = cx88_init_channel(&(sc->channel), &(sc->common), CX88_CH_MPEG_TS, MAX_MPEG_BUFFER_SIZE)))
   {
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
      return error;
   }
   /* Create I2C bus instance */
   if ((error = cx88_i2c_attach(&(sc->i2c), &(sc->common))))
   {
      cx88_destroy_channel (&(sc->channel));
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
      return error;
   }
   /* Create device file--this should be done as the last step in initialization */
   sc->devnode = make_dev(&cx88_mpeg_cdevsw, device_get_unit(sc->common.dev), UID_ROOT, GID_WHEEL, 0644, device_get_nameunit(sc->common.dev));
   if (sc->devnode == NULL)
   {
      cx88_i2c_detach(&(sc->i2c));
      cx88_destroy_channel (&(sc->channel));
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
      return ENOMEM;
   }
   cx88_mpeg_reset(sc, CX88_MPEG_UNCONFIGURED);
   return 0;
}

/* Cleanup callback to be called by cx88 common layer during device detach */
static int cx88_mpeg_teardown (void *arg)
{
   int error = 0;
   cx88_mpeg_softc *sc = arg; /* Argument will always be client driver's softc */
   sx_slock(&(sc->sxlock));
   cx88_mpeg_reset(sc, CX88_MPEG_INVALID);
   sx_sunlock(&(sc->sxlock));
   sx_xlock(&(sc->sxlock));
   cx88_release_channel (&(sc->channel)); /* Release all channel and buffer resources */
   sx_xunlock(&(sc->sxlock));
   if (!error && (sc->devnode != NULL))
   {
      /* Device file should be destroyed as the second step in cleanup*/
      destroy_dev(sc->devnode);
      sc->devnode = NULL;
   }
   /* Now remove static parts of the softc */
   if (!error)
   {
      error = cx88_i2c_detach(&(sc->i2c));
   }
   if (!error)
   {
      cx88_destroy_channel (&(sc->channel));
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
   }
   return error;
}

/* Open file operation */
static int cx88_mpeg_open(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   cx88_mpeg_softc *sc = devclass_get_softc(cx88_mpeg_devclass, minor(dev));
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Opening device\n"))
   sx_xlock(&(sc->sxlock));
   if (sc->in_use)
   {
      sx_xunlock(&(sc->sxlock));
      return EBUSY;
   }
   if (sc->state == CX88_MPEG_CONFIGURED)
   {
      cx88_i2c_request_devices(&(sc->i2c));
   }
   sc->in_use = TRUE;
   sx_xunlock(&sc->sxlock);
   return 0;
}

/* Close file operation */
static int cx88_mpeg_close(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   cx88_mpeg_softc *sc = devclass_get_softc(cx88_mpeg_devclass, minor(dev));
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Closing device\n"))
   sx_xlock(&(sc->sxlock));
   cx88_mpeg_reset(sc, sc->state);
   /* If we're using user-allocated buffers, release them as
   * they can't be persisted across open() and close() in
   * different processes */
   cx88_free_userbufs(&(sc->channel), NULL);
   if (sc->reconfig_requested)
   {
      cx88_mpeg_reconfigure(sc);
      sc->reconfig_requested = FALSE;
   }
   /* Release ownership of I2C devices */
   cx88_i2c_release_devices(&(sc->i2c));
   sc->in_use = FALSE;
   sx_xunlock(&(sc->sxlock));
   return 0;
}

/* Configures channel, buffers, and I2C devices */
static int cx88_mpeg_configure(cx88_mpeg_softc *sc, cx88_mpeg_config *config, struct thread *context)
{
   int error = 0;
   if (sc->state < CX88_MPEG_UNCONFIGURED)
   {
      return ENXIO;
   }
   cx88_mpeg_reset(sc, CX88_MPEG_UNCONFIGURED); /* Ensure there is no outstanding capture */
   size_t buffer_size, last_buffer_size;
   uint32_t num_packets, packets_per_buffer, packets_last_buffer, max_packets_per_buffer;
   uint8_t num_buffers = 0;
   /* Validate ioctl parameters */
   if ((config->inout_num_packets == 0) || (config->in_packet_size == 0) || 
      ((config->in_buftype == CX88_MPEG_BUF_USER) && (config->bufconfig.in_userbuf == NULL)) ||
      ((config->in_buftype != CX88_MPEG_BUF_USER) && (config->in_buftype != CX88_MPEG_BUF_KERNEL)))
   {
      return EINVAL;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Configuring MPEG TS channel\n"))
   /* Allocate the channel */
   error = cx88_configure_channel(
      &(sc->channel),
      0,
      config->in_packet_size,
      MAX_MPEG_LINE_SIZE,
      ((config->in_buftype == CX88_MPEG_BUF_KERNEL) ? 0 : config->inout_num_packets));
   if (!error)
   {
      /* Ensure the total number of packets in our buffer is a multiple of the line size,
       * since our DMA program runs in a continuous loop */
      num_packets = cx88_risc_check_bufsize(&(sc->channel), config->inout_num_packets);
      max_packets_per_buffer = MAX_MPEG_BUFFER_SIZE / config->in_packet_size;
      if (max_packets_per_buffer == 0)
      {
         return EINVAL;
      }
      num_buffers = num_packets / max_packets_per_buffer;
      packets_per_buffer = 0;
      if (num_packets % max_packets_per_buffer)
      {
         ++num_buffers; /* Ensure buffers will be no larger than MAX_MPEG_BUFFER_SIZE */
      }
      if (num_buffers < 2)
      {
         /* We need at least two buffers so the application can process one while the
          * other is being filled by the driver */
         num_buffers = 2;
      }
      packets_per_buffer = num_packets / num_buffers;
      if (packets_per_buffer == 0)
      {
         return EINVAL;
      }
      if (num_packets % num_buffers)
      {
         /* Ensure we don't have a trailing buffer that's too large */
         ++packets_per_buffer;
      }
      if ((packets_per_buffer * config->in_packet_size) < MIN_MPEG_BUFFER_SIZE)
      {
         return EINVAL;
      }
      /* Number of packets in last buffer will be less than packets_per_buffer
       * if num_packets is not a multiple of num_buffers */
      packets_last_buffer = num_packets - ((num_buffers - 1) * packets_per_buffer);
      config->inout_num_packets = num_packets;
      config->out_num_buffers = num_buffers;
      config->out_packets_per_buffer = packets_per_buffer;
      buffer_size = packets_per_buffer * config->in_packet_size;
      last_buffer_size = packets_last_buffer * config->in_packet_size;
      if (config->in_buftype == CX88_MPEG_BUF_KERNEL)
      {
         /* Calculate the total amount of memory the application will need to mmap(), along
          * with the buffer size the application will need to use in calculating the offsets
          * of individual buffers within that mmap'ed region */
         config->bufconfig.out_kernelbuf.out_buffer_size = round_page(buffer_size);
         config->bufconfig.out_kernelbuf.out_mmap_size = 
            ((num_buffers - 1) * config->bufconfig.out_kernelbuf.out_buffer_size) + last_buffer_size;
      }
      DIAGNOSTIC(printf("Packets per buffer = %d\n", packets_per_buffer))
      /* Request the buffer structures */
      error = cx88_request_buffers (&(sc->channel), (uint8_t)num_buffers, 0);
   }
   if (!error)
   {
      uint32_t i;
      for (i = 0; i < num_buffers; ++i)
      {
         void *bufptr = NULL;
         struct thread *bufcontext = NULL;
         if (config->in_buftype == CX88_MPEG_BUF_USER)
         {
            /* Each cx88 buffer is a segment of the user-allocated buffer */
            bufptr = (uint8_t *)(config->bufconfig.in_userbuf) + (i * buffer_size);
            bufcontext = context;
         }
         /* Configure each buffer */
         error = cx88_configure_buffer(
            &(sc->channel.buffers[i]),
            bufptr, /* This will be NULL for kernel-allocated buffers */
            bufcontext,
            ((i == (num_buffers - 1)) ? last_buffer_size : buffer_size),
            ((i == (num_buffers - 1)) ? 0 : (i + 1)),
            cx88_mpeg_dmaprog,
            CX88_REPROGRAM_ON_BUFFER_CHANGE);
         if (error)
         {
            break;
         }
      }
   }
   if (!error)
   {
      /* Now commit all changes we've made to the channel configuration */
      error = cx88_commit_channel(&(sc->channel));
   }
   if (!error)
   {
      sc->state = CX88_MPEG_CONFIGURED;
   }
   else
   {
      cx88_release_channel(&(sc->channel));
   }
   return error;
}

static void cx88_mpeg_reset(cx88_mpeg_softc *sc, cx88_mpeg_state_t state)
{
   mtx_lock(&(sc->isrlock));
   /* Stop any data transfer */
   cx88_write_reg(&(sc->common), MPEG_REG_DMA_CONTROL, 0);
   /* Disable interrupts */
   cx88_write_reg(&(sc->common), MPEG_REG_INT_MASK, 0);
   /* Reset capture status values */
   sc->data_avail = FALSE;
   sc->isr_status = CX88_MPEG_WAIT_ABORTED;
   sc->overflows = 0;
   sc->bad_packets = 0;
   /* Our state value should never be higher than CX88_MPEG_CONFIGURED, as we won't be capturing once reset is done */
   sc->state = CX88_MPEG_CONFIGURED;
   if (state < sc->state)
   {
      /* Reset alone should never raise the state value */
      sc->state = state;  
   }
   wakeup(&(sc->data_avail));
   selwakeup(&(sc->selectors));
   mtx_unlock(&(sc->isrlock));
   
   /* Protect access to the shared PCI interrupt and PINMUX I/O registers. */
   cx88_lock_board(&(sc->common));
   cx88_write_reg(&(sc->common), CX88_REG_PCI_INTMASK,
      (cx88_read_reg(&(sc->common), CX88_REG_PCI_INTMASK) & ~CX88_PCI_INT_MPEGTS));
   if (!(sc->gen_ctrl & CX88_MPEG_SERIAL_INPUT))
   {
      cx88_write_reg(&(sc->common), CX88_REG_PINMUX_IO,
         (cx88_read_reg(&(sc->common), CX88_REG_PINMUX_IO) & ~CX88_PINMUX_MPEG_PAR));
   }
   cx88_unlock_board(&(sc->common));

   sc->gen_ctrl = 0;
   cx88_write_reg(&(sc->common), MPEG_REG_GEN_CONTROL, (MPEG_GEN_STAT_CLEAR | MPEG_GEN_SW_RESET));
   pause("cx88p", 1); /* Wait for reset state to clear */
   cx88_write_reg(&(sc->common), MPEG_REG_GEN_CONTROL, 0);
}

/* Capture data--'capture' parameter may be NULL if we are re-starting an
 * existing capture. */
static int cx88_mpeg_start_capture(cx88_mpeg_softc *sc, cx88_mpeg_capture *capture)
{
   int error = 0;
   cx88_mpeg_reset(sc, sc->state); /* Make sure we're in a clean, valid state first */
   if (sc->state != CX88_MPEG_CONFIGURED)
   {
      return ENXIO;
   }
   if ((sc->channel.packet_size > 0xFFF) || (capture->in_start_byte_count > 0xF))
   {
      device_printf(sc->common.dev, "Invalid packet size or start byte count\n");
      return EINVAL;  
   }
   sc->isr_status = CX88_MPEG_CAP_NOERR;
   sc->buffer_available = 0;
   sc->packet_count = 0;
   sc->state = CX88_MPEG_CAPTURING;
   cx88_write_reg(&(sc->common), MPEG_REG_LINE_LENGTH, sc->channel.line_size);
   /* Configure start byte and packet length--we must see 1 start byte before 
    * data transfer can begin */
   cx88_write_reg(&(sc->common), MPEG_REG_SOP_CONTROL,
      (cx88_write_field(MPEG_SOP_START_BYTE, capture->in_start_byte) |
      cx88_write_field(MPEG_SOP_PACKET_LENGTH, sc->channel.packet_size) |
      cx88_write_field(MPEG_SOP_START_COUNT, capture->in_start_byte_count)));

   /* Configure general control register based on application input */
   sc->gen_ctrl = ((capture->in_options & CX88_MPEG_SERIAL_INPUT) ? MPEG_GEN_SERIAL : 0);
   sc->gen_ctrl |= ((capture->in_options & CX88_MPEG_PUNCTURED_CLOCK) ? MPEG_GEN_PUNC_CLK : 0);
   sc->gen_ctrl |= ((capture->in_options & CX88_MPEG_POS_POLARITY) ? MPEG_GEN_POSEDGE : 0);
   sc->gen_ctrl |= ((capture->in_options & CX88_MPEG_LITTLE_ENDIAN) ? MPEG_GEN_BIT_REVERSE : 0);

   if ((sc->gen_ctrl & CX88_MPEG_SERIAL_INPUT) && (capture->in_options & CX88_MPEG_BYTEWIDE_SOP))
   {
      cx88_write_reg(&(sc->common), MPEG_REG_SOP_STATUS, MPEG_SOP_BYTEWIDE);
   }
   /* Enable FIFO overflow and bad packet checks */
   cx88_write_reg(&(sc->common), MPEG_REG_FIFO_STATUS, MPEG_FIFO_OVERFLOW_INTR);
   cx88_write_reg(&(sc->common), MPEG_REG_BAD_PACKET_STAT, MPEG_BAD_PACKET_ENABLE);
   
   /* Protect access to the shared PCI interrupt and PINMUX I/O registers. */
   cx88_lock_board(&(sc->common));
   cx88_write_reg(&(sc->common), CX88_REG_PCI_INTSTAT, CX88_PCI_INT_MPEGTS);
   cx88_write_reg(&(sc->common), CX88_REG_PCI_INTMASK,
      (cx88_read_reg(&(sc->common), CX88_REG_PCI_INTMASK) | CX88_PCI_INT_MPEGTS));
   if (!(sc->gen_ctrl & CX88_MPEG_SERIAL_INPUT))
   {
      /* Configure parallel input */
      cx88_write_reg(&(sc->common), CX88_REG_PINMUX_IO,
         (cx88_read_reg(&(sc->common), CX88_REG_PINMUX_IO) | CX88_PINMUX_MPEG_PAR));
   }
   else
   {
      cx88_write_reg(&(sc->common), CX88_REG_PINMUX_IO,
         (cx88_read_reg(&(sc->common), CX88_REG_PINMUX_IO) & (~CX88_PINMUX_MPEG_PAR)));
   }
   cx88_unlock_board(&(sc->common));
   cx88_write_reg(&(sc->common), MPEG_REG_GEN_CONTROL, sc->gen_ctrl);
   /* Acknowledge any outstanding interrupts */
   cx88_write_reg(&(sc->common), MPEG_REG_INT_STAT,
      cx88_read_reg(&(sc->common), MPEG_REG_INT_STAT));
   /* Enable interrupts */
   cx88_write_reg(&(sc->common), MPEG_REG_INT_MASK,
      (MPEG_INT_RISCI1 | MPEG_INT_SYNC | MPEG_INT_OPCODE_ERR |
      MPEG_INT_PAR_ERR | MPEG_INT_RIP_ERR | MPEG_INT_PCI_ABORT));
   cx88_write_reg(&(sc->common), CX88_REG_DEVCNTL2, CX88_RUN_RISC);
   /* Start the data transfer */
   cx88_write_reg(&(sc->common), MPEG_REG_DMA_CONTROL, (MPEG_RISC_ENABLE | MPEG_FIFO_ENABLE));
   return error;
}

/* Either wait for data to become available (if !nonblocking), or check if data is available.
 * If data is available, then return buffer count and transfer status to application */
static int cx88_mpeg_transfer(cx88_mpeg_softc *sc, cx88_mpeg_wait *wait, boolean_t nonblocking)
{
   int error = 0;
   uint16_t prev_overflows, overflows, prev_bad_packets, bad_packets;
   wait->out_buffer_available = 0;
   wait->out_status = CX88_MPEG_CAP_NOERR;
   mtx_lock(&(sc->isrlock));
   if (sc->state != CX88_MPEG_CAPTURING)
   {
      mtx_unlock(&(sc->isrlock));
      return ENXIO;
   }
   while (!error && !(sc->isr_status & CX88_MPEG_CAP_FATAL) && !sc->data_avail)
   {
      if (!nonblocking)
      {
         error = msleep(&(sc->data_avail), &(sc->isrlock), PZERO | PCATCH, "cx88mp", (wait->in_timeout_s * hz));
      }
      else
      {
         error = EWOULDBLOCK;
      }
   }
   if (!error)
   {
      sc->data_avail = FALSE;
   }
   /* Copy buffer and transfer status */
   wait->out_buffer_available = sc->buffer_available;
   wait->out_status = sc->isr_status;
   sc->isr_status = CX88_MPEG_CAP_NOERR;
   wait->out_packet_count = sc->packet_count;
   prev_bad_packets = sc->bad_packets;
   prev_overflows = sc->overflows;
   /* Read FIFO overflow and bad packet counters */
   overflows = cx88_read_field(cx88_read_reg(&(sc->common), MPEG_REG_FIFO_STATUS), MPEG_FIFO_OVERFLOW_COUNT);
   bad_packets = cx88_read_field(cx88_read_reg(&(sc->common), MPEG_REG_BAD_PACKET_STAT), MPEG_BAD_PACKET_COUNT);
   sc->overflows = overflows;
   sc->bad_packets = bad_packets;
   mtx_unlock(&(sc->isrlock));
   /* For blocking I/O, msleep() timed out */
   if ((error == EWOULDBLOCK) && wait->in_timeout_s)
   {
      wait->out_status |= CX88_MPEG_CAP_TIMEOUT;
      error = 0;
   }
   else if (error)
   {
      wait->out_status |= CX88_MPEG_WAIT_ABORTED;
   }
   /* Compute the number of FIFO overflows since the last call to cx88_mpeg_transfer() */
   if (overflows != prev_overflows)
   {
      wait->out_status |= CX88_MPEG_HW_OVERFLOW;
      if (overflows < prev_overflows)
      {
         /* The overflow counter has wrapped around */
         wait->out_num_overflows = overflows + 0x1000 - prev_overflows;
      }
      else
      {
         wait->out_num_overflows = overflows - prev_overflows;
      }
   }
   /* Compute the number of bad packets since the last call to cx88_mpeg_transfer() */
   if (bad_packets != prev_bad_packets)
   {
      wait->out_status |= CX88_MPEG_BAD_PACKET;
      if (bad_packets < prev_bad_packets)
      {
         /* The bad packet counter has wrapped around */
         wait->out_num_bad_packets = bad_packets + 0x1000 - prev_bad_packets;
      }
      else
      {
         wait->out_num_bad_packets = bad_packets - prev_bad_packets;
      }
   }
   return error;
}

/* Ioctl file operation */
static int cx88_mpeg_ioctl(struct cdev *dev, u_long cmd, caddr_t data, int flags, struct thread *td)
{
   cx88_mpeg_softc *sc = devclass_get_softc(cx88_mpeg_devclass, minor(dev));
   int error = 0;
   if (sc == NULL)
   {
      return ENXIO;
   }
   switch (cmd)
   {
      case CX88_COMMON_CARD_ID: /* Retrieve vendor-specific identifiers */
         *((uint32_t*)data) = cx88_get_card(sc->common.dev);
         break;
      case CX88_MPEG_CONFIG:
         sx_xlock(&(sc->sxlock));
         error = cx88_mpeg_configure(sc, (cx88_mpeg_config*)data, td);
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_COMMON_I2C_CONFIG:
         sx_xlock(&(sc->sxlock));         
         if (sc->state < CX88_MPEG_UNCONFIGURED)
         {
            error = ENXIO;
         }
         else
         {
            error = cx88_i2c_update_devices(&(sc->i2c), (cx88_common_i2c_config*)data, IICDEV_NAME);
            if (!error)
            {
               error = cx88_i2c_request_devices(&(sc->i2c));
            }
         }
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_MPEG_CAPTURE:
         sx_xlock(&(sc->sxlock));
         error = cx88_mpeg_start_capture(sc, (cx88_mpeg_capture*)data);
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_MPEG_STOP:
         /* Wait and stop are slock'ed so that a stop can be issued with a pending wait */
         sx_slock(&(sc->sxlock));
         cx88_mpeg_reset(sc, sc->state);
         sx_sunlock(&(sc->sxlock));
         break;
      case CX88_MPEG_WAIT:
         sx_slock(&(sc->sxlock));
         error = cx88_mpeg_transfer(sc, (cx88_mpeg_wait*)data, ((flags & O_NONBLOCK) ? TRUE : FALSE));
         sx_sunlock(&(sc->sxlock));
         break;
      case CX88_COMMON_GPIO:
         cx88_gpioctl(&(sc->common), (cx88_common_gpio*)data);
         break;
      default:
         error = ENOTTY;
         break;
   }
   return error;
}

/* Select/poll file operation */
static int cx88_mpeg_poll(struct cdev *dev, int events, struct thread *td)
{
   cx88_mpeg_softc *sc = devclass_get_softc(cx88_mpeg_devclass, minor(dev));
   if (sc == NULL)
   {
      return ENXIO;
   }
   int revents = 0;
   sx_xlock(&(sc->sxlock));
   mtx_lock(&(sc->isrlock));
   if ((sc->state == CX88_MPEG_CAPTURING) && (events & (POLLIN | POLLRDNORM)))
   {
      if (sc->data_avail)
      {
         /* Notify the application that data is available */
         revents = events & (POLLIN | POLLRDNORM);
      }
      else
      {
         selrecord(td, &(sc->selectors));
      }
   }
   mtx_unlock(&(sc->isrlock));
   sx_xunlock(&(sc->sxlock));
   return revents;
}

/* Interrupt service routine */
static void cx88_mpeg_intr(void *arg)
{
   cx88_mpeg_softc *sc = arg;
   mtx_lock(&(sc->isrlock));
   /* Ensure we actually have an interrupt--use MSTAT register since we
    * only care about interrupts we've enabled */
   uint32_t intrs = cx88_read_reg(&(sc->common), MPEG_REG_INT_MSTAT);
   if (intrs)
   {
      if (sc->state != CX88_MPEG_CAPTURING)
      {
         mtx_unlock(&(sc->isrlock));
         return;  
      }
      /* Check for fatal errors */
      if (intrs & (MPEG_INT_OPCODE_ERR | MPEG_INT_SYNC))
      {
         sc->isr_status |= CX88_MPEG_DMA_PROGRAM;
      }
      if (intrs & (MPEG_INT_PCI_ABORT | MPEG_INT_PAR_ERR | MPEG_INT_RIP_ERR))
      {
         sc->isr_status |= CX88_MPEG_BUS_ERROR;
      }
      if (sc->isr_status & CX88_MPEG_CAP_FATAL)
      {
         /* We should stop the transfer as soon as possible if we've encountered a fatal error */
         cx88_write_reg(&(sc->common), MPEG_REG_DMA_CONTROL, 0);
      }
      if (intrs & MPEG_INT_RISCI1)
      {
         /* The counter indicates the most recently completed buffer */
         uint32_t bufindex = cx88_read_reg(&(sc->common), MPEG_REG_GP_COUNT);
         if (bufindex < sc->channel.num_buffers)
         {
            /* Perform any necessary cache/bounce buffer synchronization.
             * Do this in a loop to account for the unlikely case where
             * we've missed an interrupt and multiple buffers need to be 
             * synchronized */
            uint8_t sync_bufindex = sc->buffer_available;
            do
            {
               if (sync_bufindex >= sc->channel.num_buffers)
               {
                  sync_bufindex = 0;
               }
               cx88_buffer *buffer = &(sc->channel.buffers[sync_bufindex]);
               /*Synchronize the buffer */
               cx88_mark_transfer_complete(buffer);
               cx88_prepare_for_transfer(buffer);
            }
            while ((sync_bufindex++) != bufindex);
            /* Available index is 1-based */
            sc->buffer_available = bufindex + 1;
            /* Adjust packet count */
            sc->packet_count += (sc->channel.buffers[bufindex].data_buffer_size / sc->channel.packet_size);
            sc->data_avail = TRUE;
         }
      }
      /* Acknowledge the interrupt */
      cx88_write_reg(&(sc->common), MPEG_REG_INT_STAT, intrs);
      wakeup(&(sc->data_avail));
      selwakeup(&(sc->selectors));
   }
   mtx_unlock(&(sc->isrlock));
}

/* RISC program generator */
static int cx88_mpeg_dmaprog(cx88_buffer *buffer, uint8_t bufindex)
{
   int error = 0;
   DIAGNOSTIC(device_printf(buffer->channel->sc->dev, "Generating DMA program\n"))
   if (bufindex == 0)
   {
      /* Initial SYNC at beginning of capture */
      cx88_risc_sync(buffer, 0, TRUE, CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
   }
   /* Transfer the entire buffer */
   error = cx88_risc_transfer_dmabuf (buffer, 0, (uint32_t)(buffer->data_buffer_size), CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
   if (!error)
   {
      /* Jump to next buffer, adjust the counter to reflect the index of the completed
       * buffer, and trigger RISC interrupt 1 so that the buffer can be processed */
      error = cx88_risc_next_buffer(
         buffer,
         (bufindex == (buffer->channel->num_buffers - 1) ? CX88_RISC_SYNC_SIZE : 0),
         CX88_TRIGGER_RISCI1,
         ((bufindex == 0) ? CX88_RESET_COUNT : CX88_INCREMENT_COUNT));
   }
   return error;
}

/* Mmap file operation */
static int cx88_mpeg_mmap(
   struct cdev *dev,
   vm_offset_t offset,
   vm_paddr_t *paddr,
   int nprot)
{
   cx88_mpeg_softc *sc = devclass_get_softc(cx88_mpeg_devclass, minor(dev));
   sx_slock(&(sc->sxlock));
   DIAGNOSTIC(device_printf(sc->common.dev, "Attempting to mmap device memory\n"))
   if (nprot & (PROT_EXEC | PROT_WRITE))
   {
      /* MPEG TS is output-only */
      DIAGNOSTIC(device_printf(sc->common.dev, "cx88_mpeg_mmap: Invalid device memory permissions\n"))
      return EINVAL;
   }
   int error = cx88_common_mmap(&(sc->channel), offset, paddr, nprot);
   sx_sunlock(&(sc->sxlock));
   return error;
}

/* Probe DEVMETHOD */
static int cx88_mpeg_probe(device_t dev) 
{
   return cx88_common_probe(dev, &cx88_mpeg_client);
}

/* Attach DEVMETHOD */
static int cx88_mpeg_attach(device_t dev) 
{
   return cx88_common_attach(dev, &cx88_mpeg_client);
}

/* Shutdown DEVMETHOD */
static int cx88_mpeg_shutdown(device_t dev)
{
   cx88_mpeg_softc *sc = device_get_softc(dev);
   cx88_mpeg_reset(sc, CX88_MPEG_INVALID);
   return 0;
}

/* Suspend DEVMETHOD */
static int cx88_mpeg_suspend(device_t dev)
{
   int error = 0;
   cx88_mpeg_softc *sc = device_get_softc(dev);
   sx_slock(&(sc->sxlock));
   cx88_mpeg_reset(sc, CX88_MPEG_UNCONFIGURED);
   sx_sunlock(&(sc->sxlock));
   sx_xlock(&(sc->sxlock));
   cx88_release_channel (&(sc->channel)); /* Release all channel and buffer resources */
   sx_xunlock(&(sc->sxlock));
   return error;
}

/* I2C bus accessor DEVMETHOD */
static cx88_i2c_bus *cx88_mpeg_get_i2c_bus(device_t dev)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx88_mpeg_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return cx88_i2c_get_bus(&(sc->i2c));
}

/* I2C device accessor DEVMETHOD */
static cx88_i2c_dev *cx88_mpeg_get_i2c_dev(device_t dev, uint8_t addr)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx88_mpeg_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return cx88_i2c_get_dev(&(sc->i2c), addr);
}

static void cx88_mpeg_reconfigure(cx88_mpeg_softc *sc)
{
   cx88_free_sram(&(sc->channel));
   if (sc->state > CX88_MPEG_UNCONFIGURED)
   {
      sc->state = CX88_MPEG_UNCONFIGURED;
   }
}

/* Resource reallocation DEVMETHOD */
static int cx88_mpeg_request_reconfig(device_t dev)
{
   int error = 0;
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx88_mpeg_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return ENODEV;
   }
   /* Since this may be called from a sibling driver in a different thread
    * than any thread that is doing I/O with this driver, use trylock
    * a) because blocking on the lock could deadlock and
    * b) because the sibling wouldn't want to wait an indeterminate amount
    *    of time for I/O on this driver to complete */
   if (sx_try_xlock(&(sc->sxlock)))
   {
      if (sc->in_use)
      {
         /* If someone has a handle open to our device,
          * don't disrupt their session.  We'll just free our resources
          * when the handle is closed that they can be re-allocated
          * the next time a handle is opened. */
         sc->reconfig_requested = TRUE;
         error = EINPROGRESS;
      }
      else
      {
         cx88_mpeg_reconfigure(sc);
      }
      sx_xunlock(&(sc->sxlock));
   }
   else
   {
      sc->reconfig_requested = TRUE;
      error = EINPROGRESS;
   }
   return error;
}

/* DEVMETHODs */
static device_method_t cx88_mpeg_methods[] = 
{
   /* Device interface */
   DEVMETHOD(device_probe,             cx88_mpeg_probe),
   DEVMETHOD(device_attach,            cx88_mpeg_attach),
   DEVMETHOD(device_detach,            cx88_common_detach),
   DEVMETHOD(device_shutdown,          cx88_mpeg_shutdown),
   DEVMETHOD(device_suspend,           cx88_mpeg_suspend),
   DEVMETHOD(device_resume,            cx88_common_resume),

   /* cx88 interface */
   CX88_METHODS(cx88_mpeg_request_reconfig),
             
   /* i2c host methods */
   CX88_I2C_METHODS(cx88_mpeg_get_i2c_bus, cx88_mpeg_get_i2c_dev),

   {NULL, NULL}
};

/* Module declarations */
static driver_t cx88_mpeg_driver =
{
   "cx88mpeg",
   cx88_mpeg_methods,
   0
};

DRIVER_MODULE(cx88mpeg, pci, cx88_mpeg_driver, cx88_mpeg_devclass, NULL, NULL);
DRIVER_MODULE(cx88mpeg, cardbus, cx88_mpeg_driver, cx88_mpeg_devclass, NULL, NULL);
CX88_DRIVER_MODULE(cx88mpeg);
CX88_I2C_DRIVER_MODULE(cx88mpeg);
MODULE_DEPEND(cx88mpeg, iicdev, IICDEV_MINVER, IICDEV_MODVER, IICDEV_MAXVER);
MODULE_VERSION(cx88mpeg, 1);
