/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/module.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/selinfo.h>
#include <sys/poll.h>
#include <sys/fcntl.h>
#include <sys/malloc.h>
#include "cx88_common.h"
#include "cx88_if.h"
#include "iicdev.h"
#include "cx88_i2c.h"
#include "cx88_i2c_if.h"
#include "cx88_regs.h"
#include "cx88_video_regs.h"
#include "cx88_video_ioctl.h"

#define MAX_VIDEO_BUFFER_SIZE (1 << 21)

/* Driver states */
typedef enum
{
   CX88_VIDEO_INVALID = 0, /* Uninitialized */
   CX88_VIDEO_UNCONFIGURED, /* Not configured or formatted */
   CX88_VIDEO_FORMATTED, /* Formatted but not configured */
   CX88_VIDEO_CONFIGURED, /* Configured (I2C devices, buffers, etc.) and formatted */
   CX88_VIDEO_CAPTURING /* Configured, formatted, and capturing */
} cx88_video_state_t;

typedef struct
{
   uint32_t mmap_size;
   uint16_t x_header;
   uint16_t x_trailer;
   uint16_t y_header;
   uint16_t y_trailer;
   void *clip_patterns;
   uint32_t clip_pattern_size;
   volatile uint8_t even_frame_available;
   volatile uint8_t odd_frame_available;
   boolean_t enabled;
} cx88_video_channel_config;

typedef struct
{
   uint32_t sample_freq;
   uint32_t hactive;
   uint32_t hdelay;
   uint32_t vactive;
   uint32_t vdelay;
   uint16_t vbi_samples;
   uint16_t pixels_per_line;
} cx88_video_spec;

#define CX88_VIDEO_SPECS CX88_VIDEO_FMT_SECAM
#define CX88_FMT_TO_SPEC(fmt) (fmt - 1)

static const cx88_video_spec specs[CX88_VIDEO_SPECS] = 
{
   /*sample_freq  hactive  hdelay   vactive  vdelay   vbi_samples    pixels_per_line*/
   {28636363,     754,     135,     480,     24,      1600,          910},  /* NTSC-M */
   {28636363,     754,     135,     480,     24,      1600,          910},  /* NTSC-J */
   {35468950,     922,     186,     480,     24,      2044,          1126}, /* NTSC-4.43 */
   {35468950,     922,     186,     576,     36,      2044,          1135}, /* PAL-B/D/G/H/I */
   {28604888,     754,     135,     480,     24,      1600,          909},  /* PAL-M */
   {28656448,     754,     135,     576,     36,      1600,          917},  /* PAL-N */
   {28656448,     754,     135,     576,     36,      1600,          917},  /* PAL-Nc */
   {35468950,     922,     186,     480,     24,      2044,          1126}, /* PAL-60 */
   {35468950,     922,     186,     576,     36,      2044,          1135}  /* SECAM */
};

typedef struct
{
   cx88_common_softc common;
   cx88_channel channels[CX88_VIDEO_CHANNELS];
   cx88_video_channel_config channel_configs[CX88_VIDEO_CHANNELS];
   cx88_i2c_softc i2c;
   struct cdev *devnode;
   struct mtx isrlock;
   uint8_t channel_mask;
   uint8_t channels_avail;
   struct sx sxlock;
   struct selinfo selectors;
   boolean_t in_use;
   boolean_t reconfig_requested;
   cx88_video_state_t state;
   cx88_video_format format;
   const cx88_video_spec *spec;
   uint8_t num_frames;
   cx88_dmareprogram_t dma_reprogram;
   volatile uint8_t isr_status; /* Error status from interrupt service routine */
} cx88_video_softc;

static int cx88_video_probe(device_t dev);
static int cx88_video_attach(device_t dev);
static int cx88_video_setup(void *arg);
static int cx88_video_teardown(void *arg);
static void cx88_video_intr(void *arg);
static cx88_i2c_bus *cx88_video_get_i2c_bus(device_t dev);
static cx88_i2c_dev *cx88_video_get_i2c_dev(device_t dev, uint8_t addr);
static void cx88_video_reconfigure(cx88_video_softc *sc);
static int cx88_video_request_reconfig(device_t dev);
static int cx88_video_dmaprog(cx88_buffer *buffer, uint8_t bufindex);
static void cx88_video_reset(cx88_video_softc *sc, cx88_video_state_t state);
static int cx88_video_configure(
   cx88_video_softc *sc,
   cx88_video_config *conf,
   struct thread *context);
static void cx88_video_release_channels(cx88_video_softc *sc);
static int cx88_video_geom(cx88_video_softc *sc, cx88_video_format *fmt);
static int cx88_video_set_clip(cx88_video_softc *sc, cx88_video_clip *clip);
static int cx88_video_start_capture(cx88_video_softc *sc, cx88_video_capture *cap);
static int cx88_video_transfer(cx88_video_softc *sc, cx88_video_wait *wait, boolean_t nonblocking);
static int cx88_video_shutdown(device_t dev);
static int cx88_video_suspend(device_t dev);
static int cx88_video_lineprog(
   cx88_buffer *buffer,
   bus_size_t offset,
   cx88_video_clip_pattern *pattern);
static int cx88_video_fieldprog(
   cx88_buffer *buffer,
   cx88_video_channel_config *config,
   uint16_t *line,
   bus_size_t *offset,
   bus_size_t max_offset,
   bus_size_t line_size,
   uint8_t interlace);

static const cx88_client_driver cx88_video_client = 
{
   .softc_size               = sizeof(cx88_video_softc),
   .function                 = CX88_FUNC_VIDEO,
   .setup                    = cx88_video_setup,
   .teardown                 = cx88_video_teardown,
   .interrupt_handler        = cx88_video_intr,
   .description              = "Conexant CX2388x Analog Video"
};

static devclass_t cx88_video_devclass;

static d_open_t cx88_video_open;
static d_close_t cx88_video_close;
static d_mmap_t cx88_video_mmap;
static d_ioctl_t cx88_video_ioctl;
static d_poll_t cx88_video_poll;

static struct cdevsw cx88_video_cdevsw =
{
   .d_version = D_VERSION,
   .d_flags = D_TRACKCLOSE,
   .d_name = "cx88video",
   .d_open = cx88_video_open,   /* open */
   .d_close = cx88_video_close, /* close */
   .d_mmap = cx88_video_mmap,   /* mmap */
   .d_ioctl = cx88_video_ioctl, /* ioctl */
   .d_poll = cx88_video_poll,   /* poll */
};

static int cx88_video_setup(void *arg)
{   
   cx88_video_softc *sc = arg;
   int error = 0;
   sc->reconfig_requested = FALSE;
   sc->in_use = FALSE;
   sc->dma_reprogram = CX88_REPROGRAM_ON_BUFFER_CHANGE;
   mtx_init(&(sc->isrlock), "CX88 Video ISR lock", NULL, MTX_DEF);
   sx_init(&(sc->sxlock), "CX88 Video SX lock");
   uint8_t i, last_configured_channel = 0;   
   for (i = 0; i < CX88_VIDEO_CHANNELS; ++i, ++last_configured_channel)
   {
      sc->channel_configs[i].mmap_size = 0;
      sc->channel_configs[i].x_header = 0;
      sc->channel_configs[i].x_trailer = 0;
      sc->channel_configs[i].y_header = 0;
      sc->channel_configs[i].y_trailer = 0;
      sc->channel_configs[i].clip_patterns = NULL;
      sc->channel_configs[i].clip_pattern_size = 0;
      sc->channel_configs[i].even_frame_available = 0;
      sc->channel_configs[i].odd_frame_available = 0;
      sc->channel_configs[i].enabled = TRUE;
   }
   int disabled = 0;
   if ((resource_int_value(device_get_name(sc->common.dev), device_get_unit(sc->common.dev),
       "noplanar", &disabled) == 0) && disabled)
   {
      sc->channel_configs[CX88_VIDEO_CHANNEL_U].enabled = FALSE;
      sc->channel_configs[CX88_VIDEO_CHANNEL_V].enabled = FALSE;
   }
   disabled = 0;
   if ((resource_int_value(device_get_name(sc->common.dev), device_get_unit(sc->common.dev),
       "novbi", &disabled) == 0) && disabled)
   {
      sc->channel_configs[CX88_VIDEO_CHANNEL_VBI].enabled = FALSE;
   }
   for (i = 0; i < CX88_VIDEO_CHANNELS; ++i, ++last_configured_channel)
   {
      if (sc->channel_configs[i].enabled &&
          (error = cx88_init_channel(&(sc->channels[i]), &(sc->common),
         (cx88_channel_t)(CX88_CH_VIDEO_PACKED + i), MAX_VIDEO_BUFFER_SIZE)))
      {
         goto fail0;
      }
   }
   /* Create I2C bus instance */
   if ((error = cx88_i2c_attach(&(sc->i2c), &(sc->common))))
   {
      goto fail0;
   }
   /* Create device file--this should be done as the last step in initialization */
   sc->devnode = make_dev(&cx88_video_cdevsw, device_get_unit(sc->common.dev), UID_ROOT, GID_WHEEL, 0644, device_get_nameunit(sc->common.dev));
   if (sc->devnode == NULL)
   {
      goto fail1;
   }
   cx88_video_reset(sc, CX88_VIDEO_UNCONFIGURED);
   goto success;
   fail1:
      cx88_i2c_detach(&(sc->i2c));
   fail0:
      for (i = 0; i < last_configured_channel; ++i)
      {
         if (sc->channel_configs[i].enabled)
         {
            cx88_destroy_channel (&(sc->channels[i]));
         }
      }
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
   success:
      return error;
}

static int cx88_video_teardown(void *arg)
{
   int error = 0;
   cx88_video_softc *sc = arg;
   uint8_t i;
   sx_slock(&(sc->sxlock));
   cx88_video_reset(sc, CX88_VIDEO_INVALID);
   sx_sunlock(&(sc->sxlock));
   sx_xlock(&(sc->sxlock));
   /* Remove user-configurable data--I2C devices, buffers, channel allocation, etc. */
   cx88_video_release_channels(sc);
   sx_xunlock(&(sc->sxlock));
   if (sc->devnode != NULL)
   {
      destroy_dev(sc->devnode);
      sc->devnode = NULL;  
   }
   error = cx88_i2c_detach(&(sc->i2c));
   if (!error)
   {
      for (i = 0; i < CX88_VIDEO_CHANNELS; ++i)
      {
         if (sc->channel_configs[i].enabled)
         {
            cx88_destroy_channel (&(sc->channels[i]));
         }
         free(sc->channel_configs[i].clip_patterns, M_TEMP);
         sc->channel_configs[i].clip_patterns = NULL;
         sc->channel_configs[i].clip_pattern_size = 0;
      }
      mtx_destroy(&(sc->isrlock));
      sx_destroy(&(sc->sxlock));
   }
   return error;
}

static void cx88_video_reset(cx88_video_softc *sc, cx88_video_state_t state)
{
   mtx_lock(&(sc->isrlock));
   /* Stop any data transfer */
   cx88_write_reg(&(sc->common), VIDEO_REG_DMA_CONTROL, 0);
   /* Disable interrupts */
   cx88_write_reg(&(sc->common), VIDEO_REG_INT_MASK, 0);
   /* Disable video capture */
   cx88_write_reg(&(sc->common), VIDEO_REG_CAPTURE_CONTROL, 0);
   
   /* Reset capture status values */
   sc->channels_avail = 0;
   sc->isr_status = CX88_VIDEO_WAIT_ABORTED;
   
   /* Our state value should never be higher than CX88_VIDEO_CONFIGURED, as we won't be capturing once reset is done */
   sc->state = CX88_VIDEO_CONFIGURED;
   if (state < sc->state)
   {
      /* Reset alone should never raise the state value */
      sc->state = state;  
   }
   
   wakeup(&(sc->channels_avail));
   selwakeup(&(sc->selectors));
   mtx_unlock(&(sc->isrlock));
   
   /* Protect access to the shared PCI interrupt register. */
   cx88_lock_board(&(sc->common));
   cx88_write_reg(&(sc->common), CX88_REG_PCI_INTMASK,
      (cx88_read_reg(&(sc->common), CX88_REG_PCI_INTMASK) & ~CX88_PCI_INT_VIDEO));
   cx88_unlock_board(&(sc->common));
   
}

static int cx88_video_transfer(cx88_video_softc *sc, cx88_video_wait *wait, boolean_t nonblocking)
{
   int error = 0;
   mtx_lock(&(sc->isrlock));
   if (sc->state != CX88_VIDEO_CAPTURING)
   {
      mtx_unlock(&(sc->isrlock));
      return ENXIO;
   }
   if (sc->channel_mask & wait->in_channel_mask)
   {
      device_printf(sc->common.dev, "Wait already in progress for conflicting channel(s)\n");
      mtx_unlock(&(sc->isrlock));
      return EINVAL;
   }
   sc->channel_mask |= wait->in_channel_mask;
   while (!error &&
          !(sc->isr_status & CX88_VIDEO_CAP_FATAL) &&
          !(sc->channels_avail & wait->in_channel_mask))
   {
      if (!nonblocking)
      {
         error = msleep(&(sc->channels_avail), &(sc->isrlock), PZERO | PCATCH, 
            "cx88vi", wait->in_timeout_ms);
      }
      else
      {
         error = EWOULDBLOCK;
      }
   }
   if (!error)
   {
      sc->channels_avail &= (~(sc->channels_avail & wait->in_channel_mask));
   }
   sc->channel_mask &= (~(wait->in_channel_mask));
   wait->out_status = sc->isr_status;
   for (uint8_t i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      wait->channels[i].out_even_frame_available = sc->channel_configs[i].even_frame_available;
      wait->channels[i].out_odd_frame_available = sc->channel_configs[i].odd_frame_available;
      if (cx88_read_reg(&(sc->common), VIDEO_REG_INT_STAT) & VIDEO_INT_OVERFLOW(i))
      {
         wait->channels[i].overflow = 1;
      }
      else
      {
         wait->channels[i].overflow = 0;  
      }
   }
   sc->isr_status = CX88_VIDEO_CAP_NOERR;
   mtx_unlock(&(sc->isrlock));
   /* For blocking I/O, msleep() timed out */
   if ((error == EWOULDBLOCK) && wait->in_timeout_ms)
   {
      wait->out_status |= CX88_VIDEO_CAP_TIMEOUT;
      error = 0;
   }
   else if (error)
   {
      wait->out_status |= CX88_VIDEO_WAIT_ABORTED;
   }
   return error;
}  
   
static int cx88_video_start_capture(cx88_video_softc *sc, cx88_video_capture *cap)
{
   cx88_video_reset(sc, sc->state);
   if (sc->state != CX88_VIDEO_CONFIGURED)
   {
      return ENXIO;
   }
   sc->isr_status = CX88_VIDEO_CAP_NOERR;
   uint32_t input_format_reg = VIDEO_IFORM_VSYNC_ENABLE;
   uint32_t filter_control_reg = VIDEO_FILTER_CFILT;
   uint32_t output_format_reg = VIDEO_OFORM_FULL_LUMA | 
      cx88_write_field(VIDEO_OFORM_COMB_RANGE, 0x1F);
   if (sc->format.in_video_format == CX88_VIDEO_FMT_SECAM)
   {
      filter_control_reg |= VIDEO_FILTER_CLPF2;
      filter_control_reg |= cx88_write_field(VIDEO_FILTER_CLPF1, VIDEO_CLPF1_29_TAP);
   }
   if (cap->in_input_type == CX88_VIDEO_INPUT_SVIDEO)
   {
      filter_control_reg |= VIDEO_FILTER_LNOTCH_DISABLE;
      input_format_reg |= (VIDEO_IFORM_SVID_CHROMA | VIDEO_IFORM_SVIDEO);
   }
   else if (cap->in_input_type == CX88_VIDEO_INPUT_VIP)
   {
      input_format_reg |= VIDEO_IFORM_VIP_INPUT;
   }
   if (sc->format.in_x_size < 640)
   {
      filter_control_reg |= cx88_write_field(VIDEO_FILTER_COMB, VIDEO_COMB_CHROMA);
   }
   else
   {
      output_format_reg |= (VIDEO_OFORM_COMBALT | VIDEO_OFORM_PREV_REMOD);
      if (sc->format.in_video_format >= CX88_VIDEO_FMT_PALBDGHI)
      {
         output_format_reg |= VIDEO_OFORM_PAL_INV_PHASE;
      }
   }
   if (sc->format.in_even_field.enable && sc->format.in_odd_field.enable && 
      sc->format.in_interlace)
   {
      filter_control_reg |= VIDEO_FILTER_VINT;
   }
   if (sc->format.in_x_size < 193)
   {
      filter_control_reg |= cx88_write_field(VIDEO_FILTER_VFILT, VIDEO_VFILT_INTERP_5TAP);
   }
   else if (sc->format.in_x_size < 385)
   {
      filter_control_reg |= cx88_write_field(VIDEO_FILTER_VFILT, VIDEO_VFILT_INTERP_3TAP);
   }
   if (cap->in_input_options & CX88_VIDEO_INPUT_AGC_ENABLE)
   {
      input_format_reg |= VIDEO_IFORM_AGC_ENABLE;
   }
   if (!(cap->in_input_options & CX88_VIDEO_INPUT_AGC_COARSE))
   {
      input_format_reg |= VIDEO_IFORM_NORMAL_AGC;
   }
   if (cap->in_input_options & CX88_VIDEO_INPUT_AGC_ADAPTIVE)
   {
      input_format_reg |= VIDEO_IFORM_ADAPTIVE_AGC;
   }
   if (cap->in_input_options & CX88_VIDEO_INPUT_COLORKILLER)
   {
      input_format_reg |= VIDEO_IFORM_COLOR_KILLER;
   }
   input_format_reg |= cx88_write_field(VIDEO_IFORM_INPUT_SOURCE, cap->in_input_source);
   input_format_reg |= cx88_write_field(VIDEO_IFORM_FORMAT, sc->format.in_video_format);
   cx88_write_reg(&(sc->common), VIDEO_REG_FILTER_EVEN, filter_control_reg);
   cx88_write_reg(&(sc->common), VIDEO_REG_FILTER_ODD, filter_control_reg);
   cx88_write_reg(&(sc->common), VIDEO_REG_OUTPUT_FORMAT, output_format_reg);
   cx88_write_reg(&(sc->common), VIDEO_REG_INPUT_FORMAT, input_format_reg);
   uint32_t sample_freq = sc->spec->sample_freq;
   uint32_t fsc8 = sample_freq;
   if (cap->in_input_type == CX88_VIDEO_INPUT_VIP)
   {
      sample_freq = 27000000;
   }
   uint64_t temp = ((uint64_t)28636363 * (1 << 17)) / sample_freq;
   cx88_write_reg(&(sc->common), VIDEO_REG_SAMPLE_RATE_CONV, (uint32_t) temp);
   if (sc->format.in_video_format == CX88_VIDEO_FMT_SECAM)
   {
      temp = ((uint64_t)34000000 * (1 << 22)) / sample_freq;
      cx88_write_reg(&(sc->common), VIDEO_REG_SUBCARRIER_STEP, (uint32_t) temp);
      temp = ((uint64_t)35250000 * (1 << 22)) / sample_freq;
      cx88_write_reg(&(sc->common), VIDEO_REG_SUBCARRIER_STEP_DR, (uint32_t) temp);
   }
   else
   {   
      temp = ((uint64_t)fsc8 * (1 << 22)) / sample_freq;
      cx88_write_reg(&(sc->common), VIDEO_REG_SUBCARRIER_STEP, (uint32_t) temp);
      cx88_write_reg(&(sc->common), VIDEO_REG_SUBCARRIER_STEP_DR, (uint32_t) temp);
   }  
   uint32_t agc_burst_delay = cx88_write_field(VIDEO_BURST_DELAY, 
      ((65 * sample_freq) / 20000000) + 21);
   agc_burst_delay |= cx88_write_field(VIDEO_AGC_DELAY, ((68 * sample_freq) / 20000000) + 15);
   cx88_write_reg(&(sc->common), VIDEO_REG_AGC_BURST_DELAY, agc_burst_delay);
   temp = ((uint64_t)(sc->spec->pixels_per_line) * sample_freq) / fsc8;
   uint32_t htotal_reg = cx88_write_field(VIDEO_HTOTAL, (uint32_t)temp);
   if (cap->in_input_type == CX88_VIDEO_INPUT_VIP)
   {
      if (sc->format.in_video_format >= CX88_VIDEO_FMT_PALBDGHI)
      {
         htotal_reg |= cx88_write_field(VIDEO_HLOTCH, VIDEO_HLNOTCH_135PAL);
      }
      else
      {
         htotal_reg |= cx88_write_field(VIDEO_HLOTCH, VIDEO_HLNOTCH_135NTSC);
      }
   }
   cx88_write_reg(&(sc->common), VIDEO_REG_HLNOTCH_HTOTAL, htotal_reg);
   temp = ((uint64_t)(sample_freq) * 16 * (1 << 20)) / 28636363;
   cx88_write_reg(&(sc->common), VIDEO_REG_PLL,
      cx88_write_field(VIDEO_PLL_DIV, (uint32_t)temp));
   uint16_t ticks = 0;
   int error = ETIMEDOUT;
   do
   {
      if (cx88_read_reg(&(sc->common), VIDEO_REG_STATUS) & VIDEO_STATUS_PLL_LOCK)
      {
         error = 0;
         break;  
      }
      pause("cx88p", 10);
      ticks += 10;
   }
   while ((ticks < cap->in_lock_timeout_ms) && (ticks < 60000));
   if (error)
   {
      device_printf(sc->common.dev, "Unable to acquire signal\n");
      return error;
   }
   pause("cx88p", cap->in_lock_delay_ms);
   if (!(cx88_read_reg(&(sc->common), VIDEO_REG_STATUS) & VIDEO_STATUS_PLL_LOCK))
   {
      device_printf(sc->common.dev, "Lost signal lock\n");
      return EIO;
   }
   sc->state = CX88_VIDEO_CAPTURING;
   /* Protect access to the shared PCI interrupt and ADC config registers. */
   cx88_lock_board(&(sc->common));
   cx88_write_reg(&(sc->common), CX88_REG_PCI_INTMASK,
      (cx88_read_reg(&(sc->common), CX88_REG_PCI_INTMASK) | CX88_PCI_INT_VIDEO));
   if (cap->in_input_type == CX88_VIDEO_INPUT_SVIDEO)
   {
      cx88_write_reg(&(sc->common), CX88_REG_ADC_CONFIG,
         cx88_read_reg(&(sc->common), CX88_REG_ADC_CONFIG) & CX88_CHROMA_VIDEO); 
   }
   else
   {
      cx88_write_reg(&(sc->common), CX88_REG_ADC_CONFIG,
         cx88_read_reg(&(sc->common), CX88_REG_ADC_CONFIG) & (~CX88_CHROMA_VIDEO)); 
   }
   cx88_unlock_board(&(sc->common));
   uint32_t capture_control_reg = 0;
   if (sc->format.in_even_field.enable)
   {
      capture_control_reg |= VIDEO_CAPTURE_EVEN;
      if (sc->format.in_enable_vbi)
      {
         capture_control_reg |= VIDEO_CAPTURE_VBI_EVEN;
      }
   }
   if (sc->format.in_odd_field.enable)
   {
      capture_control_reg |= VIDEO_CAPTURE_ODD;
      if (sc->format.in_enable_vbi)
      {
         capture_control_reg |= VIDEO_CAPTURE_VBI_ODD;
      }
   }
   uint32_t dmac_reg = 0;
   uint32_t intr_mask = 
      VIDEO_INT_OPCODE_ERR | VIDEO_INT_PAR_ERR | VIDEO_INT_RIP_ERR | VIDEO_INT_PCI_ABORT;
   for (uint8_t i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      sc->channel_configs[i].even_frame_available = 0;
      sc->channel_configs[i].odd_frame_available = 0;
      if (sc->format.out_geoms[i].line_size != 0)
      {
         intr_mask |= (VIDEO_INT_OVERFLOW(i) | VIDEO_INT_SYNC(i));
         if (!(cap->in_no_frame_intrs))
         {
            intr_mask |= (VIDEO_INT_RISCI1(i) | VIDEO_INT_RISCI2(i));
         }
         dmac_reg |= (VIDEO_FIFO_ENABLE(i) | VIDEO_RISC_ENABLE(i));
      }
   } 
   cx88_write_reg(&(sc->common), VIDEO_REG_CAPTURE_CONTROL, capture_control_reg); 
   /* Acknowledge any outstanding interrupts */
   cx88_write_reg(&(sc->common), VIDEO_REG_INT_STAT,
      cx88_read_reg(&(sc->common), VIDEO_REG_INT_STAT));
   cx88_write_reg(&(sc->common), VIDEO_REG_INT_MASK, intr_mask);
   cx88_write_reg(&(sc->common), CX88_REG_DEVCNTL2, CX88_RUN_RISC);
   cx88_write_reg(&(sc->common), VIDEO_REG_DMA_CONTROL, dmac_reg);
   return 0;
}

/* Configures channel, buffers, and I2C devices */
static int cx88_video_configure(
   cx88_video_softc *sc,
   cx88_video_config *config,
   struct thread *context)
{
   int error = 0;
   uint8_t i;
   if (sc->state < CX88_VIDEO_FORMATTED)
   {
      return ENXIO;
   }
   cx88_video_reset(sc, CX88_VIDEO_FORMATTED); /* Ensure there is no outstanding capture */
   /* Validate ioctl parameters */
   if ((config->in_num_frames == 0) || (config->in_num_frames > CX88_VIDEO_MAX_FRAMES))
   {
      return EINVAL;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Configuring video channels\n"))
   uint8_t old_num_frames = sc->num_frames;
   sc->num_frames = (config->in_one_shot ? config->in_num_frames : 0);
   for (i = 0; (i < CX88_VIDEO_CHANNELS) && !error; ++i)
   {
      if (sc->format.out_geoms[i].line_size != 0)
      {
         cx88_dmareprogram_t reprogram = sc->dma_reprogram;
         if ((config->bufconfig[i].in_buftype != CX88_VIDEO_BUF_KERNEL) &&
             (config->bufconfig[i].in_buftype != CX88_VIDEO_BUF_USER))
         {
            error = EINVAL;  
         }
         if ((config->bufconfig[i].in_x_header != sc->channel_configs[i].x_header) ||
             (config->bufconfig[i].in_x_trailer != sc->channel_configs[i].x_trailer))
         {
            sc->channel_configs[i].x_header = config->bufconfig[i].in_x_header;
            sc->channel_configs[i].x_trailer = config->bufconfig[i].in_x_trailer;
            reprogram = CX88_REPROGRAM_ALWAYS;
         }
         size_t line_size = sc->format.out_geoms[i].line_size +
            (sc->channel_configs[i].x_header + sc->channel_configs[i].x_trailer);
         size_t buffer_size = line_size * sc->format.out_geoms[i].num_lines;
         if (!error)
         {
            /* Allocate the channel */
            error = cx88_configure_channel(
               &(sc->channels[i]),
               0,
               sc->format.out_geoms[i].line_size,
               sc->format.out_geoms[i].line_size,
               0);
         }
         if (!error)
         {      
            if (config->bufconfig[i].in_buftype == CX88_VIDEO_BUF_KERNEL)
            {   
               if ((config->bufconfig[i].in_y_header != sc->channel_configs[i].y_header) ||
                  (config->bufconfig[i].in_y_trailer != sc->channel_configs[i].y_trailer))
               {
                  sc->channel_configs[i].y_header = config->bufconfig[i].in_y_header;
                  sc->channel_configs[i].y_trailer = config->bufconfig[i].in_y_trailer;
                  reprogram = CX88_REPROGRAM_ALWAYS;
               }
               buffer_size += line_size * 
                  (sc->channel_configs[i].y_header + sc->channel_configs[i].y_trailer);
               /* Calculate the total amount of memory the application will need to mmap(), along
               * with the buffer size the application will need to use in calculating the offsets
               * of individual buffers within that mmap'ed region */
               config->bufconfig[i].u.out_kernelbuf.out_buffer_size = round_page(buffer_size);
               config->bufconfig[i].u.out_kernelbuf.out_mmap_size = 
                  (config->in_num_frames * config->bufconfig[i].u.out_kernelbuf.out_buffer_size);
               sc->channel_configs[i].mmap_size = config->bufconfig[i].u.out_kernelbuf.out_mmap_size;
            }
            else
            {
               sc->channel_configs[i].y_header = 0;
               sc->channel_configs[i].y_trailer = 0;
            }
            /* Request the buffer structures */
            error = cx88_request_buffers (&(sc->channels[i]), config->in_num_frames, 0);
         }
         uint8_t j;
         for (j = 0; (j < config->in_num_frames) && !error; ++j)
         {
            void *bufptr = NULL;
            struct thread *bufcontext = NULL;
            if (config->bufconfig[i].in_buftype == CX88_VIDEO_BUF_USER)
            {
               if (config->bufconfig[i].u.in_userbuf.buffers[j] == NULL)
               {
                  error = EINVAL;
                  break;  
               }
               /* Each cx88 buffer is a segment of the user-allocated buffer */
               bufptr = (uint8_t *)(config->bufconfig[i].u.in_userbuf.buffers[j]) + 
                  (line_size * config->bufconfig[i].in_y_header);
               bufcontext = context;
            }
            if ((old_num_frames != sc->num_frames) && 
                (((old_num_frames != 0) && (j == (old_num_frames - 1))) ||
                 ((sc->num_frames != 0) && (j == (sc->num_frames - 1)))))
            {
               reprogram = CX88_REPROGRAM_ALWAYS;   
            }
            /* Configure each buffer */
            error = cx88_configure_buffer(
               &(sc->channels[i].buffers[j]),
               bufptr, /* This will be NULL for kernel-allocated buffers */
               bufcontext,
               buffer_size,
               ((j == (config->in_num_frames - 1)) ? 0 : (j + 1)),
               cx88_video_dmaprog,
               reprogram);
         }
         if (!error)
         {
            /* Now commit all changes we've made to the channel configuration */
            error = cx88_commit_channel(&(sc->channels[i]));
         }
      }
      else
      {      
         cx88_release_channel (&(sc->channels[i])); /* Release all channel and buffer resources */
         sc->channel_configs[i].mmap_size = 0;
      }
   }
   if (!error)
   {
      sc->dma_reprogram = CX88_REPROGRAM_ON_BUFFER_CHANGE;
      sc->state = CX88_VIDEO_CONFIGURED;
   }
   else
   {
      cx88_video_release_channels(sc);
   }
   return error;
}

static void cx88_video_release_channels(cx88_video_softc *sc)
{
   uint8_t i;
   for (i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      cx88_release_channel (&(sc->channels[i])); /* Release all channel and buffer resources */
      sc->channel_configs[i].mmap_size = 0;
   }
}

/* Open file operation */
static int cx88_video_open(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{
   cx88_video_softc *sc = devclass_get_softc(cx88_video_devclass, minor(dev));
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Opening device\n"))
   sx_xlock(&(sc->sxlock));
   if (sc->in_use)
   {
      sx_xunlock(&(sc->sxlock));
      return EBUSY;
   }   
   if (sc->state == CX88_VIDEO_CONFIGURED)
   {
      cx88_i2c_request_devices(&(sc->i2c));
   }
   sc->in_use = TRUE;
   sx_xunlock(&sc->sxlock);
   return 0;
}

/* Close file operation */
static int cx88_video_close(
   struct cdev *dev,
   int flags,
   int fmt,
   struct thread *td)
{   
   cx88_video_softc *sc = devclass_get_softc(cx88_video_devclass, minor(dev));
   if (sc == NULL)
   {
      return ENXIO;
   }
   DIAGNOSTIC(device_printf(sc->common.dev, "Closing device\n"))
   sx_xlock(&(sc->sxlock));
   cx88_video_reset(sc, sc->state);
   uint8_t i;
   for (i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      /* If we're using user-allocated buffers, release them as
      * they can't be persisted across open() and close() in
      * different processes */
      cx88_free_userbufs(&(sc->channels[i]), NULL);
   }
   if (sc->reconfig_requested)
   {
      cx88_video_reconfigure(sc);
      sc->reconfig_requested = FALSE;
   }
   /* Release ownership of I2C devices */
   cx88_i2c_release_devices(&(sc->i2c));
   sc->in_use = FALSE;
   sx_xunlock(&(sc->sxlock));
   return 0;
}

/* Mmap file operation */
static int cx88_video_mmap(
   struct cdev *dev,
   vm_offset_t offset,
   vm_paddr_t *paddr,
   int nprot)
{
   cx88_video_softc *sc = devclass_get_softc(cx88_video_devclass, minor(dev));
   sx_slock(&(sc->sxlock));
   DIAGNOSTIC(device_printf(sc->common.dev, "Attempting to mmap device memory\n"))
   if (nprot & (PROT_EXEC | PROT_WRITE))
   {
      /* Video is output-only */
      DIAGNOSTIC(device_printf(sc->common.dev, "cx88_video_mmap: Invalid device memory permissions\n"))
      return EINVAL;
   }
   uint8_t i;
   int error = EINVAL;
   uint32_t total_mmap_size = 0;
   for (i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      if ((total_mmap_size + sc->channel_configs[i].mmap_size) > offset)
      {
         error = cx88_common_mmap(&(sc->channels[i]), offset - total_mmap_size, paddr, nprot);
         break;
      }
      total_mmap_size += sc->channel_configs[i].mmap_size;
   }
   sx_sunlock(&(sc->sxlock));
   return error;
}

/* Establish video frame geometry and pixel format */
static int cx88_video_geom(cx88_video_softc *sc, cx88_video_format *fmt)
{
   if (sc->state < CX88_VIDEO_UNCONFIGURED)
   {
      return ENXIO;
   }
   uint8_t i = CX88_FMT_TO_SPEC(fmt->in_video_format);
   if ((i >= CX88_VIDEO_SPECS) ||
       (!(fmt->in_even_field.enable) && !(fmt->in_odd_field.enable)) ||
       (fmt->in_rows == 0) || (fmt->in_cols == 0) ||
       (fmt->in_x_size == 0) || (fmt->in_y_size == 0) ||
       ((fmt->in_x_offset + fmt->in_x_size) > fmt->in_cols) ||
       ((fmt->in_y_offset + fmt->in_y_size) > fmt->in_rows))
   {
      return EINVAL;
   }
   sc->spec = &(specs[i]);
   uint32_t hscale = ((sc->spec->hactive * 4096) / fmt->in_cols) - 4096;
   uint32_t hdelay = ((fmt->in_cols * sc->spec->hdelay) / sc->spec->hactive) & 0x3FE;
   hdelay += fmt->in_x_offset;
   uint32_t vscale = 0x10000 - ((512 * sc->spec->vactive) / fmt->in_rows) + 512;
   if ((fmt->in_cols > sc->spec->hactive) || (hscale > 0xFFFF) ||
       (fmt->in_rows > sc->spec->vactive) || (vscale <= 0xE000))
   {
      return EINVAL;
   } 
   vscale &= 0x1FFF;
   uint32_t vactive = (fmt->in_y_size * sc->spec->vactive) / fmt->in_rows;
   uint32_t vdelay = sc->spec->vdelay + ((fmt->in_y_offset * sc->spec->vactive) / fmt->in_rows);
   if ((vdelay % 2) != 0)
   {
      vdelay -= 1;
   }
   for (i = CX88_VIDEO_CHANNEL_U; i < CX88_VIDEO_CHANNELS; ++i)
   {
      fmt->out_geoms[i].line_size = 0;
      fmt->out_geoms[i].num_lines = 0;
   }
   uint16_t num_lines = (vactive * fmt->in_rows) / sc->spec->vactive;
   uint16_t vbi_lines = sc->spec->vdelay;
   if (!(fmt->in_odd_field.enable && fmt->in_even_field.enable))
   {
      vbi_lines /= 2;
      if (!fmt->in_interlace)
      {
         num_lines /= 2;
      }
   }
   fmt->out_geoms[CX88_VIDEO_CHANNEL_Y].num_lines = num_lines;
   switch (fmt->in_pixel_format)
   {
      case CX88_VIDEO_RGB32:
         fmt->out_geoms[CX88_VIDEO_CHANNEL_Y].line_size = 4 * fmt->in_x_size;
         break;
      case CX88_VIDEO_BGR24:
         fmt->out_geoms[CX88_VIDEO_CHANNEL_Y].line_size = 3 * fmt->in_x_size;
         break;
      case CX88_VIDEO_RGB16:
      case CX88_VIDEO_RGB15:
      case CX88_VIDEO_YUV422:
         if (fmt->in_x_size % 2)
         {
            return EINVAL;  
         }
         fmt->out_geoms[CX88_VIDEO_CHANNEL_Y].line_size = 2 * fmt->in_x_size;
         break;
      case CX88_VIDEO_BTYUV411:
         if (fmt->in_x_size % 4)
         {
            return EINVAL;  
         }
         fmt->out_geoms[CX88_VIDEO_CHANNEL_Y].line_size = (3 * fmt->in_x_size) / 2;
         break;
      case CX88_VIDEO_Y8:
      case CX88_VIDEO_TGB8:
      case CX88_VIDEO_YCC422:
      case CX88_VIDEO_YCC411:
         fmt->out_geoms[CX88_VIDEO_CHANNEL_Y].line_size = fmt->in_x_size;
         break;
      default:
         return EINVAL;
   }
   if ((fmt->in_pixel_format == CX88_VIDEO_YCC422) || (fmt->in_pixel_format == CX88_VIDEO_YCC411))
   {
      if (!(sc->channel_configs[CX88_VIDEO_CHANNEL_U].enabled) ||
          !(sc->channel_configs[CX88_VIDEO_CHANNEL_V].enabled))
      {
         device_printf(sc->common.dev, "Planar capture disabled by device hint\n");
         return ENXIO;
      }
      fmt->out_geoms[CX88_VIDEO_CHANNEL_U].num_lines = num_lines;
      fmt->out_geoms[CX88_VIDEO_CHANNEL_V].num_lines = num_lines;
      uint8_t scale_factor = ((fmt->in_pixel_format == CX88_VIDEO_YCC422) ? 2 : 4);
      if (fmt->in_x_size % scale_factor)
      {
         return EINVAL;
      }
      fmt->out_geoms[CX88_VIDEO_CHANNEL_U].line_size = fmt->in_x_size / scale_factor;
      fmt->out_geoms[CX88_VIDEO_CHANNEL_V].line_size = fmt->in_x_size / scale_factor;  
   }
   if (fmt->in_enable_vbi)
   {
      if (!(sc->channel_configs[CX88_VIDEO_CHANNEL_VBI].enabled))
      {
         device_printf(sc->common.dev, "VBI capture disabled by device hint\n");
         return ENXIO;
      }
      fmt->out_geoms[CX88_VIDEO_CHANNEL_VBI].line_size = sc->spec->vbi_samples;
      fmt->out_geoms[CX88_VIDEO_CHANNEL_VBI].num_lines = vbi_lines;  
      cx88_write_reg(&(sc->common), VIDEO_REG_VBI_PACKET, 
         cx88_write_field(VIDEO_VBI_PACKET_SIZE, (sc->spec->vbi_samples / 4)) |
         cx88_write_field(VIDEO_VBI_VDELAY, 10));
   }
   if ((sc->state < CX88_VIDEO_FORMATTED) || 
       (sc->format.in_interlace != fmt->in_interlace) ||
       (sc->format.in_odd_field.enable != fmt->in_odd_field.enable) ||
       (sc->format.in_even_field.enable != fmt->in_even_field.enable))
   {
      sc->dma_reprogram = CX88_REPROGRAM_ALWAYS;
   }
   sc->format = *fmt;
   cx88_video_reset(sc, CX88_VIDEO_FORMATTED);
   cx88_write_reg(&(sc->common), VIDEO_REG_HSCALE_EVEN, hscale);
   cx88_write_reg(&(sc->common), VIDEO_REG_HSCALE_ODD, hscale);
   cx88_write_reg(&(sc->common), VIDEO_REG_VSCALE_EVEN, vscale);
   cx88_write_reg(&(sc->common), VIDEO_REG_VSCALE_ODD, vscale);
   cx88_write_reg(&(sc->common), VIDEO_REG_HDELAY_EVEN, hdelay);
   cx88_write_reg(&(sc->common), VIDEO_REG_HDELAY_ODD, hdelay);
   cx88_write_reg(&(sc->common), VIDEO_REG_VDELAY_EVEN, vdelay);
   cx88_write_reg(&(sc->common), VIDEO_REG_VDELAY_ODD, vdelay);
   cx88_write_reg(&(sc->common), VIDEO_REG_VDELAY_VIP_EVEN, vdelay);
   cx88_write_reg(&(sc->common), VIDEO_REG_VDELAY_VIP_ODD, vdelay);
   cx88_write_reg(&(sc->common), VIDEO_REG_HACTIVE_EVEN, fmt->in_x_size);
   cx88_write_reg(&(sc->common), VIDEO_REG_HACTIVE_ODD, fmt->in_x_size);
   cx88_write_reg(&(sc->common), VIDEO_REG_VACTIVE_EVEN, vactive);
   cx88_write_reg(&(sc->common), VIDEO_REG_VACTIVE_ODD, vactive);
   uint32_t color_format_reg = 0;
   if (fmt->in_even_field.enable)
   {
      if (fmt->in_even_field.swap & CX88_VIDEO_SWAP_WORDS)
      {
         color_format_reg |= VIDEO_COLOR_WSWAP_EVEN;
      }
      if (fmt->in_even_field.swap & CX88_VIDEO_SWAP_BYTES)
      {
         color_format_reg |= VIDEO_COLOR_BSWAP_EVEN;
      }
      color_format_reg |= cx88_write_field(VIDEO_COLOR_EVEN, fmt->in_pixel_format);
   }
   if (fmt->in_odd_field.enable)
   {
      if (fmt->in_odd_field.swap & CX88_VIDEO_SWAP_WORDS)
      {
         color_format_reg |= VIDEO_COLOR_WSWAP_ODD;
      }
      if (fmt->in_odd_field.swap & CX88_VIDEO_SWAP_BYTES)
      {
         color_format_reg |= VIDEO_COLOR_BSWAP_ODD;
      }
      color_format_reg |= cx88_write_field(VIDEO_COLOR_ODD, fmt->in_pixel_format);
   }
   cx88_write_reg(&(sc->common), VIDEO_REG_COLOR_FORMAT, color_format_reg);
   return 0;
}

static int cx88_video_set_clip(cx88_video_softc *sc, cx88_video_clip *clip)
{
   for (uint8_t i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      if (clip->channels[i].in_pattern_size == 0)
      {
         if (sc->channel_configs[i].clip_pattern_size != 0)
         {
            free(sc->channel_configs[i].clip_patterns, M_TEMP);
            sc->channel_configs[i].clip_patterns = NULL;
            sc->channel_configs[i].clip_pattern_size = 0;
            sc->dma_reprogram = CX88_REPROGRAM_ALWAYS;  
         }
         continue;
      }
      else if (clip->channels[i].in_patterns == NULL)
      {
         return EINVAL;  
      }
      void *patterns = realloc(sc->channel_configs[i].clip_patterns,
         clip->channels[i].in_pattern_size, M_TEMP, M_WAITOK);
      if (patterns == NULL)
      {
         return ENOMEM;
      }
      sc->channel_configs[i].clip_patterns = patterns;
      sc->channel_configs[i].clip_pattern_size = clip->channels[i].in_pattern_size;
      sc->dma_reprogram = CX88_REPROGRAM_ALWAYS;
      int error = copyin(clip->channels[i].in_patterns, sc->channel_configs[i].clip_patterns,
         sc->channel_configs[i].clip_pattern_size);
      if (error)
      {
         return error;
      }
   }
   return 0;
}

/* Ioctl file operation */
static int cx88_video_ioctl(struct cdev *dev, u_long cmd, caddr_t data, int flags, struct thread *td)
{
   cx88_video_softc *sc = devclass_get_softc(cx88_video_devclass, minor(dev));
   int error = 0;
   if (sc == NULL)
   {
      return ENXIO;
   }
   switch (cmd)
   {
      case CX88_COMMON_CARD_ID: /* Retrieve vendor-specific identifiers */
         *((uint32_t*)data) = cx88_get_card(sc->common.dev);
         break;
      case CX88_VIDEO_FORMAT: /* Establish video geometry */
         sx_xlock(&(sc->sxlock));
         error = cx88_video_geom(sc, (cx88_video_format*)data);
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_CONFIG: /* Configure video buffers */
         sx_xlock(&(sc->sxlock));
         error = cx88_video_configure(sc, (cx88_video_config*)data, td);
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_COMMON_I2C_CONFIG: /* Configure video buffers */
         sx_xlock(&(sc->sxlock));
         if (sc->state < CX88_VIDEO_UNCONFIGURED)
         {
            error = ENXIO;
         }
         else
         {
            error = cx88_i2c_update_devices(&(sc->i2c), (cx88_common_i2c_config*)data, IICDEV_NAME);
            if (!error)
            {
               error = cx88_i2c_request_devices(&(sc->i2c));
            }
         }
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_CAPTURE: /* Start capturing data */
         sx_xlock(&(sc->sxlock));
         error = cx88_video_start_capture(sc, (cx88_video_capture*)data);
         sx_xunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_STOP: /* Stop capturing data */
         /* Wait and stop are slock'ed so that a stop can be issued with a pending wait */
         sx_slock(&(sc->sxlock));
         cx88_video_reset(sc, sc->state);
         sx_sunlock(&(sc->sxlock));
         break;
      case CX88_VIDEO_WAIT: /* Wait for data to become available */
         sx_slock(&(sc->sxlock));
         error = cx88_video_transfer(sc, (cx88_video_wait*)data, ((flags & O_NONBLOCK) ? TRUE : FALSE));
         sx_sunlock(&(sc->sxlock));
         break;
      case CX88_COMMON_GPIO: /* Modify GPIO pin state */
         cx88_gpioctl(&(sc->common), (cx88_common_gpio*)data);
         break;
      case CX88_VIDEO_HUE:
         cx88_write_reg(&(sc->common), VIDEO_REG_HUE, (uint32_t)(*((uint8_t*)data)));
         break;
      case CX88_VIDEO_BRIGHTNESS:
         cx88_write_reg(&(sc->common), VIDEO_REG_BRIGHT_CONT,
            cx88_update_field(cx88_read_reg(&(sc->common), VIDEO_REG_BRIGHT_CONT), 
               VIDEO_BRIGHTNESS, (uint32_t)(*((uint8_t*)data))));
         break;
      case CX88_VIDEO_CONTRAST:
         cx88_write_reg(&(sc->common), VIDEO_REG_BRIGHT_CONT,
            cx88_update_field(cx88_read_reg(&(sc->common), VIDEO_REG_BRIGHT_CONT), 
               VIDEO_CONTRAST, (uint32_t)(*((uint8_t*)data))));
         break;
      case CX88_VIDEO_SAT_U:
         cx88_write_reg(&(sc->common), VIDEO_REG_UV_SAT,
            cx88_update_field(cx88_read_reg(&(sc->common), VIDEO_REG_UV_SAT), 
               VIDEO_USAT, (uint32_t)(*((uint8_t*)data))));
         break;
      case CX88_VIDEO_SAT_V:
         cx88_write_reg(&(sc->common), VIDEO_REG_UV_SAT,
            cx88_update_field(cx88_read_reg(&(sc->common), VIDEO_REG_UV_SAT), 
               VIDEO_VSAT, (uint32_t)(*((uint8_t*)data))));
         break;
      case CX88_VIDEO_DROP_FRAMES:
         cx88_write_reg(&(sc->common), VIDEO_REG_TDEC, 0);
         cx88_write_reg(&(sc->common), VIDEO_REG_TDEC, 
            cx88_write_field(VIDEO_TDEC_COUNT, (uint32_t)(*((uint8_t*)data))));
         break;
      case CX88_VIDEO_GET_STATUS:
      {
         cx88_video_status *status = (cx88_video_status*)data;
         uint32_t status_reg = cx88_read_reg(&(sc->common), VIDEO_REG_STATUS);
         status->out_video_present = (status_reg & VIDEO_STATUS_VPRES) ? 1 : 0;
         status->out_play_mode = cx88_read_field(status_reg, VIDEO_STATUS_NSPLAY);
         status->out_horizontal_lock = (status_reg & VIDEO_STATUS_HLOCK) ? 1 : 0;
         status->out_system_type = cx88_read_field(status_reg, VIDEO_STATUS_NUML);
         break;
      }
      case CX88_VIDEO_CLIP:
         sx_xlock(&(sc->sxlock));
         error = cx88_video_set_clip(sc, (cx88_video_clip*)data);
         sx_xunlock(&(sc->sxlock));
         break;
      default:
         error = ENOTTY;
         break;
   }
   return error;
}

/* Select/poll file operation */
static int cx88_video_poll(struct cdev *dev, int events, struct thread *td)
{   
   cx88_video_softc *sc = devclass_get_softc(cx88_video_devclass, minor(dev));
   if (sc == NULL)
   {
      return ENXIO;
   }
   int revents = 0;
   sx_xlock(&(sc->sxlock));
   mtx_lock(&(sc->isrlock));
   if ((sc->state == CX88_VIDEO_CAPTURING) && (events & (POLLIN | POLLRDNORM)))
   {
      if (sc->channels_avail)
      {
         /* Notify the application that data is available */
         revents = events & (POLLIN | POLLRDNORM);
      }
      else
      {
         selrecord(td, &(sc->selectors));
      }
   }
   mtx_unlock(&(sc->isrlock));
   sx_xunlock(&(sc->sxlock));
   return revents;
}

static void cx88_video_intr(void *arg)
{   
   cx88_video_softc *sc = arg;
   mtx_lock(&(sc->isrlock));
   /* Ensure we actually have an interrupt--use MSTAT register since we
    * only care about interrupts we've enabled */
   uint32_t intrs = cx88_read_reg(&(sc->common), VIDEO_REG_INT_MSTAT);
   if (intrs)
   {
      if (sc->state != CX88_VIDEO_CAPTURING)
      {
         mtx_unlock(&(sc->isrlock));
         return;  
      }
      if (intrs &  (VIDEO_INT_Y_SYNC | VIDEO_INT_U_SYNC | VIDEO_INT_V_SYNC | VIDEO_INT_VBI_SYNC))
      {
         sc->isr_status |= CX88_VIDEO_SYNC_ERROR;
      }
      /* Check for fatal errors */
      if (intrs & VIDEO_INT_OPCODE_ERR)
      {
         sc->isr_status |= CX88_VIDEO_DMA_PROGRAM;
      }
      if (intrs & (VIDEO_INT_PCI_ABORT | VIDEO_INT_PAR_ERR | VIDEO_INT_RIP_ERR))
      {
         sc->isr_status |= CX88_VIDEO_BUS_ERROR;
      }
      if (sc->isr_status & CX88_VIDEO_CAP_FATAL)
      {
         /* We should stop the transfer as soon as possible if we've encountered a fatal error */
         cx88_write_reg(&(sc->common), VIDEO_REG_DMA_CONTROL, 0);
      }
      for (uint8_t i = 0; i < CX88_VIDEO_CHANNELS; ++i)
      {
         if (intrs & (VIDEO_INT_RISCI1(i) | VIDEO_INT_RISCI2(i)))
         {
            /* The counter indicates the most recently completed buffer */
            uint32_t counter = cx88_read_reg(&(sc->common), VIDEO_REG_GP_COUNT(i));
            uint32_t bufindex = counter;
            if (sc->format.in_even_field.enable && sc->format.in_odd_field.enable)
            {
               bufindex /= 2;
            }
            if (bufindex < sc->channels[i].num_buffers)
            {
               /* Perform any necessary cache/bounce buffer synchronization.
               * Do this in a loop to account for the unlikely case where
               * we've missed an interrupt and multiple buffers need to be 
               * synchronized */
               uint8_t sync_bufindex = (sc->format.in_even_field.enable ? 
                  sc->channel_configs[i].even_frame_available : 
                  sc->channel_configs[i].odd_frame_available);
               do
               {
                  if (sync_bufindex >= sc->channels[i].num_buffers)
                  {
                     sync_bufindex = 0;
                  }
                  cx88_buffer *buffer = &(sc->channels[i].buffers[sync_bufindex]);
                  /*Synchronize the buffer */
                  cx88_mark_transfer_complete(buffer);
                  cx88_prepare_for_transfer(buffer);
               }
               while ((sync_bufindex++) != bufindex);
               if (intrs & VIDEO_INT_RISCI1(i))
               {
                  sc->channel_configs[i].odd_frame_available = bufindex + 1;
                  sc->channels_avail |= CX88_VIDEO_FIELD_ODD(i);              
               }
               if (intrs & VIDEO_INT_RISCI2(i))
               {
                  if (!(sc->format.in_odd_field.enable) || ((counter % 2) == 1))
                  {
                     sc->channel_configs[i].even_frame_available = bufindex + 1;
                  }
                  else if (bufindex == 0)
                  {
                     sc->channel_configs[i].even_frame_available = sc->channels[i].num_buffers;
                  }
                  else
                  {
                     sc->channel_configs[i].even_frame_available = bufindex;
                  }
                  sc->channels_avail |= CX88_VIDEO_FIELD_EVEN(i);
               }
            }
         }
      }
      /* Acknowledge the interrupt */
      cx88_write_reg(&(sc->common), VIDEO_REG_INT_STAT, intrs);
      if ((sc->channels_avail & sc->channel_mask) ||
          (sc->isr_status & CX88_VIDEO_CAP_FATAL))
      {
         wakeup(&(sc->channels_avail));
      }
      selwakeup(&(sc->selectors));
   }
   mtx_unlock(&(sc->isrlock));
}

static int cx88_video_lineprog(
   cx88_buffer *buffer,
   bus_size_t offset,
   cx88_video_clip_pattern *pattern)
{
   int error = 0;  
   uint16_t count = 0;
   boolean_t skip = FALSE;
   uint8_t i = 0;
   if ((pattern != NULL) && (pattern->in_clip_type == CX88_CLIP_EXCLUDE))
   {
      skip = TRUE;
   }
   do
   {
      uint16_t point = buffer->channel->line_size;
      if ((pattern != NULL) && (i < pattern->in_num_points))
      {
         point = pattern->in_points[i++];
      }
      if (point <= count)
      {
         device_printf(buffer->channel->sc->dev, "Out-of-order point %d in clipping pattern", point);
         return EINVAL;
      }
      if (point > buffer->channel->line_size)
      {
         point = buffer->channel->line_size;
      }
      uint16_t segcount = point - count;
      if (skip)
      {
         error = cx88_risc_skip_dmabuf(buffer, segcount, CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
      }
      else
      {
         error = cx88_risc_transfer_dmabuf (buffer, offset + count,
            segcount, CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
      }
      count += segcount;
      skip = !skip;
   } while (!error && (count < buffer->channel->line_size));
   return error;
}

static int cx88_video_fieldprog(
   cx88_buffer *buffer,
   cx88_video_channel_config *config,
   uint16_t *line,
   bus_size_t *offset,
   bus_size_t max_offset,
   bus_size_t line_size,
   uint8_t interlace)
{
   int error = 0;
   bus_size_t current_offset = *offset;
   uint16_t current_line = *line;
   cx88_video_clip_pattern *current_pattern = config->clip_patterns;
   uint32_t pattern_size = config->clip_pattern_size;
   while (!error && (current_offset < max_offset))
   {
      cx88_video_clip_pattern *pattern = NULL;
      while (pattern_size > sizeof(*current_pattern))
      {
         size_t current_size = 
            sizeof(*current_pattern) + (current_pattern->in_num_points * sizeof(uint16_t));
         if ((current_size > pattern_size) || (current_pattern->in_start_line > current_line))
         {
            break;    
         }
         else if (current_pattern->in_end_line >= current_line)
         {
            pattern = current_pattern;
            break;
         }
         current_pattern = (cx88_video_clip_pattern*)((uint8_t*)current_pattern + current_size);
         pattern_size -= current_size;
      }
      error = cx88_video_lineprog(buffer, current_offset, pattern);
      current_offset += (interlace * line_size);
      current_line += interlace;
   }
   *offset = current_offset;
   *line = current_line;
   return error;
}

static int cx88_video_dmaprog(cx88_buffer *buffer, uint8_t bufindex)
{
   int error = 0;
   cx88_video_softc *sc = (cx88_video_softc*) (buffer->channel->sc);
   size_t channel_index = buffer->channel - sc->channels;
   DIAGNOSTIC(device_printf(buffer->channel->sc->dev, "Generating DMA program\n"))
   if (channel_index >= CX88_VIDEO_CHANNELS)
   {
      return EINVAL;
   }
   cx88_video_channel_config *config = &(sc->channel_configs[channel_index]);
   bus_size_t line_size = buffer->channel->line_size + config->x_header + config->x_trailer;
   bus_size_t frame_start = config->y_header * line_size;
   bus_size_t frame_end = buffer->data_buffer_size - (line_size * config->y_trailer);
   bus_size_t current_offset = frame_start + config->x_header;
   uint16_t line_number = 0;
   uint8_t interlace = 
      ((sc->format.in_interlace && (buffer->channel->descriptor->number != CX88_CH_VIDEO_VBI)) ? 2 : 1);
   cx88_count_control countctl = ((bufindex == 0) ? CX88_RESET_COUNT : CX88_INCREMENT_COUNT);
   if (sc->format.in_odd_field.enable)
   {
      bus_size_t max_offset = frame_end;
      if (sc->format.in_even_field.enable && (interlace == 1))
      {
         max_offset = (frame_start + frame_end) / 2;
      }
      cx88_risc_sync(buffer, 0, TRUE, CX88_NO_TRIGGER, CX88_IGNORE_COUNT);
      error = cx88_video_fieldprog(buffer, config, &line_number,
         &current_offset, max_offset, line_size, interlace);
   }
   if (!error && (sc->format.in_even_field.enable))
   {
      if (sc->format.in_odd_field.enable)
      {
         cx88_risc_sync(buffer, 0x200, TRUE, CX88_TRIGGER_RISCI1, countctl);
         countctl = CX88_INCREMENT_COUNT;
      }
      else
      {
         cx88_risc_sync(buffer, 0x200, TRUE, CX88_NO_TRIGGER, CX88_IGNORE_COUNT);   
      }
      if (interlace == 2)
      {
         current_offset = ((config->y_header + 1) * line_size) + config->x_header;
         line_number = 1;
      }
      error = cx88_video_fieldprog(buffer, config, &line_number,
         &current_offset, frame_end, line_size, interlace);
   }
   if (!error)
   {
      if ((sc->num_frames != 0) && (bufindex == (sc->num_frames - 1)))
      {
         cx88_risc_write_const_to_reg(
            buffer,
            0,
            VIDEO_FIFO_ENABLE((buffer->channel->descriptor->number - CX88_CH_VIDEO_PACKED)),
            VIDEO_REG_DMA_CONTROL,
            CX88_NO_TRIGGER,
            CX88_IGNORE_COUNT);
      }
      
      /* Jump to next buffer, adjust the counter to reflect the index of the completed
      * buffer, and trigger RISC interrupt 1 so that the buffer can be processed */
      error = cx88_risc_next_buffer(
         buffer,
         0,
         (sc->format.in_even_field.enable ? CX88_TRIGGER_RISCI2 : CX88_TRIGGER_RISCI1),
         countctl);
   }
   return error;
}

static int cx88_video_probe(device_t dev) 
{
   return cx88_common_probe(dev, &cx88_video_client);
}

static int cx88_video_attach(device_t dev) 
{
   return cx88_common_attach(dev, &cx88_video_client);
}

/* I2C bus accessor DEVMETHOD */
static cx88_i2c_bus *cx88_video_get_i2c_bus(device_t dev)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx88_video_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return cx88_i2c_get_bus(&(sc->i2c));
}

/* I2C device accessor DEVMETHOD */
static cx88_i2c_dev *cx88_video_get_i2c_dev(device_t dev, uint8_t addr)
{
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx88_video_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return NULL;
   }
   return cx88_i2c_get_dev(&(sc->i2c), addr);
}

static void cx88_video_reconfigure(cx88_video_softc *sc)
{
   uint8_t i;
   for (i = 0; i < CX88_VIDEO_CHANNELS; ++i)
   {
      cx88_free_sram (&(sc->channels[i]));
   }
   if (sc->state > CX88_VIDEO_FORMATTED)
   {
      sc->state = CX88_VIDEO_FORMATTED;
   }
}

static int cx88_video_request_reconfig(device_t dev)
{   
   int error = 0;
   /* Siblings may call this at arbitrary times, so use cx88_get_softc()
    * to ensure we're in a valid & initialized state. */
   cx88_video_softc *sc = cx88_get_softc(dev);
   if (sc == NULL)
   {
      return ENODEV;
   }
   /* Since this may be called from a sibling driver in a different thread
    * than any thread that is doing I/O with this driver, use trylock
    * a) because blocking on the lock could deadlock and
    * b) because the sibling wouldn't want to wait an indeterminate amount
    *    of time for I/O on this driver to complete */
   if (sx_try_xlock(&(sc->sxlock)))
   {
      if (sc->in_use)
      {
         /* If someone has a handle open to our device,
          * don't disrupt their session.  We'll just free our resources
          * when the handle is closed that they can be re-allocated
          * the next time a handle is opened. */
         sc->reconfig_requested = TRUE;
         error = EINPROGRESS;
      }
      else
      {
         cx88_video_reconfigure(sc);  
      }
      sx_xunlock(&(sc->sxlock));
   }
   else
   {
      sc->reconfig_requested = TRUE;
      error = EINPROGRESS;
   }
   return error; 
}

/* Shutdown DEVMETHOD */
static int cx88_video_shutdown(device_t dev)
{
   cx88_video_softc *sc = device_get_softc(dev);
   cx88_video_reset(sc, CX88_VIDEO_INVALID);
   return 0;
}

/* Suspend DEVMETHOD */
static int cx88_video_suspend(device_t dev)
{
   int error = 0;
   cx88_video_softc *sc = device_get_softc(dev);
   sx_slock(&(sc->sxlock));
   cx88_video_reset(sc, CX88_VIDEO_UNCONFIGURED);
   sx_sunlock(&(sc->sxlock));
   sx_xlock(&(sc->sxlock));
   cx88_video_release_channels(sc);
   sx_xunlock(&(sc->sxlock));
   return error;
}

static device_method_t cx88_video_methods[] = 
{
   /* Device interface */
   DEVMETHOD(device_probe,             cx88_video_probe),
   DEVMETHOD(device_attach,            cx88_video_attach),
   DEVMETHOD(device_detach,            cx88_common_detach),
   DEVMETHOD(device_shutdown,          cx88_video_shutdown),
   DEVMETHOD(device_suspend,           cx88_video_suspend),
   DEVMETHOD(device_resume,            cx88_common_resume),

   /* cx88 interface */
   CX88_METHODS(cx88_video_request_reconfig),

   /* i2c host methods */
   CX88_I2C_METHODS(cx88_video_get_i2c_bus, cx88_video_get_i2c_dev),

   {NULL, NULL}
};

static driver_t cx88_video_driver =
{
   "cx88video",
   cx88_video_methods,
   0
};

DRIVER_MODULE(cx88video, pci, cx88_video_driver, cx88_video_devclass, NULL, NULL);
DRIVER_MODULE(cx88video, cardbus, cx88_video_driver, cx88_video_devclass, NULL, NULL);
CX88_DRIVER_MODULE(cx88video);
CX88_I2C_DRIVER_MODULE(cx88video);
MODULE_DEPEND(cx88video, iicdev, IICDEV_MINVER, IICDEV_MODVER, IICDEV_MAXVER);
MODULE_VERSION(cx88video, 1);
